#!/bin/sh
#-*- mode: Tcl;time-stamp-start:"TimeStamp[ 	]+\\\\?[\"<]+";-*-
# the next line restarts using tclsh \
exec tclsh $0 $@
set TimeStamp "2008-12-29 19:18:14 poser"
# Copyright (C) 2005-2008 William J. Poser (billposer@alum.mit.edu)
# This program is free software; you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as published by
# the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is contained in the
# procedure "License" in this file.
# If it is not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# or go to the web page:  http://www.gnu.org/licenses/gpl.txt.

set Version "2.4";
set DebugP 0;
set SynchJournalP 0;
set VerboseP 1;
package require msgcat
#The GUI requires Tk and Iwidgets but we defer requiring them since
#we don't use them in batch mode.
proc _ {s} {return [::msgcat::mc $s]};	# Define shorthand for gettext
proc ProgramTimeDateStamp {} {
    set sts [split $::TimeStamp]
    return "[lindex $sts 0] [lindex $sts 1]"
}

set InitFile     ".xlitrc";
set JournalFile  ".xlitlog";
set NonBinPath [file join /usr share Xlit];

proc dmsg {msg} {
    if {$::DebugP} {
	puts stderr $msg;
    }
}

#Portability
#Figure out what system we are running on
if {[string equal $tcl_platform(platform) windows]} {
    set System MSWindows;
    dmsg "Running under MS Windows";
} elseif {[string equal $tcl_platform(platform) unix]} {
    if {[string equal $tcl_platform(os) Darwin]} {
	set System MacOSX;
	dmsg "Running under Mac OS X";
    } else {
	set System Unix;
	dmsg "Running under Unix";
    }
}

if {[string match $System MSWindows]} {
    set NonBinPath [file dirname [info script]];
    set ::env(CYGWIN) "";
    set InitFile "XlitInit";
    set JournalFile "XlitLog";
}

if {[info exists ::env(TMP)]} {
    set TempDir $::env(TMP);
} elseif {[info exists ::env(TEMP)]} {
    set TempDir $::env(TEMP);
} else {
    if { $tcl_platform(platform) == "windows"} {
	set TempDir "C:\tmp";
    } else {
	set TempDir "/tmp";
    }
}

set ExternTempFile [file join $TempDir "xlitext.tmp"]

proc SetupEvents {sys} {
    switch $sys {
	Unix {
	    event add <<B3>> <ButtonPress-3>
	    event add <<B3Release>> <ButtonRelease-3>
	}
	MacOSX {
	    event add <<B3>> <Control-ButtonPress-1>
	    event add <<B3Release>> <Control-ButtonRelease-1>
	}
	MSWindows {
	    event add <<B3>> <ButtonPress-3>
	    event add <<B3Release>> <ButtonRelease-3>
	}
    }
}

proc DetermineGraphicsSystem {} {
    global AquaP
    global System
    global WindowSystem

    if {[string match X11*  [winfo server .]]} {
	set AquaP 0
	set WindowSystem X11
    } else {
	if {[string match $System MSWindows]} {
	    set AquaP 0;
	    set WindowSystem MSWindows;
	}
	if {[string match $System MacOSX]} {
	    set AquaP 1
	    set WindowSystem Aqua
	}
    }
}

proc CheckTkEightFivePlusOnX11 {} {
    set VLevelList [split [info patchlevel] "."]
    set Major [lindex $VLevelList 0]
    set Minor [lindex $VLevelList 1]
    if {$::WindowSystem eq "X11"} {
	if {$Major > 8} {
	    return 1
	} else {
	    if {($Major == 8) && ($Minor >= 5)} {
		return 1
	    }
	} 
    }
    return 0
}

proc LoadMessageCatalog {} {
    global DebugP;

    if { $DebugP } {
	if { [::msgcat::mcload "msgs"] == 0} {
	    puts stderr "No message catalog loaded."
	} else {
	    puts stderr "Message catalog loaded."
	}
    } else  {
	::msgcat::mcload [file join [file dirname [info script]] msgs];
    }
}

set DefaultBrowser dillo
#For this purpose dillo is in many respects arguably the best choice because
#it is very lightweight and starts up fast. However, it isn't as fully
#featured as the others, and in particular, doesn't handle the #full range of
#Unicode characters.
set BrowserList [list firefox mozilla epiphany galeon konqueror dillo netscape opera]
set BrowserPIDS [list];
set MSG     .mf.msg;
set PBAR    .mf.pbar 
set CTRLENT1 .cf.ctr.ebx.ent1
set CTRLENT2 .cf.ctr.ebx.ent2
set XLINP   .cf.inp
set XLOUT   .cf.out
set TXTINP  .txt.inp
set TXTOUT  .txt.out
set TXTISB  .txt.isbar
set TXTOSB  .txt.osbar

set MainWidth  10;
set MainHeight 6;
set HPWidth 64;				# Width of help popups.
set HPLines 8;				# Number of lines for help popups.

set ParseErrorCnt 0;

set BalloonHelpP 1;
set BatchModeP 0;
set IPACIsDisplayedP 0;
set IPADIsDisplayedP 0;
set IPAVIsDisplayedP 0;
set IPAAIsDisplayedP 0;
set CharEntryByCodeIsDisplayedP 0;
set DisplayEditingControlsP 0;
set StrictParseP 0;
set ReadInitFileP 1;
set CustomChartList "";
set TkPresentP 0;
#This determines whether we use the internal map (0),
#an external command (1), or a plugin (2)
set WhichProcedure 0
set PluginName "";
set PluginPath "";
set ExternalCommandList ""
set InputText ""

set ColorSpecs($XLINP,Background) 		"\#DDDDFF";
set ColorSpecs($XLOUT,Background) 		"\#DDDDFF";
set ColorSpecs(Alert,Background)		"\#c36176";
set ColorSpecs(Alert,Foreground)		"\#FFFFFF";
set ColorSpecs(AlertDismiss,Background)	"\#2200FF";
set ColorSpecs(AlertDismiss,Foreground)	"\#FFCC00";
set ColorSpecs(Backdrop,Background)		salmon;
set ColorSpecs(Button,bg)		salmon
set ColorSpecs(CharacterEntry,Background)		"\#FFFFFF";
set ColorSpecs(CharacterEntry,Foreground)		"\#000000";
set ColorSpecs(CharacterEntryHeadings,Background)		"\#09ffb2";
set ColorSpecs(ConfirmationDialogue,Background)	 		"\#DFC0C0";
set ColorSpecs(ConfirmationDialogueLabel,Background)	 		"\#c36176";
set ColorSpecs(ConfirmationDialogueLabel,Foreground)	 		"\#DFC0C0";
set ColorSpecs(ConfirmationDialogueTitle,Background)	 		"\#c36176";
set ColorSpecs(Default,Background)		"\#d4e4fe";
set ColorSpecs(Entry,Background)	"navajo white"
set ColorSpecs(Entry,DisabledBackground)		blue
set ColorSpecs(General,Background)	 		"\#c36176";
set ColorSpecs(IPAHeadings,Background)		"\#09ffb2";
set ColorSpecs(LabelText.inp,Background)	 	"\#FFDDDD";
set ColorSpecs(LabelText.out,Background)	 	"\#FFDDDD";
set ColorSpecs(MainFrame,Background) "\#4400FF"
set ColorSpecs(Match,Background)		"\#FFBBDD";
set ColorSpecs(Menu,ActiveBackground)	        NavajoWhite
set ColorSpecs(Menu,ActiveForeground)	        maroon
set ColorSpecs(Menu,Background)			azure3
set ColorSpecs(Menu,Foreground)			black
set ColorSpecs(Menu,Select)			coral
set ColorSpecs(Menubar,ActiveBackground)	salmon
set ColorSpecs(Menubar,ActiveForeground)	black
set ColorSpecs(Menubar,Background)	 	"\#c36176";
set ColorSpecs(Menubar,Foreground)		"\#fee4a9";
set ColorSpecs(Message,Background) 		"\#e9c4dc";
set ColorSpecs(Message,Foreground) 		"\#000000";
set ColorSpecs(PopupWidgetDefault,Background)    "\#fee4a9";
set ColorSpecs(PopupWidgetDefault,Foreground)    "\#fee4a9";
set ColorSpecs(PopupWidgetButton,Background)    "\#FBAA9A"
set ColorSpecs(PopupWidgetButton,Foreground)    black
set ColorSpecs(PopupWidgetCheckbutton,Foreground)    black
set ColorSpecs(PopupWidgetEntry,Background)    "\#BBBBFF"
set ColorSpecs(PopupWidgetEntry,Foreground)    black
set ColorSpecs(PopupWidgetLabel,Background)    "\#fee4a9";
set ColorSpecs(PopupWidgetLabel,Foreground)    black
set ColorSpecs(RadioButton,ActiveBackground) "\#FF8080"
set ColorSpecs(RadioButton,Background) "\#A0A0FF"
set ColorSpecs(RadioButton,SelectedBackground) red
set ColorSpecs(Substitutions,Background)		"\#FFFFFF";
set ColorSpecs(InputText,Background)	 		"\#88FF88"
set ColorSpecs(OutputText,Background)	 		"\#88FF88"
set ColorSpecs(TransliterationIn,Background)	 		"\#DDDDFF";
set ColorSpecs(TransliterationOut,Background)	 		"\#DDDDFF";
set ColorSpecs(TransliterationIn,ActiveBackground)	 		"\#DDDDFF";
set ColorSpecs(TransliterationOut,ActiveBackground)	 		"\#DDDDFF";
set ColorSpecs(TransliterationIn,Foreground)	 		"\#333300";
set ColorSpecs(TransliterationOut,Foreground)	 		"\#333300";
set ColorSpecs(TransliterationLineNumbers,Background)	NavajoWhite 
set ColorSpecs(TransliterationLineNumbers,Foreground)	black
set ColorSpecs(UserText,Foreground)		\#0000FF
set ColorSpecs(UserTextEntry,Background)	"\#FFFFFF";
set ColorSpecs(UserTextEntry,Foreground)	black

#This is a tab, which is the normal delimiter.
#The user is presented with the option of using 0xFDDF
#instead. This is guaranteed not be a character, so it
#can be safely used without interference with the content
#of the transliteration. On the other hand, it is more
#difficult to handle since it cannot be displayed and
#cannot be inserted by normal editing operation.
set TransliterationDelimiter "\u0009";

#Defaults
set FoldCaseP 1;
set MapLongestFirstP 0;
set SideBySideLayoutP 0;
set UseScrollbarsP 1;
set CommandLineTransliterateWhat ""
set CommandLineDuplicateSelection ""
set DefaultCommentCharacter \#;
set WhichFontToSet MainFont
# Determines whether to duplicate embedded region and if so where
#  0 do not duplicate
#  1 duplicate before transliteration
#  2 duplicate after transliteration
set DuplicateSelection   0;
set DelimitDuplicateP 0;
# What sort of delimiters to put around duplicate
set DuplicateDelimiterChoice 0;
set DupPairedDelimiter 1;

#This variable controls what gets transliterated.
#  1  transliterate the entire input
#  2  transliterate embedded domains only
#  3  transliterate matrix domains only
set XLWhat 1;
#ParseChoice controls how if at all the input text
#is parsed into domains
#  0  do not parse - implies XLWhat = 1
#  1  use HZ escapes
#  2  use simple paired delimiters (e.g. <>)
#  3  use simple unpaired delimiters  (e.g. ||)
#  4  XML
#  5  use paired strings
set ParseChoice 0;
set OutputDelimiterChoice 0;
array set Parsers {
    1 HZParse
    2 PairedSingleParse
    3 UnpairedSingleParse
    4 XMLParse
    5 PairedStringParse
}

set XMLTag "";
set DuplicateXMLTag "";
set OutputXMLTag "";
set UnpairedDelimiter "";
set DuplicateUnpairedDelimiter "";
set OutputUnpairedDelimiter "";
set PairedDelimiter 1;
set OutputPairDelimiters 1;
set PairedStringOpen "";
set PairedSTringClose "";
set DuplicatePairedStringOpen "";
set DuplicatePairedSTringClose "";
set OutputPairedStringOpen "";
set OutputPairedStringClose "";

set ResultGenTime  0;
set ResultSaveTime 0;
set XLGenTime      0;
set XLSaveTime     0;
array set CommandGlosses {};
set XLRulesLoaded 0;
set XLRulesInWindows 0;
set wret 0;
set TopEntry 1;
set CommandLineTextFile "";
set CommandLineTransliterationDefinition "";
set CommandLineLHSTransliterationDefinition "";
set CommandLineParseChoice "";

set PairedStringOpen  "";
set PairedStringClose "";
set DuplicatePairedStringOpen  "";
set DuplicatePairedStringClose "";

set Locale [::msgcat::mclocale];
LoadMessageCatalog;

#Shut down gracefully.
proc ShutDown {} {
    global BrowserPIDS;
    if {[$::TXTOUT get 1.0 end] != ""} {
	if {$::ResultSaveTime < $::ResultGenTime} {
	    set resp [ConfirmationDialog .sdt \
			  [_ "The Transliterated Text Has Not Been Saved"]\
			  [_ "Save Transliterated Text?"] [_ "No, Don't Save It"] [_ "Yes, Save It"]\
			  [_ "Click here if you don't really want to quit after all."]]
	    switch -exact -- $resp {
		1 {SaveOutput}
		-1 {return}
		default {}
	    }
	}
    }
    GetTransliterationFromWindows;
    if {$::XLRulesInWindows != 0} {
	if {$::XLSaveTime < $::XLGenTime} {
	    set resp [ConfirmationDialog .sdx \
			  [_ "The Transliteration Has Not Been Saved"]\
			  [_ "Save Transliteration?"] [_ "No, Don't Save It"] [_ "Yes, Save It"]\
			  [_ "Click here if you don't really want to quit after all."]]
	    switch -exact -- $resp {
		1 {SaveTransliteration.}
		-1 {return}
		default {}
	    }
	}
    }
    WriteJournal [format [_ "Session ended:   %s"] [clock format [clock seconds]]];
    CloseJournal;
    #Shutdown child browsers
    foreach pid $BrowserPIDS {
	dmsg "$pid"
	catch {exec kill $pid};
    }
    exit 0;
}


#Internationalized version of tk_optionMenu.
proc OptionMenu {w varName firstValue args} {
    upvar #0 $varName var

    if {![info exists var]} {
	set var $firstValue
    }
    menubutton $w -text [::msgcat::mc $firstValue] -indicatoron 1 -menu $w.menu \
	    -relief raised -bd 2 -highlightthickness 2 -anchor c \
	    -direction flush
    menu $w.menu -tearoff 0
    $w.menu add radiobutton -label [::msgcat::mc $firstValue] -value $firstValue -variable $varName \
	    -command "UpdateOptionMenuLabel $w $varName"
    foreach i $args {
    	$w.menu add radiobutton -label [::msgcat::mc $i] -value $i -variable $varName\
	    -command "UpdateOptionMenuLabel $w $varName"
    }
    return $w.menu
}

proc DelimitNumber {number {delim ","} {GroupSize 3}} {
     # First, extract right hand part of number, up to and including decimal point
     set point [string last "." $number];
     if {$point >= 0} {
         set PostDecimal [string range $number [expr $point + 1] end];
         set PostDecimalP 1;
     } else {
         set point [expr [string length $number] + 1]
         set PostDecimal "";
         set PostDecimalP 0;
     }

     # Now extract any leading spaces.
     set ind 0;
     while {[string equal [string index $number $ind] \u0020]} {
         incr ind;
     }
     set FirstNonSpace $ind;
     set LastSpace [expr $FirstNonSpace - 1];
     set LeadingSpaces [string range $number 0 $LastSpace];

     # Now extract the non-fractional part of the number, omitting leading spaces.
     set MainNumber [string range $number $FirstNonSpace [expr $point -1]];

     # Insert commas into the non-fractional part.
     set Length [string length $MainNumber];
     set Phase  [expr $Length % $GroupSize]
     set PhaseMinusOne  [expr $Phase -1];
     set DelimitedMain "";

     #First we deal with the extra stuff.
     if {$Phase > 0} {
         append DelimitedMain [string range $MainNumber 0 $PhaseMinusOne];
     }
     set FirstInGroup $Phase;
     set LastInGroup [expr $FirstInGroup + $GroupSize -1];
     while {$LastInGroup < $Length} {
         if {$FirstInGroup > 0} {
             append DelimitedMain $delim;
         }
         append DelimitedMain [string range $MainNumber $FirstInGroup $LastInGroup];
         incr FirstInGroup $GroupSize
         incr LastInGroup  $GroupSize
     }

     # Reassemble the number.
     if {$PostDecimalP} {
         return [format "%s%s.%s" $LeadingSpaces $DelimitedMain $PostDecimal];
     } else {
         return [format "%s%s" $LeadingSpaces $DelimitedMain];
     }
 }

proc UpdateOptionMenuLabel {w v} {
    upvar #0 $v x
    $w configure -text [::msgcat::mc $x]
}

#This provides an addition to the default methods of the Scrollbar
#widget. The default method, bound to <B1>, #increments or decrements
#the value by the resolution, by default 1. If you want a larger
#increment, you can use a larger resolution, but sometimes it is
#convenient to have a finer resolutio available while also being
#able to move in larger increments. A solution is to bind to
#<B3>, or some other event if you wish, a larger movement.
#This procedure moves by the specified multiplier k times
#the resolution of the Scrollbar. 

proc ScrollbarMoveBigIncrement {w f x y} {
    set part [$w identify $x $y]
    switch -exact -- $part {
	trough1 {
	    set dir -1;
	}
	arrow1 {
	    set dir -1;
	}
	trough2 {
	    set dir  1;
	}
	arrow2 {
	    set dir  1;
	}
	default {
	    return ;
	}
    }
    set CurrentFraction [lindex [$w get] 0]
    set NewFraction [expr $CurrentFraction + ($dir * $f)]
    eval [concat [$w cget -command] moveto $NewFraction]
}

proc ScaleMoveBigIncrement {w k x y} {
    set part [$w identify $x $y]
    set inc 0;
    switch -exact -- $part {
	trough1 {
	    set dir -1;
	}
	trough2 {
	    set dir  1;
	}
	default {
	    return ;
	}
    }
    set Resolution [$w cget -resolution]
    set CurrentValue [$w get]
    set Delta [expr $dir * $k * $Resolution]
    $w set [expr $CurrentValue + $Delta]
}

#Font stuff begins here
# Default defaults
set FontInfo(family) courier
set FontInfo(size) 12
set FontInfo(weight) normal
set FontInfo(slant) roman
set FontInfo(underline) 0
set FontInfo(overstrike) 0

namespace eval fontsel {
    variable FontPropertyList [list family size weight slant underline overstrike]
    variable FontPropertyClass
    variable FSCColors

    foreach p $FontPropertyList {
	set FontPropertyClass($p) generic;
    }
    set FontPropertyClass(underline) boolean
    set FontPropertyClass(overstrike) boolean

    #We check the existence of the variable so that this
    #can safely be called AFTER non-default values are set.
    proc SetFontInfoDefaults {} {
	global FontInfo;
	foreach ft $::FontList {
	    if {![info exist FontInfo($ft,family)]} {
		set FontInfo($ft,family)	$FontInfo(family)
	    }
	    if {![info exist FontInfo($ft,size)]} {
		set FontInfo($ft,size)	$FontInfo(size)
	    }
	    if {![info exist FontInfo($ft,weight)]} {
		set FontInfo($ft,weight)	$FontInfo(weight)
	    }
	    if {![info exist FontInfo($ft,slant)]} {
		set FontInfo($ft,slant)	$FontInfo(slant)
	    }
	    if {![info exist FontInfo($ft,underline)]} {
		set FontInfo($ft,underline)	$FontInfo(underline)
	    }
	    if {![info exist FontInfo($ft,overstrike)]} {
		set FontInfo($ft,overstrike)	$FontInfo(overstrike)
	    }
	    if {![info exist FontInfo($ft,gloss)]} {
		set FontInfo($ft,gloss)	$ft;
	    }
	    if {![info exist FontInfo($ft,help)]} {
		set FontInfo($ft,help)	"?"
	    }
	}
    }

    #For a given font assign font-specific properties from generics.
    proc FontSet {which} {
	global FontInfo
	variable FontPropertyList
	foreach prop $FontPropertyList {
	    set FontInfo($which,$prop) $FontInfo($prop);
	}
    }

    #Assign the proprety values of a specified font to the generics.
    proc InverseFontSet {which} {
	global FontInfo
	variable FontPropertyList
	foreach prop $FontPropertyList {
	    set FontInfo($prop) $FontInfo($which,$prop);
	}
    }

    #Configure a specified font according to the font-specific properties.
    proc ConfigureFont {which} {
	global FontInfo
	variable FontPropertyList
	foreach prop $FontPropertyList {
	    font configure $which -$prop $FontInfo($which,$prop)
	}
    }

    #Record as default values the font-specific values of properties at the
    #time of the call. The recorded values are intended for use by
    #ResetToDefaults 
    proc RecordDefaults {} {
	global FontInfo
	global FontList;
	variable FontPropertyList

	foreach f $FontList {
	    foreach p $FontPropertyList {
		set FontInfo($f,$p,Default) $FontInfo($f,$p);
	    }
	}
    }

    #Create all of the fonts on the font list and configure them
    #according to the generic values of the properties, which are
    #also assigned at this time to the font-specific variables.
    proc CreateFonts {} {
	global FontList;

	foreach ft $FontList {
	    font create $ft;
	    InverseFontSet $ft;
	    ConfigureFont $ft;
	}
    
    }

    #Get the values from the three scales, combine them into a hex RGB spec,
    #assign its value to FontExampleColor, and configure the two examples.
    proc UpdateFontExampleColor {x} {
	variable FSCColors
	global FontExampleColorConfigureWhich

	UpdateFontExampleColorHex;
	.selectFont.curcan.att.r1c1 configure \
	    -$::FontExampleColorConfigureWhich  \#$::FontExampleColor
	.selectFont.curcan.att.r2c1 configure \
	    -$::FontExampleColorConfigureWhich \#$::FontExampleColor
	set FSCColors(red,$FontExampleColorConfigureWhich)   $::FontRed;
	set FSCColors(green,$FontExampleColorConfigureWhich) $::FontGreen;
	set FSCColors(blue,$FontExampleColorConfigureWhich)  $::FontBlue;
    }

    proc UpdateFontExampleColorHex {} {
	set ::FontExampleColor [format "%02X%02X%02X" $::FontRed $::FontGreen $::FontBlue];
    }

    proc RestoreCSColors {} {
	variable FSCColors
	global FontExampleColorConfigureWhich
	set ::FontRed   $FSCColors(red,$FontExampleColorConfigureWhich)
	set ::FontGreen $FSCColors(green,$FontExampleColorConfigureWhich)
	set ::FontBlue  $FSCColors(blue,$FontExampleColorConfigureWhich)
	UpdateFontExampleColorHex;
    }

    proc CloneMainFontAsFontControlPanelFont {} {
	variable FontPropertyList
	global FontInfo

	foreach p $FontPropertyList {
	    font configure FontControlPanelFont -$p $FontInfo(MainFont,$p);
	}
    }

    #This is the main procedure of this package.
    #It creates the font control panel.
    proc CreateFontControlPanel {args} {
	global FontInfo;
	global WhichFontToSet
	global FontList;
	variable FontPropertyList
	variable FSCColors

	set General 1
	set exwid 10

	set w .selectFont
	if {[winfo exists $w]} {
	    wm deiconify $w;
	    raise $w;
	    return ;
	}
	toplevel $w
	wm title $w [_ "Font Selection"]
	BindKeys $w
	wm protocol $w WM_DELETE_WINDOW "wm withdraw $w"

	#We create this separately because we don't want it on the FontList.
	font create FontControlPanelFont;
	
	#The idea is that in general we want to use the main
	#font here, but we keep them separate so that we
	#can wait before switching the control panel to
	#the new main font configuration so that
	#the user isn't confronted with an unusable
	#control panel if he or she sets the mainfont to something
	#crazy.
	CloneMainFontAsFontControlPanelFont;

	if {[llength $args] > 0} {
	    set ::WhichFontToSet [lindex $args 0];
	    set General 0;
	} 
	set FontTypes [llength $FontList];
	if {$FontTypes == 1} {
	    set ::WhichFontToSet [lindex $FontList 0];
	    set General 0;
	}

	frame $w.which -relief ridge -border 3
	if {$General} {
	    set msg [_ "For what aspect of the program\ndo you wish to set the font?"]
	    balloonhelp_for $w.which $msg
	    label $w.which.title -text [_ "Use for which to set font"] \
		-anchor w -font FontControlPanelFont
	    frame $w.which.btns;
	    set k 0;
	    set PerRow 4;
	    set ypad 3
	    set xpad 10
	    foreach ft $FontList {
		set bn $w.which.btns.f$ft;
		radiobutton $bn -text [_ $FontInfo($ft,gloss)]  \
		    -variable WhichFontToSet -value $ft \
		    -command fontsel::SetFontSelectionDefaults \
		    -font FontControlPanelFont \
		    -indicatoron 0 -selectcolor $::ColorSpecs(Default,Background)
		balloonhelp_for $w.which.btns.f$ft $FontInfo($ft,help)
		set row [expr 1+ $k/$PerRow]
		set col [expr $k%$PerRow]
		if {$FontTypes >= $PerRow} {
		    grid $w.which.btns.f$ft -row $row -column $col \
			-sticky we -ipadx 3 -padx $xpad -pady $ypad
		}
		incr k
	    }
	    if {$FontTypes < $PerRow} {
		foreach ft $FontList {
		    pack $w.which.btns.f$ft -side left -expand 1 -fill x \
			-ipadx 5 -ipady 3 -padx 40 -pady 3
		}
	    }
	    frame $w.which.pad  -height 4
	    pack $w.which.title -side top -expand 1 -fill both -anchor w
	    pack $w.which.btns  -side top -expand 1 -fill both -anchor w
	    pack $w.which.pad   -side top -expand 0 -fill x    -anchor w

	}

	frame $w.mid 
	frame $w.mid.opts  -relief ridge -border 3
	frame $w.mid.families  -relief ridge -border 3
	label $w.mid.families.lab -text [_ "Family"] -relief ridge \
	    -justify center -font FontControlPanelFont
	listbox $w.mid.families.lb -height 1 -exportselection 0 \
	    -yscrollcommand "$w.mid.families.sbar set" -selectmode single \
	    -font FontControlPanelFont
	scrollbar $w.mid.families.sbar -orient vertical -command "$w.mid.families.lb yview"
	bind $w.mid.families.sbar <<B3>> \
	    "ScrollbarMoveBigIncrement $w.mid.families.sbar 0.20 %x %y"

	bind $w.mid.families.lb <<B3>> ProvideFontDescription

	pack $w.mid.families.lab  -side top -expand 0 -fill x
	pack $w.mid.families.lb   -side left -expand 1 -fill both
	pack $w.mid.families.sbar -side left -expand 1 -fill y

	label $w.mid.opts.lab -text [_ "Attributes"] -relief ridge \
	    -justify center -font FontControlPanelFont
	label $w.mid.opts.sizel -text [_ "Size:"]  -font FontControlPanelFont
	scale $w.mid.opts.size -orient h -digit 1 -from 5 -to 55 \
	    -variable FontInfo(size)  -tickinterval 0 -length 150 \
	    -font FontControlPanelFont
	bind $w.mid.opts.size <<B3>> "ScaleMoveBigIncrement $w.mid.opts.size 5 %x %y"

	label $w.mid.opts.weightl -text [::msgcat::mc "Bold:"] \
	    -font FontControlPanelFont	    
	checkbutton $w.mid.opts.weight -variable FontInfo(weight) \
	    -onvalue bold -offvalue normal  -font FontControlPanelFont

	label $w.mid.opts.slantl -text [::msgcat::mc "Italic:"] \
	    -font FontControlPanelFont
	checkbutton $w.mid.opts.slant -variable FontInfo(slant) \
	    -onvalue italic -offvalue roman -font FontControlPanelFont

	label $w.mid.opts.ulinel -text [::msgcat::mc "Underline:"] \
	    -font FontControlPanelFont
	checkbutton $w.mid.opts.uline -variable FontInfo(underline) \
	    -onvalue 1 -offvalue 0  -font FontControlPanelFont

	label $w.mid.opts.strkovl -text [::msgcat::mc "Overstrike:"] \
	    -font FontControlPanelFont
	checkbutton $w.mid.opts.strkov -variable FontInfo(overstrike) \
	    -onvalue 1 -offvalue 0  -font FontControlPanelFont

	grid $w.mid.opts.lab   -row 0 -column 0 -columnspan 2 -sticky ew
	grid $w.mid.opts.sizel -row 1 -column 0 -sticky w
	grid $w.mid.opts.size -row 1 -column 1 -sticky w
	grid $w.mid.opts.weightl -row 2 -column 0 -sticky w
	grid $w.mid.opts.weight -row 2 -column 1 -sticky w
	grid $w.mid.opts.slantl -row 3 -column 0 -sticky w
	grid $w.mid.opts.slant -row 3 -column 1 -sticky w
	grid $w.mid.opts.ulinel -row 4 -column 0 -sticky w
	grid $w.mid.opts.uline -row 4 -column 1 -sticky w
	grid $w.mid.opts.strkovl -row 5 -column 0 -sticky w
	grid $w.mid.opts.strkov -row 5 -column 1 -sticky w

	pack $w.mid.families -side left -expand 1 -fill both -padx 5 -pady 2
	pack $w.mid.opts     -side left -expand 1 -fill both -padx 5 -pady 2

	eval $w.mid.families.lb insert 0 [lsort [font families]]

	bind $w.mid.families.lb <ButtonPress-1> {fontsel::SelectFontFamily %W %y}
	bind $w.mid.families.lb <B1-Motion> {fontsel::SelectFontFamily %W %y}
	bind $w.mid.families.lb <Activate> {fontsel::SelectFontFamily %W %y}

	frame $w.cntls -relief ridge -border 3
	button $w.cntls.sav -text [_ "Apply"]  -command fontsel::FontSave \
	    -font FontControlPanelFont
	button $w.cntls.res -text [_ "Reset"] -command  fontsel::ResetToDefaults \
	    -font FontControlPanelFont
	button $w.cntls.can -text [_ "Dismiss"] -command "wm withdraw $w" \
	    -font FontControlPanelFont
	button $w.cntls.cpanel -text [_ "Here Too?"] \
	    -command fontsel::CloneMainFontAsFontControlPanelFont \
	-font FontControlPanelFont
	pack $w.cntls.cpanel -side left  -expand 1 -fill both -padx 3
	pack $w.cntls.can    -side left  -expand 1 -fill both -padx 3
	pack $w.cntls.res    -side left  -expand 1 -fill both -padx 3
	pack $w.cntls.sav    -side right -expand 1 -fill both -padx 3

	set msg [_ "Press this button to dismiss the popup."];
	balloonhelp_for $w.cntls.can $msg
	set msg [_ "Press this button to reset the font to its defaults."];
	balloonhelp_for $w.cntls.res $msg
	set msg [_ "Press this button to make the changes\nyou have made take effect."]
	balloonhelp_for $w.cntls.sav $msg
	set msg [format \
	     [_ "Press this button to set the fonts\nin this control panel to the %s font."] \
		     [_ $::FontInfo(MainFont,gloss)]]
	balloonhelp_for $w.cntls.cpanel $msg

	frame $w.curcan -relief ridge -border 1
	set ATT [frame $w.curcan.att]
	set Row 0
	foreach r {Header Current Candidate} {
	    set Column 2
	    foreach p $FontPropertyList {
		set v $w.curcan.att.r${Row}c${Column};
		label $v  -font FontControlPanelFont
		incr Column;
	    } 
	    incr Row;
	}

	font create CurrentFontExampleFont \
	    -family $FontInfo($::WhichFontToSet,family) \
	    -size $FontInfo($::WhichFontToSet,size) \
	    -weight $FontInfo($::WhichFontToSet,weight) \
	    -slant $FontInfo($::WhichFontToSet,slant) \
	    -underline $FontInfo($::WhichFontToSet,underline) \
	    -overstrike $FontInfo($::WhichFontToSet,overstrike);

	font create CandidateFontExampleFont \
	    -family $FontInfo($::WhichFontToSet,family) \
	    -size $FontInfo($::WhichFontToSet,size) \
	    -weight $FontInfo($::WhichFontToSet,weight) \
	    -slant $FontInfo($::WhichFontToSet,slant) \
	    -underline $FontInfo($::WhichFontToSet,underline) \
	    -overstrike $FontInfo($::WhichFontToSet,overstrike);

	label $w.curcan.att.r0c0 -font FontControlPanelFont
	label $w.curcan.att.r0c1 -font FontControlPanelFont
	label $w.curcan.att.r1c0  -text [_ "Current"] -font FontControlPanelFont
	entry $w.curcan.att.r1c1 -font CurrentFontExampleFont -relief flat -width $exwid
	$w.curcan.att.r1c1 insert 0 [_ "example"]
	label $w.curcan.att.r2c0  -text [_ "Candidate"] -font FontControlPanelFont
	entry $w.curcan.att.r2c1 -font CandidateFontExampleFont -relief flat  -width $exwid
	$w.curcan.att.r2c1 insert 0 [_ "example"]

	#This is for use in conjunction with my character insertion library.
	#It arranges for the various character insertion widgets to insert
	#characters into the focussed window.
	if {[llength [info commands BindInsertionTarget]]} {
	    BindInsertionTarget $w.curcan.att.r1c1;
	    BindInsertionTarget $w.curcan.att.r2c1;
	}

	set CLR [frame $w.curcan.clr]
	set ln 60;
	set wd 10
	set trc \#D4B8C1
	scale $CLR.red -orient v -digit 1 -from 255 -to 0 -tickinterval 0 \
	    -variable FontRed  -showvalue 0 -length $ln -width $wd\
	    -troughcolor $trc -activebackground \#F0AAAA \
	    -bg red -fg yellow -command fontsel::UpdateFontExampleColor
	scale $CLR.grn -orient v -digit 1 -from 255 -to 0 -tickinterval 0 \
	    -variable FontGreen  -showvalue 0 -length $ln -width $wd\
	    -troughcolor $trc -activebackground \#AAF0AA \
	    -bg green -fg yellow -command fontsel::UpdateFontExampleColor
	scale $CLR.blu -orient v -digit 1 -from 255 -to 0 -tickinterval 0 \
	    -variable FontBlue  -showvalue 0 -length $ln -width $wd\
	    -troughcolor $trc -activebackground \#AAAAF0 \
	    -bg blue -fg yellow -command fontsel::UpdateFontExampleColor
	set FSCColors(red,bg)   0x00
	set FSCColors(green,bg) 0x00
	set FSCColors(blue,bg)  0x00
	set FSCColors(red,fg)   0x00
	set FSCColors(green,fg) 0x00
	set FSCColors(blue,fg)  0x00

	bind $CLR.red <<B3>> "ScaleMoveBigIncrement $CLR.red -20 %x %y"
	bind $CLR.grn <<B3>> "ScaleMoveBigIncrement $CLR.grn -20 %x %y"
	bind $CLR.blu <<B3>> "ScaleMoveBigIncrement $CLR.blu -20 %x %y"

	set ::FontRed   255
	set ::FontGreen 255
	set ::FontBlue  255
	set FontExampleColor \#FFFFFF
	label $CLR.rdt -textvariable FontExampleColor -width 7 -relief raised
	bind $CLR.rdt <Button-1> "pack forget $CLR"
	bind $ATT <Button-1> "pack $w.curcan.clr -before $ATT -side left -expand 1 -fill both"
	set msg [_ "Left click here to remove the color selector."]
	balloonhelp_for $CLR.rdt $msg;
	set msg [_ "Use this to set the colors in the example text boxes."]
	balloonhelp_for $CLR.red $msg;
	balloonhelp_for $CLR.grn $msg;
	balloonhelp_for $CLR.blu $msg;

	frame $CLR.whi
	set ::FontExampleColorConfigureWhich bg;
	radiobutton $CLR.whi.t -variable FontExampleColorConfigureWhich \
	    -value fg -command fontsel::RestoreCSColors -text [_ "fg"] \
	    -indicatoron 0 -activebackground gray -selectcolor gray \
	    -font FontControlPanelFont
	radiobutton $CLR.whi.b -variable FontExampleColorConfigureWhich \
	    -value bg -command fontsel::RestoreCSColors -text [_ "bg"] \
	    -indicatoron 0 -activebackground gray -selectcolor gray \
	    -font FontControlPanelFont
	pack $CLR.whi.t -side left -expand 1 -fill both
	pack $CLR.whi.b -side left -expand 1 -fill both

	pack $CLR.rdt -side top -expand 1  -fill both
	pack $CLR.whi -side bottom -expand 1  -fill both
	pack $CLR.red -side left -expand 1 -fill y
	pack $CLR.grn -side left -expand 1 -fill y
	pack $CLR.blu -side left -expand 1 -fill y

	set msg [_ "Choose the color for the sample text."]
	balloonhelp_for $CLR.whi.t $msg;
	set msg [_ "Choose the background color for the sample text."]
	balloonhelp_for $CLR.whi.b $msg;

	#Now grid them all
	set xp 5
	set yp 3
	set CLimit [expr [llength $FontPropertyList] + 2];
	for {set r 0} {$r < 3} {incr r} {
	    for {set c 0} {$c < $CLimit} {incr c} {
		set v $w.curcan.att.r${r}c${c};
		grid $v -row $r -column $c -sticky w -padx $xp -pady $yp
	    } 
	}

	set xp 3
	set yp 1
	if {!$General} {
	    $w.curcan.att.r0c0 configure -text [string totitle [_ $FontInfo($::WhichFontToSet,gloss)]]
	}

	pack $w.curcan.clr -side left -expand 1 -fill both -padx 9 -pady 5
	pack $w.curcan.att -side left -expand 1 -fill both -padx 3 -pady 5

	#Put it all together
	frame $w.botsep -height 4
	pack $w.which -side top -expand 1 -fill x -padx 8 -pady 6
	pack $w.curcan -side top -expand 1 -fill x -padx 8 -pady 6
	pack $w.mid   -side top -expand 1 -fill x -padx 8 -pady 1
	pack $w.cntls -side top -expand 1 -fill x -padx 8 -pady 2
	pack $w.botsep -side top -expand 1 -fill x -padx 8 -pady 2
	if {$General} {
	    $w.which.btns.f[lindex $FontList 0] invoke;#  Default to first on list
	}
	foreach f $FontPropertyList {
	    trace add variable FontInfo($f) write fontsel::FontShow;
	}
	return $w;
    }

    #This resets the properties of the font specified by WhichFontToSet
    #to their default values, where "default" means whatever values
    #are stored in FontInfo(<font>,<property>,Default). The intention
    #is that these values will have been set by RecordDefaults at
    #a suitable point. A suitable point might be program startup or it
    #might be after init files have been read.
    #After resetting the attributes for the current font to their
    #default values, it configures the fonts for both the current and candidate
    #examples in the control panel to match.
    proc ResetToDefaults {} {
	global FontInfo;
	global WhichFontToSet
	variable FontPropertyList

	foreach prop $FontPropertyList {
	    set FontInfo($::WhichFontToSet,$prop) $FontInfo($::WhichFontToSet,$prop,Default);
	    set FontInfo($prop) $FontInfo($::WhichFontToSet,$prop,Default);
	}
	ConfigureFont $::WhichFontToSet;
	ConfigureCurrentExampleFont;
	ConfigureCandidateExampleFont;
	DescribeBothFonts;
    }

    #This is the procedure called from the control panel when the user decides
    #to apply and save the values set on the control panel. It sets the font-specific
    #attributes from the generic ones, configures the target font as well as the
    #current example font, and updates the description of the current font
    #in the control panel.
    proc FontSave {} {
	global FontInfo

	set f $::WhichFontToSet;
	FontSet $f;
	ConfigureFont $f
	ConfigureCurrentExampleFont;
	DescribeBothFonts;
    } 

    #This procedure configures the font used to display the candidate example
    #in the control panel. It is intended to be called whenever a property
    #is changed in the control panel or when the target font is changed.
    proc ConfigureCandidateExampleFont {} {
	global FontInfo
	global WhichFontToSet
	variable FontPropertyList

	foreach prop $FontPropertyList {
	    font configure CandidateFontExampleFont -$prop $FontInfo($prop)
	}
    }

    #This procedure configures the font used to display the current example
    #in the control panel. It is intended to be called when the user
    #applies new values or when the target font is changed.
    proc ConfigureCurrentExampleFont {} {
	global FontInfo
	global WhichFontToSet
	variable FontPropertyList

	foreach prop $FontPropertyList {
	    font configure CurrentFontExampleFont -$prop $FontInfo($::WhichFontToSet,$prop)
	}
    }

    #This procedure configures the candidate example font#and updates the
    #description of the candidate font in the control panel. It is intended
    #to be called whenever a proprty of the candidate font is changed.
    proc FontShow {args} {
	ConfigureCandidateExampleFont;
	DescribeBothFonts;
    }

    #This procedure generates the descriptions of the current and candidate
    #fonts that are shown in the control panel.
    proc DescribeBothFonts {} {
	global FontInfo
	global WhichFontToSet;
	variable FontPropertyList

	set f $::WhichFontToSet;
	set c 2;
	foreach prop $FontPropertyList {
	    .selectFont.curcan.att.r0c$c configure -text  [string totitle [_ $prop]] 
	    set value $FontInfo($f,$prop)
	    if {$value == 1} {set value [_ "Yes"]} else {
		if {$value == 0} {set value [_ "No"]}}
	    .selectFont.curcan.att.r1c$c configure -text $value;
	    set value $FontInfo($prop)
	    if {$value == 1} {set value [_ "Yes"]} else {
		if {$value == 0} {set value [_ "No"]}}
	    .selectFont.curcan.att.r2c$c configure -text $value; 
	    incr c;
	}
    }

#This is to be called when we change font targets.
#It resets the control panel to show the current
#configuration for the newly targeted font.
    proc SetFontSelectionDefaults {} {
	global FontInfo;
	global WhichFontToSet;
	variable FontPropertyList

	foreach prop $FontPropertyList {
	    set FontInfo($prop) $FontInfo($WhichFontToSet,$prop);
	}
	set FontChoices [.selectFont.mid.families.lb size];
	set k 1;
	while {$k <= $FontChoices} {
	    set Family [.selectFont.mid.families.lb get $k];
	    if {[string equal $Family [string tolower $FontInfo($WhichFontToSet,family)]]} {
		.selectFont.mid.families.lb selection clear 0 end;
		.selectFont.mid.families.lb see $k
		.selectFont.mid.families.lb selection set $k $k;
		.selectFont.mid.families.lb activate $k;
		set FontInfo(family) [.selectFont.mid.families.lb get $k]
		break
	    }
	    incr k;
	}
	ConfigureCandidateExampleFont
	ConfigureCurrentExampleFont;
	DescribeBothFonts;
    }

    #This is the callback that extracts a new font family
    #setting from the control panel.
    proc SelectFontFamily {w y} {
	global FontInfo

	set index [$w nearest $y]
	if {$index != ""} {
	    set FontInfo(family) [$w get $index]
	}
    }

    #This is the callback that extracts a new font size
    #from the control panel.
    proc SelectFontSize {w} {
	global WhichFontToSet
	global FontInfo

	set size [$w get]
	if {[catch {expr round($size)} size] == 0} {
	    if { ($size > 0) && ($size <= 60) } {
		set FontInfo(size) $size
		return ;
	    }
	}
	$w delete 0 end
	$w insert 0 $FontInfo($WhichFontToSet,size)
    }

    #This is presently a stub.
    #It is intended to be called on an event, probably <B3>,
    #in the font family listbox in the control panel.
    #Its purpose is to provide the user with information about
    #the particular font. 
    proc ProvideFontDescription {w y} {
	variable FontDescriptions

	set index [$w nearest $y]
	if {$index != ""} {
	    FontDescriptions([$w get $index])
	}
    }

    #Create a command for setting each combination of font and property.
    proc DefineFontSettingProcs {} {
	variable FontPropertyList
	variable FontPropertyClass
	global FontList;
	global FontProcList;

	foreach ft $FontList {
	    foreach prop $fontsel::FontPropertyList {
		if {[string equal $FontPropertyClass($prop) boolean]} {
		    set ProcName  [format "::fontsel::Set%s%sP" $ft [string totitle $prop]];
		    lappend FontProcList $ProcName;
		    set cmd [list "proc"  $ProcName  "\{v\}" \
			 "set ::FontInfo($ft,$prop) \[SlaveBoolean \$v\];\nConfigureFont $ft;"]
		} else {
		    set ProcName  [format "::fontsel::Set%s%s" $ft [string totitle $prop]];
		    lappend FontProcList $ProcName;
		    set cmd [list "proc" $ProcName "\{v\}" \
			 "set ::FontInfo($ft,$prop) \$v;\nConfigureFont $ft;"]
		}
		eval $cmd;
	    }
	}
    }

    #This is used when we are running without Tk.
    proc DefineNoopFontSettingProcs {} {
	variable FontPropertyList
	variable FontPropertyClass
	global FontList;
	global FontProcList;

	foreach ft $FontList {
	    foreach prop $fontsel::FontPropertyList {
		set ProcName  [format "::fontsel::Set%s%s" $ft [string totitle $prop]];
		lappend FontProcList $ProcName;
		set cmd [list "proc" $ProcName "\{v\}" \
			     ";"]
		eval $cmd;
	    }
	}
    }


    #Create aliases in the named daughter interpreter for the font setting commands
    proc AliasFontSettings {interp} {
	variable FontPropertyList
	variable FontPropertyClass
	global FontList;
	
	foreach ft $FontList {
	    foreach prop $FontPropertyList {
		set Tprop [string totitle $prop]
		set Tprop2 $Tprop;
		if {[string equal $FontPropertyClass($prop) boolean]} {
		    set Tprop ${Tprop}P
		}
		eval [split [format "%s alias %s%s ::fontsel::Set%s%s" \
			 $interp $ft $Tprop $ft $Tprop2]]
	    }
	}
    }

    #Returns a list of init file commands representing the current font settings.
    proc SaveFontSettings {} {
	variable FontPropertyList
	variable FontPropertyClass
	global FontInfo
	global FontList;
	
	set cl [list]
	foreach ft $FontList {
	    foreach prop $FontPropertyList {
		set Tprop [string totitle $prop]
		if {[string equal $FontPropertyClass($prop) boolean]} {
		    set Tprop ${Tprop}P
		}
		lappend cl [format "%s%s %s" $ft $Tprop [list $FontInfo($ft,$prop)]]
	    }
	}
	return [lsort $cl];
    }
}

# End of namespace fontsel

#Set your font list here.
set FontList { \
MainFont\
LHSFont\
RHSFont\
BalloonHelpFont\
MenuFont\
CharacterEntryFont\
CharacterEntryLabelFont\
DiacriticFont\
AccentedLetterFont}

#If you want to override defaults in a general way (that is, not specific to
#a particular font, do it here. 
set FontInfo(family) clearlyu

# Set the properties that you want to be different from the defaults
# At a minimum these will probably include the glosses and balloon help text.
set FontInfo(BalloonHelpFont,family) helvetica
set FontInfo(CharacterEntryFont,size) 12
set FontInfo(LHSFont,size) 12
set FontInfo(MainFont,family) Courier
set FontInfo(MenuFont,size) 12
set FontInfo(RHSFont,size) 12
set FontInfo(family) code2000
set FontInfo(BalloonHelpFont,family) lucida
set FontInfo(BalloonHelpFont,size) 13
set FontInfo(MainFont,family) courier
set FontInfo(MainFont,size) 12
set FontInfo(MainFont,weight) bold
set FontInfo(MenuFont,family) courier
set FontInfo(MenuFont,size) 12

set FontInfo(AccentedLetterFont,family) "code2000"
set FontInfo(AccentedLetterFont,size) 14
set FontInfo(CharacterEntryLabelFont,family) bible
set FontInfo(CharacterEntryLabelFont,size) 13
set FontInfo(CharacterEntryFont,size) 14
set FontInfo(CharacterEntryFont,family) "doulos sil"
set FontInfo(PanelCharacterEntryFont,size) 13
set FontInfo(PanelCharacterEntryFont,family)  "doulos sil"
set FontInfo(DiacriticFont,size)  [expr $FontInfo(CharacterEntryFont,size) + 40];

#We package this in a procedure because in other programs in which the
#same system is used the use may change the locale in the middle of a session,
#in which case we need to obtain new translations. For this program we could
#just set these here once-and-for-all.
proc SetFontGlossHelpTranslations {} {
    global FontInfo;
    set FontInfo(BalloonHelpFont,gloss) [_ "balloon help"]
    set FontInfo(BalloonHelpFont,help) [_ "This font is used in help balloons."]
    set FontInfo(DiacriticFont,gloss) [_ "diacritics"]
    set FontInfo(DiacriticFont,help) \
	[_ "This font is used on the labels of the\ndiacritic insertion widget."]
    set FontInfo(CharacterEntryLabelFont,gloss) [_ "character entry labels"]
    set FontInfo(CharacterEntryLabelFont,help) \
	[_ "This font is used for the row and column\nlabels in character entry widgets."]
    set FontInfo(MainFont,gloss) [_ "general"]
    set FontInfo(MainFont,help) [_ "This is the font used for most things."]
    set FontInfo(MenuFont,gloss) [_ "menu"]
    set FontInfo(MenuFont,help) [_ "This font is used for menu labels."]
    set FontInfo(CharacterEntryFont,gloss) [_ "character entry"]
    set FontInfo(CharacterEntryFont,help) \
	[_ "This font is used on the labels of\ncharacter insertion widgets."]
    set FontInfo(PanelCharacterEntryFont,gloss) [_ "character entry panel"]
    set FontInfo(PanelCharacterEntryFont,help) \
	[_ "This font is used on the labels of\n the character insertion panel."]
    set FontInfo(AccentedLetterFont,gloss) [_ "accented letters"]
    set FontInfo(AccentedLetterFont,help) \
	[_ "This font is used on the labels of the\naccented character insertion widget."]
    set FontInfo(LHSFont,gloss) "left hand side";
    set FontInfo(LHSFont,help) "This font is used on the left hand side\nof the transliteration definition window\nand for the text to be translitered."
    set FontInfo(RHSFont,gloss) "right hand side";
    set FontInfo(RHSFont,help) "This font is used on the right hand side\nof the transliteration definition window\nand for the results of transliteration."

}

SetFontGlossHelpTranslations;

#End of font stuff

proc Abort {} {
    if {$::TransliterationInProgressP} {
	set ::AbortP 1;
    } else {
	ShowMessage [_ "There is no transliteration in progress so there is nothing to abort."]
    }
}

proc GetHomeDir {} {
    set cwd [pwd];
    cd;
    set hd [pwd];
    cd $cwd;
    return $hd;
}

proc GetUser {} {
    if {[info exists ::env(USER)]} {
	return $::env(USER);
    }
}

proc LockInputText {} {
    $::TXTINP configure -state disabled
    ShowMessage [_ "The text to be transliterated is now locked."]
}

proc UnlockInputText {} {
    $::TXTINP configure -state normal
    ShowMessage [_ "The text to be transliterated is now editable."]
}


set ParseDialogIsDisplayedP 0;
proc ToggleParseDialog {} {
    global ParseDialogIsDisplayedP;
    global m;

    if {$ParseDialogIsDisplayedP == 0} {
	if {[winfo exists .par]} {
	    wm deiconify .par;
	    raise .par;
	} else { 
	    PopupParseDialog;
	}
	set ParseDialogIsDisplayedP 1;
	$m.misc entryconfigure $::ToggleParseDialogIndex \
	    -label [_ "Remove Parse Dialog"];
    } else {
	if {[winfo exists .par]} {
	    wm iconify .par;
	}
	set ParseDialogIsDisplayedP 0;
	$m.misc entryconfigure $::ToggleParseDialogIndex \
	    -label [_ "Display Parse Dialog"];
    }
}

proc ExplainHZEscapes {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(HZEscapes)]} {
	if {[winfo exists $PopupList(HZEscapes)]} {
	    destroy $PopupList(HZEscapes);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "HZ Escapes"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(HZEscapes) $po;
    AppendToTextDisplay $po [_ "HZ escapes are a device for separating materials in different encodings developed originally for use on the internet. They are also used in some other computational contexts. Unlike most delimitation systems, which were not intended for mechanical processing of the text, HZ escapes provide a means of separating matrix and embedded text unambiguously, without interfering with the use of the delimiters in other contexts in the language.  The sequence ~\{ initiates an embedded region; the sequence ~\} terminates an embedded region. A literal tilde is represented by a sequence of two tildes. A singleton tilde, that is, a tilde followed by any character other than itself, \}, \{, or newline, is an error. The sequence ~\\ at the end of a line serves as a
line-continuation marker and is consumed without output."];
    AppendToTextDisplay $po [_ "This system allows the delimiters \{ and \} to be included in the text, since they only have a special meaning when immediately preceded by a tilde. The sequences tilde \{ and tilde \} are also representable as ~~\{ and ~~\} respectively."];
    AppendToTextDisplay $po [_ "HZ escapes were originally proposed as a means of separating English and Chinese text in bilingual documents in "];
    AppendLinkToTextDisplay $po [_ "RFC 1843."] {ShowWebPage http://www.ietf.org/rfc/rfc1843.txt};
}

proc ExplainXML {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(XML)]} {
	if {[winfo exists $PopupList(XML)]} {
	    destroy $PopupList(XML);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "XMLDelimiters"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(XML) $po;
    AppendToTextDisplay $po [_ "XML delimiters consisting of an opening tag and a distinct, but \
systematically related, closing tag. The opening tag has the form \'<x>\' where x is the name \
of the tag. The closing tag has the form \'</x>\'. For example, the HTML paragraph tags \
\'<p>\' and \'</p>\' are valid XML tags, as are the HTML emphasis tags \'<em>\' and \'</em>\'."];
}

proc ExplainSeparatorCharacter {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(SeparatorCharacter)]} {
	if {[winfo exists $PopupList(SeparatorCharacter)]} {
	    destroy $PopupList(SeparatorCharacter);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "Domain Seprators"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(SeparatorCharacter) $po;
    AppendToTextDisplay $po [_ "Separator characters separate one domain from the next, serving simultaneously to end one portion and begin the next."];
}

proc ExplainPairedDelimiters {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(PairedDelimiters)]} {
	if {[winfo exists $PopupList(PairedDelimiters)]} {
	    destroy $PopupList(PairedDelimiters);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "Paired Delimiters"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(PairedDelimiters) $po;
    AppendToTextDisplay $po [_ "Paired delimiters are those that come in pairs, an opening delimiter and a closing delimiter, such as \'(\' and \')\'."];
}


proc ExplainNeedForParsing {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(NeedForParsing)]} {
	if {[winfo exists $PopupList(NeedForParsing)]} {
	    destroy $PopupList(NeedForParsing);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "The Need for Parsing"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(NeedForParsing) $po;
    AppendToTextDisplay $po [_ "It is often the case that only part of a text should be transliterated. In many cases, a text is written in one language, the matrix language, but contains quotations or other material in a second language. These latter may be considered to be embedded in the matrix language text. Material in the two languages may be distinguished and treated differently if the material in the embedded language is enclosed within delimiters. In some cases these delimiters are inserted for the express purpose of facilitating mechanical processing. This is the case with XML tags and HZ escapes. In other cases, the delimiters are inserted by the original author of the text for other reasons. Examples of this type are the use by linguists of square brackets \[ \] for phonetically transcribed material, slashes / / for phonologically transcribed material, and angle brackets (for which less-than and greater-than signs are frequently substituted), for orthographic representations. The use of quotation marks of various sorts is another example of the same type."];
    AppendToTextDisplay $po [_ "In order to take advantage of the existence of such delimiters, it is necessary for the transliteration program to parse the input text into a sequence of matrix and embedded regions, to which transliteration can be applied selectively."];
}

proc ExplainOutputDelimiters {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(OutputDelimiters)]} {
	if {[winfo exists $PopupList(OutputDelimiters)]} {
	    destroy $PopupList(OutputDelimiters);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "Output Delimiters?"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(OutputDelimiters) $po;
    AppendToTextDisplay $po [_ "When delimiters are used to separate some portions of the text from others, they may be of no further use and considered clutter. In this case, it is desirable to remove them once the transliteration is done. In other cases, they may be needed for further processing or they may be the punctuation intended to be used in the ultimate document. You therefore have the choice of removing them, keeping them, or replacing them with other delimiters. This means that you can also use this program to convert one type of delimiter to another, even if you don't want to transliterate anything."]
}

proc ExplainStrictParse {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(StrictParse)]} {
	if {[winfo exists $PopupList(StrictParse)]} {
	    destroy $PopupList(StrictParse);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "Strict Parse?"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(StrictParse) $po;
    AppendToTextDisplay $po [_ "For a particular approach to delimiting embedded regions, some sequences of characters in the input text are ill-formed. Such sequences lead to parse errors. For example, if the left delimiter is < and the right delimiter is >, it is an error for a second < to occur before a > closes the domain opened by the first <. A sequence like <xxx>yyy<zzz> is well-formed, while a sequence like <<xxx>yyy<zzz> is ill-formed. When a parser encounters such ill-formed input, it has two options. One is to issue an error message and abort. The other is to try to recover from the error and continue with the parse. If the parser is able to do a good job of recovery, it is often most useful to let it do so since it will then generate a list of all of the errors in the input, which the user can then correct. Aborting on the first error may be necessary in some cases, but it has the disadvantage that the user only learns about the location of the first error. You may choose which approach to take."]
}


proc ShowWebPage {url} {
    global BrowserPIDS;

    if { $::tcl_platform(platform) == "macintosh"} {
	lappend BrowserPIDS [exec osascript -e "\"open location $url\""]
	return 
    }
    set BrowserFound 0;
    foreach Browser $::BrowserList {
	if { [string length [auto_execok $Browser]]} {
	    set BrowserFound 1;
	    ShowMessage [format [_ "Browser %s located."] $Browser];
	    break ;
	} else {
	    ShowMessage [format \
	     [_ "The browser %s is not available on this machine or not in your path."]\
		 $Browser];
	}
    }
    if {$BrowserFound} {
	lappend BrowserPIDS [exec $Browser $url &]
    } else {
	ShowMessage [_ "No browser on the browser list was located."]
    }
}

proc ConfirmationDialog {WindowName TitleText MainText LeftText RightText CancelHelp} {
    global wret;
    global MainFont;
    
    toplevel $WindowName -borderwidth 4 -relief raised -bg $::ColorSpecs(ConfirmationDialogueTitle,Background)
    wm title $WindowName "";
    label $WindowName.title -text $TitleText -font MainFont -bg $::ColorSpecs(ConfirmationDialogueTitle,Background) -fg \#00FFFF
    label $WindowName.txt -width 50 -height 1 -font MainFont\
	-relief flat -anchor c -text $MainText -bg $::ColorSpecs(ConfirmationDialogueLabel,Background) -fg \#00FFFF
    button $WindowName.right   -text $RightText\
	-bg $::ColorSpecs(ConfirmationDialogue,Background) -font MainFont -command {set ::wret 1}
    button $WindowName.left -text $LeftText\
	-bg $::ColorSpecs(ConfirmationDialogue,Background) -font MainFont -command {set ::wret 0}
    button $WindowName.cancel -text [_ "Cancel"]\
	-bg $::ColorSpecs(ConfirmationDialogue,Background) -font MainFont -command {set ::wret -1}
    grid $WindowName.title     -row 0 -column 0 -sticky ew -columnspan 3
    grid $WindowName.txt       -row 1 -column 0 -sticky ew -columnspan 3
    grid $WindowName.left      -row 2 -column 0
    grid $WindowName.right     -row 2 -column 2
    grid $WindowName.cancel    -row 2 -column 1
    balloonhelpd_for $WindowName.cancel $CancelHelp;
    tkwait variable wret;
    destroy $WindowName;
    return $wret;
}

proc WriteJournal {msg} {
    global JournalHandle;
    global SynchJournalP;
    puts $JournalHandle $msg;
    if {$SynchJournalP} {
	FlushJournal;
    }
}

proc CloseJournal {} {
    close $::JournalHandle;
}

proc FlushJournal {} {
    global JournalHandle;
    flush $JournalHandle;
}

proc ShowMessage {msg} {
    if {$::TkPresentP} {
	if {[winfo exists $::MSG]} {
	    $::MSG configure -state normal;
	    $::MSG delete 1.0 end;
	    $::MSG insert 1.0 $msg;
	    $::MSG configure -state disabled;
	    WriteJournal $msg;
	    return
	}
    } 
    puts stderr $msg;
    WriteJournal $msg;
}

proc ClearMessageWindow {} {
    if {[winfo exists $::MSG]} {
	$::MSG configure -state normal;
	$::MSG delete 1.0 end;
	$::MSG configure -state disabled;
    }
}

proc ClearInputText {} {
    if {[string length [$::TXTINP get 1.0 end]] > 1 } {
	$::TXTINP configure -state normal;
	$::TXTINP delete 1.0 end;
	$::TXTINP configure -state disabled;
	ShowMessage [_ "Input text cleared."];
    } else {
	ShowMessage [_ "There is no input text to clear."];
    }
    if {[string length [$::TXTOUT get 1.0 end]] > 1 } {
	set resp [ConfirmationDialog .crd "" "Clear Transliterated Text?" "No, Leave it Alone" "Yes, Clear It"  ""];
	if {$resp} {
	    ClearTransliteratedText;
	}
    }
}

proc ClearTransliteratedText {} {
    $::TXTOUT configure -state normal;
    $::TXTOUT delete 1.0 end;
    $::TXTOUT configure -state disabled;
    ShowMessage [_ "Transliterated text cleared."];
}

proc ClearTransliteration {} {
    if {[expr ([string length [$::XLINP get 1.0 end]] > 1) || ([string length [$::XLOUT get 1.0 end]] > 1)] } {
	StashTransliteration;
	$::XLINP configure -state normal;
	$::XLINP delete 1.0 end;
	$::XLOUT configure -state normal;
	$::XLOUT delete 1.0 end;
	ShowMessage [_ "Transliteration cleared."];
    } else {
	ShowMessage [_ "There is no transliteration to clear"]
    }
}

proc LoadInputText {args} {		# Optional filename argument

    if {[llength $args] > 0} {
	set InDataFile [lindex $args 0];
    } else {
	set InDataFile [tk_getOpenFile -title [_ "Load Input Text"]];
    }
    if {$InDataFile != ""} {
	if {[catch {open $InDataFile "r"} InDataHandle ] != 0} {
	    ShowMessage [format [_ "Unable to open input text file %s."] $InDataFile];
	    return;
	}
	set new [read $InDataHandle];
	close $InDataHandle;
    } else {
	ShowMessage [_ "No file name specified for input text."];
	return;
    }
    if {$new == ""} {
	ShowMessage [format [_ "Unable to read text to be transliterated from file %s."] $InDataFile];
	return;
    }
    if {$::BatchModeP} {
	set ::InputText $new;
    } else {
	$::TXTINP configure -state normal
	$::TXTINP  insert insert $new;
	$::TXTINP configure -state disabled
    }
    if {$::VerboseP} {
	ShowMessage [format [_ "Loaded text to be transliterated from %s."] $InDataFile]
    }
}

proc SaveOutput {} {
    
    set XFormedText [$::TXTOUT get 1.0 end];
    if {$XFormedText == ""} {
	ShowMessage [_ "There is no output to save."];
	return ;
    }
    set OutputSaveFile [tk_getSaveFile\
			    -initialfile [_ "TransliteratedText"]\
			    -title [_ "Save Transliterated Text"]
		       ];
    if {$OutputSaveFile == ""} {
	ShowMessage [format [_ "File selection cancelled."]];
	return ;
    }
    if {[catch {open $OutputSaveFile "w"} OutputSaveHandle ] != 0} {
	ShowMessage [format [_ "Unable to open file %s to save transliterated text."] $OutputSaveFile];
	return ;
    }
    puts $OutputSaveHandle $XFormedText;
    close $OutputSaveHandle;
    set ::ResultSaveTime [clock seconds]
    ShowMessage [format [_ "Transliterated text saved in file %s."] $OutputSaveFile];
}

proc SaveInputText {} {
    set Txt [$::TXTINP get 1.0 end];
    if {$Txt == ""} {
	ShowMessage [_ "There is no input to save."];
	return ;
    }
    set InputSaveFile [tk_getSaveFile\
			    -initialfile [_ "ModifiedInputText"]\
			    -title [_ "Save Input Text"]
		      ];
    if {$InputSaveFile == ""} {
	ShowMessage [format [_ "File selection cancelled."]];
	return ;
    }
    if {[catch {open $InputSaveFile "w"} InputSaveHandle ] != 0} {
	ShowMessage [format [_ "Unable to open file %s to save input text."] $InputSaveFile];
	return ;
    }
    puts $InputSaveHandle $Txt;
    close $InputSaveHandle;
    ShowMessage [format [_ "Input text saved in file %s."] $InputSaveFile];
}

proc PopupAlert {name txt} {
    global ColorSpecs;

    destroy .alert_$name;
    toplevel .alert_$name -borderwidth 4 -relief raised -bg $ColorSpecs(Alert,Background)
    wm overrideredirect .alert_$name 1;
    label .alert_$name.title -text $txt -bg  $::ColorSpecs(Alert,Background) -fg $::ColorSpecs(Alert,Foreground);
    button .alert_$name.dismiss -text [_ "Dismiss"] -bg $::ColorSpecs(AlertDismiss,Background) -fg $::ColorSpecs(AlertDismiss,Foreground) -command "destroy .alert_$name";
    pack .alert_$name.title -side top -fill x -padx 8 -pady 8;
    pack .alert_$name.dismiss -side right;
    catch {.alert_$name.title configure -font -*-Courier-bold-o-normal--*-180-*}
}

set SavedXLSrc "";
set SavedXLTgt "";

proc StashTransliteration {} {
    if {!$::BatchModeP} {
	if {[winfo exists $::XLINP]} {
	    set ::SavedXLSrc [string trimright [$::XLINP get 1.0 end]]
	    set ::SavedXLTgt [string trimright [$::XLOUT get 1.0 end]]
	}
    }
}

proc GetTransliterationFromXL {} {
    set r [list]
    foreach k [array names ::XL] {
	lappend r $k
	lappend r $::XL($k);
    }
    return $r;
}

proc GetTransliterationFromWindows {} {
    StashTransliteration;
    set SrcLines [split $::SavedXLSrc "\n"];
    set TgtLines [split $::SavedXLTgt "\n"];
    set SrcCnt [llength $SrcLines];
    set TgtCnt [llength $TgtLines];
    set Limit $SrcCnt;
    if {$SrcCnt != $TgtCnt} {
	PopupAlert Mismatch [format [_ "Warning: transliteration inputs and outputs do not match.\nLeft Hand Side  = %d Right Hand Side = %d"] $SrcCnt $TgtCnt];
	#Ignore the excess lines.
	if {$SrcCnt > $TgtCnt} {
	    set Limit $TgtCnt;
	}
    }
    set map [list];
    set ::XLRulesInWindows 0;
    foreach s $SrcLines t $TgtLines {
	lappend map $s
	lappend map $t
	set ::XL($s) $t;
	incr ::XLRulesInWindows;
	if {$::XLRulesInWindows == $Limit} { break}
    }
    return $map;
}

proc RestoreSavedTransliteration {} {
    $::XLINP delete 1.0 end;
    $::XLOUT delete 1.0 end;
    set SrcLines [split $::SavedXLSrc "\n"];
    set TgtLines [split $::SavedXLTgt "\n"];
    foreach s $SrcLines t $TgtLines {
	$::XLINP insert end $s;
	$::XLINP insert end "\n";
	$::XLOUT insert end $t;
	$::XLOUT insert end "\n";
    }
}

proc SaveTransliterationNativeFormat {} {
    SaveTransliteration 1;
}

proc SaveTransliteration {NativeFormatP} {
    global TransliterationDelimiter;
    
    set map [GetTransliterationFromWindows];
    if {$::XLRulesInWindows == 0} {
	ShowMessage [_ "There is nothing to save."];
	return;
    }
    set TransliterationSaveFile [tk_getSaveFile\
        -initialfile [_ "SavedTransliteration"]\
	-title [_ "Save Transliteration Definition"]];
    if {[ string equal $TransliterationSaveFile ""] ==1} {
	ShowMessage [format [_ "File selection cancelled."]];
	return;
    } else {
	if { [catch {open $TransliterationSaveFile "w+"} TransliterationSaveHandle] != 0} {
	    ShowMessage [format [_ "Unable to open file %s to save transliteration."] \
			     $TransliterationSaveFile];
	    return ;
	}
    }
    if {$NativeFormatP == 0} {
	puts $TransliterationSaveHandle  "//Yudit .kmap."
	puts $TransliterationSaveHandle \
	    [format [_ "//Generated %s for %s by xlit."] [clock format [clock seconds]] [GetUser]]
    } 
    set EltCnt 1;
#The LHS in a Yudit keymap must be ASCII, so we check here.
    if {$NativeFormatP == 0} {
	set NonAsciiCharFoundP 0;
	foreach elt $map {
	    if {$EltCnt % 2 == 1} {
		set i 0;
		while {$i < [string length $elt]} {
		    if {[scan "%c"  [string index $elt $i]] > 0x7F} {
			set NonAsciiCharFoundP 1;
			break ;
		    }
		    incr i;
		}
		if {$NonAsciiCharFoundP} {
		    ShowMessage [format [_ "Cannot save as Yudit kmap: LHS contains non-ASCII characters: %s."] $elt]
		    return ;
		}
	    }
	    incr EltCnt;
	}
    }
    set EltCnt 1;
    set NotFirstP 0
    foreach elt $map {
	if {$EltCnt % 2 == 0} {
	    if {$NativeFormatP} {
		puts $TransliterationSaveHandle \
		    [format "%s%s%s" $in $TransliterationDelimiter $elt];
	    } else {
		if {$NotFirstP} {
		    puts $TransliterationSaveHandle ","
		}
		set NotFirstP 1
		puts -nonewline $TransliterationSaveHandle [format "\"%s=" $in]
		set i 0;
		while {$i < [string length $elt]} {
		    scan [string index $elt $i] "%c" c;
		    puts -nonewline $TransliterationSaveHandle [format " 0x%4.4x" $c];
		    incr i;
		}
		puts -nonewline $TransliterationSaveHandle "\"";
	    }
	} else {
	    set in $elt;
	}
	incr EltCnt;
    }
    if {$NativeFormatP == 0} {puts $TransliterationSaveHandle ""};
    close $TransliterationSaveHandle;
    set ::XLSaveTime [clock seconds]
    ShowMessage [format [_ "Transliteration saved in file %s."] $TransliterationSaveFile];
}

# Returns 1 on successfully parsing a line
#         0 on reading a blank/comment line
#        -1 on error 
proc ParseNativeTransliterationEntry {entry} {
    global TransliterationDelimiter;

    set Fields [split $entry $TransliterationDelimiter];
    set FieldCnt [llength $Fields];
    if {$FieldCnt == 0} {return 0}
    if {$FieldCnt <  2} {return -1}
    set re [lindex $Fields 0];
    set se [lindex $Fields 1];
    if {$::BatchModeP} {
	set ::NewXLIn  $re;
	set ::NewXLOut $se;
    } else {
	$::XLINP insert end $re;
	$::XLINP insert end "\n";
	$::XLOUT insert end $se;
	$::XLOUT insert end "\n";
    }
    return 1;
}

# Returns 1 on successfully parsing a line
#         0 on reading a blank/comment line
#        -1 on error 
# Same as ParseNativeTransliterationEntry but for Yudit .kmap format
proc ParseYuditKmapEntry {entry} {
    set lhs "";
    set rhs "";
    set InStr "";
    set OutStr "";
    set cloc [string first "//" $entry];
    if {$cloc >= 0} {
	set entry [string range $entry 0 [expr $cloc -1]]
    }
    if {[string equal $entry ""]} {return 0}
    set sides [split [string trim [string trimright $entry ","] "\""]  "="];
    set lhs [lindex $sides 0];
    dmsg "LHS = $lhs";
    set slhs [split $lhs]
    foreach k $slhs {
	if {[scan $k "%x" num] ==1} {
	    append InStr [format %c $num];
	} else {
	    append InStr $k;
	}
    }
    set rhs [lindex $sides 1];
    dmsg "RHS = $rhs";
    set srhs [split $rhs]
    foreach k $srhs {
	if {[scan $k "%x" num] ==1} {
	    append OutStr [format %c $num];
	}
    }
    if {$::BatchModeP} {
	set ::NewXLIn  $re;
	set ::NewXLOut $se;
    } else {
	$::XLINP insert end $InStr;
	$::XLINP insert end "\n";
	$::XLOUT insert end $OutStr;
	$::XLOUT insert end "\n";
    }
    return 1;
}

#Takes as optional argument the name of the function that handles
#the particular file type.
proc LoadTransliterationFile {{ f ParseNativeTransliterationEntry} {fn ""} } {
    set cs "CommentCharacter";
    set CommentCharacter $::DefaultCommentCharacter;
    if {$fn == ""} {
	set XLFile [tk_getOpenFile -title [_ "Load Transliteration Definition"]];
    } else {
	set XLFile $fn;
    }
    if {$XLFile != ""} {
	if { [catch {open $XLFile "r"} XLHandle ] != 0} {
	    ShowMessage [format [_ "Unable to open transliteration file %s."] $XLFile];
	    return ;
	}
	set lineno 0;
	set ::XLRulesLoaded 0;
	while {![eof $XLHandle]} {
	    gets $XLHandle line;
	    incr lineno;
	    if {[string index $line 0] == $CommentCharacter} {
		if {$lineno == 1} {
		    set line [string trimright [string range $line 1 end]]
		    if {[string first $cs $line] >= 0} {
			set CommentCharacter [string range $line [expr [string length $cs] + 1] end]
			puts "Comment Character redefined as $CommentCharacter";
		    }
		}
		continue
	    }
	    set rval [$f $line]; #Here's where the real work is done.
	    if {$::BatchModeP} {
		set ::XL($::NewXLIn) $::NewXLOut;
	    }
	    if {$rval > 0} {
		incr ::XLRulesLoaded;
	    } elseif {$rval < 0} {
		ShowMessage [format [_ "Ill-formed entry at line %d."] $lineno]
		return ;
	    }
	}
	close $XLHandle;
	StashTransliteration;
    } else {
	ShowMessage [_ "File selection cancelled."];
	return ;
    }
    ShowMessage [format [_ "Loaded %d rules from %s."] $::XLRulesLoaded $XLFile]
    set ::XLRulesInWindows $::XLRulesLoaded;
    if {!$::BatchModeP} {UpdateEntryLineNumbers}
    return ;
}

proc LoadTransliterationLHS {{fn ""}} {
    if {$fn == ""} {
    set XLFile [tk_getOpenFile -title [_ "Load Transliteration Definition LHS"]];
    } else {
	set XLFile $fn;
    }
    if {$XLFile != ""} {
	if { [catch {open $XLFile "r"} XLHandle ] != 0} {
	    ShowMessage [format [_ "Unable to open file %s."] $XLFile];
	    return ;
	}
	while { [gets $XLHandle line] > 0} {
	    $::XLINP insert insert $line;
	    $::XLINP insert insert "\n";
	}
	close $XLHandle;

    } else {
	ShowMessage [_ "File selection cancelled."];
	return ;
}
    ShowMessage [format [_ "Loaded transliteration lefthand side from %s."] $XLFile]
    return ;
}

proc LoadTransliterationRHS {} {
    set XLFile [tk_getOpenFile -title [_ "Load Transliteration Definition RHS"]];
    if {$XLFile != ""} {
	if { [catch {open $XLFile "r"} XLHandle ] != 0} {
	    ShowMessage [format [_ "Unable to open file %s."] $XLFile];
	    return ;
	}
	while { [gets $XLHandle line] > 0} {
	    $::XLOUT insert insert $line;
	    $::XLOUT insert insert "\n";
	}
	close $XLHandle;

    } else {
	ShowMessage [_ "File selection cancelled."];
	return ;
}
    ShowMessage [format [_ "Loaded transliteration righthand side from %s."] $XLFile]
    return ;
}

proc SwapXLInOut {} {
    #Get from windows and store for backup.
    GetTransliterationFromWindows; #Discard result - we only want the side effects.
    if {$::XLRulesInWindows == 0} {
	ShowMessage [_ "There is nothing to swap."];
	return;
    }
    #Delete window contents
    $::XLINP configure -state normal;
    $::XLINP delete 1.0 end;
    $::XLOUT configure -state normal;
    $::XLOUT delete 1.0 end;
    #Cross-insert
    $::XLINP insert 1.0 $::SavedXLTgt
    $::XLOUT insert 1.0 $::SavedXLSrc
    ShowMessage [_ "Swapped transliteration input and output."]
}

#proc MakeFocusFollowMouse {} {
#    tk_focusFollowsMouse;
#    ShowMessage [_ "Focus now follows the mouse. This cannot be reversed."]
#    $::m.misc.configure delete $::FocusFollowsMouseIndex $::FocusFollowsMouseIndex
#}

proc new_dialog_create {class {win "auto"}} {
    if {$win == "auto"} {
        set count 0
        set win ".ndialog[incr count]"
        while {[winfo exists $win]} {
            set win ".ndialog[incr count]"
        }
    }
    toplevel $win -class $class;
    frame $win.info
    pack $win.info -expand yes -fill both -padx 4 -pady 4
    wm title $win $class
    wm group $win .

    after idle [format {
        update idletasks
        wm minsize %s [winfo reqwidth %s] [winfo reqheight %s]
    } $win $win $win]

    return $win
}

# The following code is taken from the Efftcl library by Mark Harrison and
# Michael McLennan, copyrighted by Mark Harrison and Lucent Technologies, available
# from http://www.awprofessional.com/content/images/0201634740/sourcecode/efftcl.zip.
# As the authors explicitly give permission to "steal the code for your own applications"
# the relevant portions are included here so as not to require the user to install
# to install the library. If you install the library, remove the following and
# uncomment the line "#package require Efftcl" by deleting the crosshatch.

#  Effective Tcl/Tk Programming
#    Mark Harrison, DSC Communications Corp.
#    Michael McLennan, Bell Labs Innovations for Lucent Technologies
#    Addison-Wesley Professional Computing Series
# ======================================================================
#  Copyright (c) 1996-1997  Lucent Technologies Inc. and Mark Harrison
# ======================================================================

proc dialog_create {class {win "auto"}} {
    if {$win == "auto"} {
        set count 0
        set win ".dialog[incr count]"
        while {[winfo exists $win]} {
            set win ".dialog[incr count]"
        }
    }
    toplevel $win -class $class -background \#2200DD;

    frame $win.info
    pack $win.info -expand yes -fill both -padx 4 -pady 4
    frame $win.sep -height 2 -borderwidth 1 -relief sunken
    pack $win.sep -fill x -pady 4
    frame $win.controls
    pack $win.controls -fill x -padx 4 -pady 1

    wm title $win $class
    wm group $win .

    after idle [format {
        update idletasks
        wm minsize %s [winfo reqwidth %s] [winfo reqheight %s]
    } $win $win $win]

    return $win
}

proc dialog_info {win} {
    return "$win.info"
}

proc dialog_controls {win} {
    return "$win.controls"
}

proc dialog_wait {win varName} {
    dialog_safeguard $win

    set x [expr [winfo rootx .]+50]
    set y [expr [winfo rooty .]+50]
    wm geometry $win "+$x+$y"

    wm deiconify $win
    grab set $win

    vwait $varName

    grab release $win
    wm withdraw $win
}

#bind modalDialog <ButtonPress> {
#    wm deiconify %W
#    raise %W
#}

proc dialog_safeguard {win} {
    if {[lsearch [bindtags $win] modalDialog] < 0} {
        bindtags $win [linsert [bindtags $win] 0 modalDialog]
    }
}

proc CreateTextDisplay {title width height {bg "\#EDC798"} {fg "\#000080"} } {
    set top [new_dialog_create Textdisplay]
    wm title $top $title
    set info [dialog_info $top]
    scrollbar $info.sbar -command "$info.text yview" 
    pack $info.sbar -side right -fill y
    text $info.text -height $height -width $width -font MainFont \
	-wrap word -yscrollcommand "$info.sbar set" \
	-background $bg -foreground $fg -exportselection 1;
    pack $info.text -side left -expand yes -fill both
    bind $info.sbar <<B3>> "ScrollbarMoveBigIncrement $info.sbar 0.20 %x %y"
    $info.text configure -state disabled
    return $top
}

proc AppendToTextDisplay {top mesg} {
    set info [dialog_info $top]
    $info.text configure -state normal
    $info.text insert end $mesg
    $info.text configure -state disabled
}

set linkNum 0;
proc AppendLinkToTextDisplay {top mesg LinkCode} {
    global linkNum
    set info [dialog_info $top]
    $info.text configure -state normal
    set tag "link[incr linkNum]"
    bind $info.text <ButtonPress> break
    $info.text insert end $mesg [list body $tag]
    $info.text tag configure $tag -foreground red -underline 1
    $info.text tag bind $tag <Enter> \
        "$info.text tag configure $tag -foreground blue"
    $info.text tag bind $tag <Leave> \
        "$info.text tag configure $tag -foreground red"
    $info.text tag bind $tag <ButtonPress> \
        "$LinkCode"
    $info.text configure -state disabled
}

proc WebManual {} {
    global ManualPath;
    global ManualURL;

    if {[file exists $ManualPath] == 0} {
	ShowMessage [format [_ "Manual not found at location %s"] $ManualPath];
	return ;
    }
    ShowWebPage $ManualURL;
}

set AboutPopup "";
proc About {} {
    global Version;
    if {[PopupDown About] ==1} {return}
    set po [CreateTextDisplay [_ "About"] 50 6];
    set ::PopupList(About) $po;
    BindKeys $po;
    AppendToTextDisplay $po [format [_ "This is Xlit version %s \[%s\]."] \
	 $Version [ProgramTimeDateStamp]];
    AppendToTextDisplay $po \
	[_ "It is a tool for transliteration, that is, for converting text from one writing system into another."];
    AppendToTextDisplay $po " The author of Xlit is Bill Poser (billposer@alum.mit.edu).";
    AppendToTextDisplay $po [_ " You can obtain the latest version of Xlit from: "];
    AppendLinkToTextDisplay $po "http://www.billposer.org/Software/xlit.html." {ShowWebPage http://www.billposer.org/Software/xlit.html};
}


proc BugReports {} {
    global Version;
    global tcl_platform;
    if {[PopupDown BugReports] ==1} {return}
    set po [CreateTextDisplay [_ "Bug Reports"] 60 21];
    set ::PopupList(BugReports) $po;
    BindKeys $po;
    AppendToTextDisplay $po [_ "Report bugs to: billposer@alum.mit.edu.\n"];
    AppendToTextDisplay $po [_ "Please include the following information:\n\n"];
    AppendToTextDisplay $po [_ "\tWhat version of the program are you using?\n"];
    AppendToTextDisplay $po [format [_ "\t\t(This is version %s \[%s\].)\n\n"] $Version [ProgramTimeDateStamp]];
    AppendToTextDisplay $po [_ "\tWhat operating system are you running?.\n"];
    set OS $tcl_platform(os);
    if {$OS == "Linux"} {set OS "GNU/Linux"};
    AppendToTextDisplay $po [format [_ "\t\t(This is %s  %s.)\n\n"] $OS  $tcl_platform(osVersion)];
    AppendToTextDisplay $po [_ "\tWhat window system are you using?.\n"];
    AppendToTextDisplay $po [format [_ "\t\t(This is %s.)\n\n"] $::WindowSystem]
    AppendToTextDisplay $po [_ "\tWhat version of tcl/tk are you using?.\n"];
    AppendToTextDisplay $po [format [_ "\t\t(This is version %s.)\n\n"] [info patchlevel]];
    AppendToTextDisplay $po [_ "Bug reports may be sent in any language that I can read without too much trouble or am trying to learn or improve. These include:\n\n"];
    AppendLinkToTextDisplay $po [_ "\tCatalan"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=cat};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tDakelh (Carrier)"] {ShowWebPage http://ydli.org}
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tDutch"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=nld};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tEnglish"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=eng};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tEsperanto"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=epo};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tFrench"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=fra};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tGerman"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=deu};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tItalian"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=ita};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tJapanese"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=jpn};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tKazakh"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=kaz};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tKorean"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=kor};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tLatin"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=lat};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tPortuguese"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=por};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tSpanish"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=spa};
    AppendToTextDisplay $po "\n";
    AppendLinkToTextDisplay $po [_ "\tTurkish"] {ShowWebPage http://www.ethnologue.com/show_language.asp?code=tur};
    AppendToTextDisplay $po "\n\n";
    AppendToTextDisplay $po "Please note that in many cases although I can understand the language my ability to respond in it may be limited.\n";
    AppendToTextDisplay $po "\n";
}

proc CommandLineOptions {} {
    global Version;
    if {[PopupDown CommandLineOptions] ==1} {return}
    set po [CreateTextDisplay [_ "Command Line Options"] 82 9];
    set ::PopupList(CommandLineOptions) $po;
    BindKeys $po;
    AppendToTextDisplay $po [_ "-b              batch mode\n"];
    AppendToTextDisplay $po \
	[_ "-d <duplicate selection>            set DuplicateSelection\n"];
    AppendToTextDisplay $po \
	[_ "-f <filename>   read configuration from <filename> rather than usual init file\n"];
    AppendToTextDisplay $po [_ "-h              help\n"];
    AppendToTextDisplay $po \
	[_ "-l <filename>   load left hand side of transliteration definition from <filename>\n"];
    AppendToTextDisplay $po [_ "-i              do not read init file\n"];
    AppendToTextDisplay $po \
	[_ "-p <ParseChoice>                    set TransliterateWhat\n"];
    AppendToTextDisplay $po [_ "-q              quiet mode - don't chat so much\n"];
    AppendToTextDisplay $po \
	[_ "-t <filename>   load text to be transliterated from <filename>\n"];
    AppendToTextDisplay $po [_ "-v              version\n"];
    AppendToTextDisplay $po   \
	[_ "-x <filename>   load transliteration definition from <filename>\n"];
}

proc ShowGPL {} {
    if {[PopupDown ShowGPL] ==1} {return}
    set po [CreateTextDisplay [_ "License"] 70 24]
    set ::PopupList(ShowGPL) $po
    BindKeys $po
    AppendToTextDisplay $po [format "%s%s" [format "%s\n\t%s\n" [_ "For this license in your language see:"] [_ "http://www.gnu.org/copyleft/gpl.html"]]  "\
\

                       Version 3, 29 June 2007\


 Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.
\

                            Preamble
\

  The GNU General Public License is a free, copyleft license for
software and other kinds of works.
\

  The licenses for most software and other practical works are designed
to take away your freedom to share and change the works.  By contrast,
the GNU General Public License is intended to guarantee your freedom to
share and change all versions of a program--to make sure it remains free
software for all its users.  We, the Free Software Foundation, use the
GNU General Public License for most of our software; it applies also to
any other work released this way by its authors.  You can apply it to
your programs, too.
\

  When we speak of free software, we are referring to freedom, not
price.  Our General Public Licenses are designed to make sure that you
have the freedom to distribute copies of free software (and charge for
them if you wish), that you receive source code or can get it if you
want it, that you can change the software or use pieces of it in new
free programs, and that you know you can do these things.
\

  To protect your rights, we need to prevent others from denying you
these rights or asking you to surrender the rights.  Therefore, you have
certain responsibilities if you distribute copies of the software, or if
you modify it: responsibilities to respect the freedom of others.
\

  For example, if you distribute copies of such a program, whether
gratis or for a fee, you must pass on to the recipients the same
freedoms that you received.  You must make sure that they, too, receive
or can get the source code.  And you must show them these terms so they
know their rights.
\

  Developers that use the GNU GPL protect your rights with two steps:
(1) assert copyright on the software, and (2) offer you this License
giving you legal permission to copy, distribute and/or modify it.
\

  For the developers' and authors' protection, the GPL clearly explains
that there is no warranty for this free software.  For both users' and
authors' sake, the GPL requires that modified versions be marked as
changed, so that their problems will not be attributed erroneously to
authors of previous versions.
\

  Some devices are designed to deny users access to install or run
modified versions of the software inside them, although the manufacturer
can do so.  This is fundamentally incompatible with the aim of
protecting users' freedom to change the software.  The systematic
pattern of such abuse occurs in the area of products for individuals to
use, which is precisely where it is most unacceptable.  Therefore, we
have designed this version of the GPL to prohibit the practice for those
products.  If such problems arise substantially in other domains, we
stand ready to extend this provision to those domains in future versions
of the GPL, as needed to protect the freedom of users.
\

  Finally, every program is threatened constantly by software patents.
States should not allow patents to restrict development and use of
software on general-purpose computers, but in those that do, we wish to
avoid the special danger that patents applied to a free program could
make it effectively proprietary.  To prevent this, the GPL assures that
patents cannot be used to render the program non-free.
\

  The precise terms and conditions for copying, distribution and
modification follow.
\

                       TERMS AND CONDITIONS
\

  0. Definitions.
\

  \"This License\" refers to version 3 of the GNU General Public License.
\

  \"Copyright\" also means copyright-like laws that apply to other kinds of
works, such as semiconductor masks.
\

  \"The Program\" refers to any copyrightable work licensed under this
License.  Each licensee is addressed as \"you\".  \"Licensees\" and
\"recipients\" may be individuals or organizations.
\

  To \"modify\" a work means to copy from or adapt all or part of the work
in a fashion requiring copyright permission, other than the making of an
exact copy.  The resulting work is called a \"modified version\" of the
earlier work or a work \"based on\" the earlier work.
\

  A \"covered work\" means either the unmodified Program or a work based
on the Program.
\

  To \"propagate\" a work means to do anything with it that, without
permission, would make you directly or secondarily liable for
infringement under applicable copyright law, except executing it on a
computer or modifying a private copy.  Propagation includes copying,
distribution (with or without modification), making available to the
public, and in some countries other activities as well.
\

  To \"convey\" a work means any kind of propagation that enables other
parties to make or receive copies.  Mere interaction with a user through
a computer network, with no transfer of a copy, is not conveying.
\

  An interactive user interface displays \"Appropriate Legal Notices\"
to the extent that it includes a convenient and prominently visible
feature that (1) displays an appropriate copyright notice, and (2)
tells the user that there is no warranty for the work (except to the
extent that warranties are provided), that licensees may convey the
work under this License, and how to view a copy of this License.  If
the interface presents a list of user commands or options, such as a
menu, a prominent item in the list meets this criterion.
\

  1. Source Code.
\

  The \"source code\" for a work means the preferred form of the work
for making modifications to it.  \"Object code\" means any non-source
form of a work.
\

  A \"Standard Interface\" means an interface that either is an official
standard defined by a recognized standards body, or, in the case of
interfaces specified for a particular programming language, one that
is widely used among developers working in that language.
\

  The \"System Libraries\" of an executable work include anything, other
than the work as a whole, that (a) is included in the normal form of
packaging a Major Component, but which is not part of that Major
Component, and (b) serves only to enable use of the work with that
Major Component, or to implement a Standard Interface for which an
implementation is available to the public in source code form.  A
\"Major Component\", in this context, means a major essential component
(kernel, window system, and so on) of the specific operating system
(if any) on which the executable work runs, or a compiler used to
produce the work, or an object code interpreter used to run it.
\

  The \"Corresponding Source\" for a work in object code form means all
the source code needed to generate, install, and (for an executable
work) run the object code and to modify the work, including scripts to
control those activities.  However, it does not include the work's
System Libraries, or general-purpose tools or generally available free
programs which are used unmodified in performing those activities but
which are not part of the work.  For example, Corresponding Source
includes interface definition files associated with source files for
the work, and the source code for shared libraries and dynamically
linked subprograms that the work is specifically designed to require,
such as by intimate data communication or control flow between those
subprograms and other parts of the work.
\

  The Corresponding Source need not include anything that users
can regenerate automatically from other parts of the Corresponding
Source.
\

  The Corresponding Source for a work in source code form is that
same work.
\

  2. Basic Permissions.
\

  All rights granted under this License are granted for the term of
copyright on the Program, and are irrevocable provided the stated
conditions are met.  This License explicitly affirms your unlimited
permission to run the unmodified Program.  The output from running a
covered work is covered by this License only if the output, given its
content, constitutes a covered work.  This License acknowledges your
rights of fair use or other equivalent, as provided by copyright law.
\

  You may make, run and propagate covered works that you do not
convey, without conditions so long as your license otherwise remains
in force.  You may convey covered works to others for the sole purpose
of having them make modifications exclusively for you, or provide you
with facilities for running those works, provided that you comply with
the terms of this License in conveying all material for which you do
not control copyright.  Those thus making or running the covered works
for you must do so exclusively on your behalf, under your direction
and control, on terms that prohibit them from making any copies of
your copyrighted material outside their relationship with you.
\

  Conveying under any other circumstances is permitted solely under
the conditions stated below.  Sublicensing is not allowed; section 10
makes it unnecessary.
\

  3. Protecting Users' Legal Rights From Anti-Circumvention Law.
\

  No covered work shall be deemed part of an effective technological
measure under any applicable law fulfilling obligations under article
11 of the WIPO copyright treaty adopted on 20 December 1996, or
similar laws prohibiting or restricting circumvention of such
measures.
\

  When you convey a covered work, you waive any legal power to forbid
circumvention of technological measures to the extent such circumvention
is effected by exercising rights under this License with respect to
the covered work, and you disclaim any intention to limit operation or
modification of the work as a means of enforcing, against the work's
users, your or third parties' legal rights to forbid circumvention of
technological measures.
\

  4. Conveying Verbatim Copies.
\

  You may convey verbatim copies of the Program's source code as you
receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy an appropriate copyright notice;
keep intact all notices stating that this License and any
non-permissive terms added in accord with section 7 apply to the code;
keep intact all notices of the absence of any warranty; and give all
recipients a copy of this License along with the Program.
\

  You may charge any price or no price for each copy that you convey,
and you may offer support or warranty protection for a fee.
\

  5. Conveying Modified Source Versions.
\

  You may convey a work based on the Program, or the modifications to
produce it from the Program, in the form of source code under the
terms of section 4, provided that you also meet all of these conditions:
\

    a) The work must carry prominent notices stating that you modified
    it, and giving a relevant date.
\

    b) The work must carry prominent notices stating that it is
    released under this License and any conditions added under section
    7.  This requirement modifies the requirement in section 4 to
    \"keep intact all notices\".
\

    c) You must license the entire work, as a whole, under this
    License to anyone who comes into possession of a copy.  This
    License will therefore apply, along with any applicable section 7
    additional terms, to the whole of the work, and all its parts,
    regardless of how they are packaged.  This License gives no
    permission to license the work in any other way, but it does not
    invalidate such permission if you have separately received it.
\

    d) If the work has interactive user interfaces, each must display
    Appropriate Legal Notices; however, if the Program has interactive
    interfaces that do not display Appropriate Legal Notices, your
    work need not make them do so.
\

  A compilation of a covered work with other separate and independent
works, which are not by their nature extensions of the covered work,
and which are not combined with it such as to form a larger program,
in or on a volume of a storage or distribution medium, is called an
\"aggregate\" if the compilation and its resulting copyright are not
used to limit the access or legal rights of the compilation's users
beyond what the individual works permit.  Inclusion of a covered work
in an aggregate does not cause this License to apply to the other
parts of the aggregate.
\

  6. Conveying Non-Source Forms.
\

  You may convey a covered work in object code form under the terms
of sections 4 and 5, provided that you also convey the
machine-readable Corresponding Source under the terms of this License,
in one of these ways:
\

    a) Convey the object code in, or embodied in, a physical product
    (including a physical distribution medium), accompanied by the
    Corresponding Source fixed on a durable physical medium
    customarily used for software interchange.
\

    b) Convey the object code in, or embodied in, a physical product
    (including a physical distribution medium), accompanied by a
    written offer, valid for at least three years and valid for as
    long as you offer spare parts or customer support for that product
    model, to give anyone who possesses the object code either (1) a
    copy of the Corresponding Source for all the software in the
    product that is covered by this License, on a durable physical
    medium customarily used for software interchange, for a price no
    more than your reasonable cost of physically performing this
    conveying of source, or (2) access to copy the
    Corresponding Source from a network server at no charge.
\

    c) Convey individual copies of the object code with a copy of the
    written offer to provide the Corresponding Source.  This
    alternative is allowed only occasionally and noncommercially, and
    only if you received the object code with such an offer, in accord
    with subsection 6b.
\

    d) Convey the object code by offering access from a designated
    place (gratis or for a charge), and offer equivalent access to the
    Corresponding Source in the same way through the same place at no
    further charge.  You need not require recipients to copy the
    Corresponding Source along with the object code.  If the place to
    copy the object code is a network server, the Corresponding Source
    may be on a different server (operated by you or a third party)
    that supports equivalent copying facilities, provided you maintain
    clear directions next to the object code saying where to find the
    Corresponding Source.  Regardless of what server hosts the
    Corresponding Source, you remain obligated to ensure that it is
    available for as long as needed to satisfy these requirements.
\

    e) Convey the object code using peer-to-peer transmission, provided
    you inform other peers where the object code and Corresponding
    Source of the work are being offered to the general public at no
    charge under subsection 6d.
\

  A separable portion of the object code, whose source code is excluded
from the Corresponding Source as a System Library, need not be
included in conveying the object code work.
\

  A \"User Product\" is either (1) a \"consumer product\", which means any
tangible personal property which is normally used for personal, family,
or household purposes, or (2) anything designed or sold for incorporation
into a dwelling.  In determining whether a product is a consumer product,
doubtful cases shall be resolved in favor of coverage.  For a particular
product received by a particular user, \"normally used\" refers to a
typical or common use of that class of product, regardless of the status
of the particular user or of the way in which the particular user
actually uses, or expects or is expected to use, the product.  A product
is a consumer product regardless of whether the product has substantial
commercial, industrial or non-consumer uses, unless such uses represent
the only significant mode of use of the product.
\

  \"Installation Information\" for a User Product means any methods,
procedures, authorization keys, or other information required to install
and execute modified versions of a covered work in that User Product from
a modified version of its Corresponding Source.  The information must
suffice to ensure that the continued functioning of the modified object
code is in no case prevented or interfered with solely because
modification has been made.
\

  If you convey an object code work under this section in, or with, or
specifically for use in, a User Product, and the conveying occurs as
part of a transaction in which the right of possession and use of the
User Product is transferred to the recipient in perpetuity or for a
fixed term (regardless of how the transaction is characterized), the
Corresponding Source conveyed under this section must be accompanied
by the Installation Information.  But this requirement does not apply
if neither you nor any third party retains the ability to install
modified object code on the User Product (for example, the work has
been installed in ROM).
\

  The requirement to provide Installation Information does not include a
requirement to continue to provide support service, warranty, or updates
for a work that has been modified or installed by the recipient, or for
the User Product in which it has been modified or installed.  Access to a
network may be denied when the modification itself materially and
adversely affects the operation of the network or violates the rules and
protocols for communication across the network.
\

  Corresponding Source conveyed, and Installation Information provided,
in accord with this section must be in a format that is publicly
documented (and with an implementation available to the public in
source code form), and must require no special password or key for
unpacking, reading or copying.
\

  7. Additional Terms.
\

  \"Additional permissions\" are terms that supplement the terms of this
License by making exceptions from one or more of its conditions.
Additional permissions that are applicable to the entire Program shall
be treated as though they were included in this License, to the extent
that they are valid under applicable law.  If additional permissions
apply only to part of the Program, that part may be used separately
under those permissions, but the entire Program remains governed by
this License without regard to the additional permissions.
\

  When you convey a copy of a covered work, you may at your option
remove any additional permissions from that copy, or from any part of
it.  (Additional permissions may be written to require their own
removal in certain cases when you modify the work.)  You may place
additional permissions on material, added by you to a covered work,
for which you have or can give appropriate copyright permission.
\

  Notwithstanding any other provision of this License, for material you
add to a covered work, you may (if authorized by the copyright holders of
that material) supplement the terms of this License with terms:
\

    a) Disclaiming warranty or limiting liability differently from the
    terms of sections 15 and 16 of this License; or
\

    b) Requiring preservation of specified reasonable legal notices or
    author attributions in that material or in the Appropriate Legal
    Notices displayed by works containing it; or
\

    c) Prohibiting misrepresentation of the origin of that material, or
    requiring that modified versions of such material be marked in
    reasonable ways as different from the original version; or
\

    d) Limiting the use for publicity purposes of names of licensors or
    authors of the material; or
\

    e) Declining to grant rights under trademark law for use of some
    trade names, trademarks, or service marks; or
\

    f) Requiring indemnification of licensors and authors of that
    material by anyone who conveys the material (or modified versions of
    it) with contractual assumptions of liability to the recipient, for
    any liability that these contractual assumptions directly impose on
    those licensors and authors.
\

  All other non-permissive additional terms are considered \"further
restrictions\" within the meaning of section 10.  If the Program as you
received it, or any part of it, contains a notice stating that it is
governed by this License along with a term that is a further
restriction, you may remove that term.  If a license document contains
a further restriction but permits relicensing or conveying under this
License, you may add to a covered work material governed by the terms
of that license document, provided that the further restriction does
not survive such relicensing or conveying.
\

  If you add terms to a covered work in accord with this section, you
must place, in the relevant source files, a statement of the
additional terms that apply to those files, or a notice indicating
where to find the applicable terms.
\

  Additional terms, permissive or non-permissive, may be stated in the
form of a separately written license, or stated as exceptions;
the above requirements apply either way.
\

  8. Termination.
\

  You may not propagate or modify a covered work except as expressly
provided under this License.  Any attempt otherwise to propagate or
modify it is void, and will automatically terminate your rights under
this License (including any patent licenses granted under the third
paragraph of section 11).
\

  However, if you cease all violation of this License, then your
license from a particular copyright holder is reinstated (a)
provisionally, unless and until the copyright holder explicitly and
finally terminates your license, and (b) permanently, if the copyright
holder fails to notify you of the violation by some reasonable means
prior to 60 days after the cessation.
\

  Moreover, your license from a particular copyright holder is
reinstated permanently if the copyright holder notifies you of the
violation by some reasonable means, this is the first time you have
received notice of violation of this License (for any work) from that
copyright holder, and you cure the violation prior to 30 days after
your receipt of the notice.
\

  Termination of your rights under this section does not terminate the
licenses of parties who have received copies or rights from you under
this License.  If your rights have been terminated and not permanently
reinstated, you do not qualify to receive new licenses for the same
material under section 10.
\

  9. Acceptance Not Required for Having Copies.
\

  You are not required to accept this License in order to receive or
run a copy of the Program.  Ancillary propagation of a covered work
occurring solely as a consequence of using peer-to-peer transmission
to receive a copy likewise does not require acceptance.  However,
nothing other than this License grants you permission to propagate or
modify any covered work.  These actions infringe copyright if you do
not accept this License.  Therefore, by modifying or propagating a
covered work, you indicate your acceptance of this License to do so.
\

  10. Automatic Licensing of Downstream Recipients.

  Each time you convey a covered work, the recipient automatically
receives a license from the original licensors, to run, modify and
propagate that work, subject to this License.  You are not responsible
for enforcing compliance by third parties with this License.
\

  An \"entity transaction\" is a transaction transferring control of an
organization, or substantially all assets of one, or subdividing an
organization, or merging organizations.  If propagation of a covered
work results from an entity transaction, each party to that
transaction who receives a copy of the work also receives whatever
licenses to the work the party's predecessor in interest had or could
give under the previous paragraph, plus a right to possession of the
Corresponding Source of the work from the predecessor in interest, if
the predecessor has it or can get it with reasonable efforts.
\

  You may not impose any further restrictions on the exercise of the
rights granted or affirmed under this License.  For example, you may
not impose a license fee, royalty, or other charge for exercise of
rights granted under this License, and you may not initiate litigation
(including a cross-claim or counterclaim in a lawsuit) alleging that
any patent claim is infringed by making, using, selling, offering for
sale, or importing the Program or any portion of it.
\

  11. Patents.
\

  A \"contributor\" is a copyright holder who authorizes use under this
License of the Program or a work on which the Program is based.  The
work thus licensed is called the contributor's \"contributor version\".
\

  A contributor's \"essential patent claims\" are all patent claims
owned or controlled by the contributor, whether already acquired or
hereafter acquired, that would be infringed by some manner, permitted
by this License, of making, using, or selling its contributor version,
but do not include claims that would be infringed only as a
consequence of further modification of the contributor version.  For
purposes of this definition, \"control\" includes the right to grant
patent sublicenses in a manner consistent with the requirements of
this License.
\

  Each contributor grants you a non-exclusive, worldwide, royalty-free
patent license under the contributor's essential patent claims, to
make, use, sell, offer for sale, import and otherwise run, modify and
propagate the contents of its contributor version.
\

  In the following three paragraphs, a \"patent license\" is any express
agreement or commitment, however denominated, not to enforce a patent
(such as an express permission to practice a patent or covenant not to
sue for patent infringement).  To \"grant\" such a patent license to a
party means to make such an agreement or commitment not to enforce a
patent against the party.
\

  If you convey a covered work, knowingly relying on a patent license,
and the Corresponding Source of the work is not available for anyone
to copy, free of charge and under the terms of this License, through a
publicly available network server or other readily accessible means,
then you must either (1) cause the Corresponding Source to be so
available, or (2) arrange to deprive yourself of the benefit of the
patent license for this particular work, or (3) arrange, in a manner
consistent with the requirements of this License, to extend the patent
license to downstream recipients.  \"Knowingly relying\" means you have
actual knowledge that, but for the patent license, your conveying the
covered work in a country, or your recipient's use of the covered work
in a country, would infringe one or more identifiable patents in that
country that you have reason to believe are valid.
\

  If, pursuant to or in connection with a single transaction or
arrangement, you convey, or propagate by procuring conveyance of, a
covered work, and grant a patent license to some of the parties
receiving the covered work authorizing them to use, propagate, modify
or convey a specific copy of the covered work, then the patent license
you grant is automatically extended to all recipients of the covered
work and works based on it.
\

  A patent license is \"discriminatory\" if it does not include within
the scope of its coverage, prohibits the exercise of, or is
conditioned on the non-exercise of one or more of the rights that are
specifically granted under this License.  You may not convey a covered
work if you are a party to an arrangement with a third party that is
in the business of distributing software, under which you make payment
to the third party based on the extent of your activity of conveying
the work, and under which the third party grants, to any of the
parties who would receive the covered work from you, a discriminatory
patent license (a) in connection with copies of the covered work
conveyed by you (or copies made from those copies), or (b) primarily
for and in connection with specific products or compilations that
contain the covered work, unless you entered into that arrangement,
or that patent license was granted, prior to 28 March 2007.
\

  Nothing in this License shall be construed as excluding or limiting
any implied license or other defenses to infringement that may
otherwise be available to you under applicable patent law.
\

  12. No Surrender of Others' Freedom.
\

  If conditions are imposed on you (whether by court order, agreement or
otherwise) that contradict the conditions of this License, they do not
excuse you from the conditions of this License.  If you cannot convey a
covered work so as to satisfy simultaneously your obligations under this
License and any other pertinent obligations, then as a consequence you may
not convey it at all.  For example, if you agree to terms that obligate you
to collect a royalty for further conveying from those to whom you convey
the Program, the only way you could satisfy both those terms and this
License would be to refrain entirely from conveying the Program.
\

  13. Use with the GNU Affero General Public License.
\

  Notwithstanding any other provision of this License, you have
permission to link or combine any covered work with a work licensed
under version 3 of the GNU Affero General Public License into a single
combined work, and to convey the resulting work.  The terms of this
License will continue to apply to the part which is the covered work,
but the special requirements of the GNU Affero General Public License,
section 13, concerning interaction through a network will apply to the
combination as such.
\

  14. Revised Versions of this License.
\

  The Free Software Foundation may publish revised and/or new versions of
the GNU General Public License from time to time.  Such new versions will
be similar in spirit to the present version, but may differ in detail to
address new problems or concerns.
\

  Each version is given a distinguishing version number.  If the
Program specifies that a certain numbered version of the GNU General
Public License \"or any later version\" applies to it, you have the
option of following the terms and conditions either of that numbered
version or of any later version published by the Free Software
Foundation.  If the Program does not specify a version number of the
GNU General Public License, you may choose any version ever published
by the Free Software Foundation.
\

  If the Program specifies that a proxy can decide which future
versions of the GNU General Public License can be used, that proxy's
public statement of acceptance of a version permanently authorizes you
to choose that version for the Program.
\

  Later license versions may give you additional or different
permissions.  However, no additional obligations are imposed on any
author or copyright holder as a result of your choosing to follow a
later version.
\

  15. Disclaimer of Warranty.
\

  THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY
OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
\

  16. Limitation of Liability.
\

  IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
\

  17. Interpretation of Sections 15 and 16.
\

  If the disclaimer of warranty and limitation of liability provided
above cannot be given local legal effect according to their terms,
reviewing courts shall apply local law that most closely approximates
an absolute waiver of all civil liability in connection with the
Program, unless a warranty or assumption of liability accompanies a
copy of the Program in return for a fee.
"]
}


proc SetupBalloonHelp {} {
    global bhInfo

    option add *Balloonhelp*background white widgetDefault
    option add *Balloonhelp*foreground black widgetDefault
    option add *Balloonhelpinfo.wrapLength 3i  widgetDefault
    option add *Balloonhelp.info.justify left widgetDefault
    toplevel .balloonhelp -class Balloonhelp -background black -borderwidth 1 -relief flat
    label .balloonhelp.info -font BalloonHelpFont;
    pack .balloonhelp.info -side left -fill y
    wm overrideredirect .balloonhelp 1
    wm withdraw .balloonhelp
    set bhInfo(active) 1
}

proc balloonhelp_control {state} {
     global bhInfo
     if {$state} {
          set bhInfo(active) 1
     } else {
	balloonhelp_cancel
	set bhInfo(active) 0
     }
}

proc balloonhelp_for {win mesg} {
    global bhInfo
    set bhInfo($win) $mesg
    set ::bhOverlapP($win) 1; 
    bind $win <Enter> {+balloonhelp_pending %W}
    bind $win <Leave> {+balloonhelp_cancel}
}

proc balloonhelpd_for {win mesg} {
    global bhInfo
    set ::bhOverlapP($win) 0;
    set bhInfo($win) $mesg
    bind $win <Enter> {+balloonhelp_show %W}
    bind $win <Leave> {+wm withdraw .balloonhelp}
}

proc balloonhelp_pending {win} {
     global bhInfo
     balloonhelp_cancel
     set bhInfo(pending) [after 1000 [list balloonhelp_show $win]]
}

proc balloonhelp_cancel {} {
    global bhInfo
    if { [info exists bhInfo(pending)]} {
	after cancel $bhInfo(pending)
	unset bhInfo(pending)
    }
    if {$::TkPresentP} {
	wm withdraw .balloonhelp
    }
}

proc balloonhelp_show {win} {
    global bhInfo;
    global bhOverlapP;
    if {$bhOverlapP($win)} {
	set Overlap 25;
    } else {
	set Overlap -10;
    }
    if {[winfo exists $win]} {
	if {$bhInfo(active)} {
	    .balloonhelp.info configure -text $bhInfo($win)
	    #Set abcissa
	    set MaxStringWidth 0;
	    foreach line [split $bhInfo($win) "\n"] {
		set StringWidth [font measure BalloonHelpFont -displayof .balloonhelp.info $line]
		if {$StringWidth > $MaxStringWidth} {
		    set MaxStringWidth $StringWidth;
		}
	    }
	    set ScreenWidth [winfo screenwidth $win]
	    set Width [winfo width $win];
	    set LeftEdge  [winfo rootx $win];
	    set RightEdge [expr $LeftEdge + $Width];
	    if {$ScreenWidth - $RightEdge < $MaxStringWidth} { 
		if {$LeftEdge > $MaxStringWidth} {
		    set x [expr $LeftEdge - $MaxStringWidth + $Overlap];
		} else {
		    if {$ScreenWidth - $MaxStringWidth > 0} {
			set x [expr $RightEdge - $MaxStringWidth];
		    } else {
			set x [expr $ScreenWidth - $MaxStringWidth];
		    }
		}
	    } else {
		set x [expr $RightEdge - $Overlap];
	    }
	    #Set ordinate
	    set Height [winfo height $win];
	    set TopEdge [winfo rooty $win];
#	    set y [expr $TopEdge + ($Height/2)];
	    set y [expr $TopEdge + int(($Height/1.25))];
	    wm geometry .balloonhelp +$x+$y
	    wm deiconify .balloonhelp
	    raise .balloonhelp
	}
    }
    if {[info exist bhInfo(pending)]} {
	unset bhInfo(pending)
    }
}

proc ToggleBalloonHelp {} {
    global BalloonHelpIndex;   
    global m;
    if {$::BalloonHelpP} {
	set ::BalloonHelpP 0;
	balloonhelp_control 0
	ShowMessage [_ "Irritating Balloon Help Disabled"];
	$m.misc.configure entryconfigure $BalloonHelpIndex -label [_ "Show Balloon Help"];
    } else {
	set ::BalloonHelpP 1;
	balloonhelp_control 1
	ShowMessage [_ "Balloon Help Enabled"];
	$m.misc.configure entryconfigure $BalloonHelpIndex -label [_ "Hide Irritating Balloon Help"];
    }
}

#Character entry code begins here
#Set this to the path to character entry menu
set CEM .menubar.charentry
#Set the default insertion target here.
set InsertionTarget $XLINP
set UseNoGlyphsP 0; #Should we include in widgets characters not in Code2000?

set CharentryMenuItems 6;
set IPAAIsDisplayedP 0;
set IPACIsDisplayedP 0;
set IPAVIsDisplayedP 0;
set IPADIsDisplayedP 0;
set CharEntryByCodeIsDisplayedP 0;

set DisplayConsonantChartColumnLabelsP 1
set DisplayConsonantChartRowLabelsP 1
set DisplayVowelChartColumnLabelsP 1
set DisplayVowelChartRowLabelsP 1

proc BindInsertionTarget {v} {
    bind $v <FocusIn> "+SetInsertionTargets $v"
}

proc SetCharacterEntryOptions {} {
    global ColorSpecs

    option add *ipaec.Button.font CharacterEntryFont 100
    option add *ipaec.Label.background  $ColorSpecs(CharacterEntry,Background)
    option add *ipaec.Button.background $ColorSpecs(CharacterEntry,Background)
    option add *ipaec.Button.foreground $ColorSpecs(CharacterEntry,Foreground)
    option add *ipaec.Label.relief raised

    option add *ipaev.Button.font CharacterEntryFont 100
    option add *ipaev.Label.background  $ColorSpecs(CharacterEntry,Background)
    option add *ipaev.Button.background $ColorSpecs(CharacterEntry,Background)
    option add *ipaev.Button.foreground $ColorSpecs(CharacterEntry,Foreground)
    option add *ipaev.Label.relief raised

    option add *ipaea.Button.font AccentedLetterFont 100
    option add *ipaea.Label.background  "\#E0E0E0"
    option add *ipaea.Button.background "\#FFFFFF"
    option add *ipaea.Button.foreground "\#000000"
    option add *ipaea.Label.relief raised

    option add *ipaed.Button.font DiacriticFont 100
    option add *ipaed.Button.background "\#E0E0FF"
    option add *ipaed.Label.background "\#E0E0E0"
    option add *ipaed.Button.foreground "\#000000"
    option add *ipaed.Label.relief raised

    option add *ipae*.*Label.font CharacterEntryLabelFont 100
    option add *ipaec.*Button.font CharacterEntryFont 100
}

proc SetInsertionTargets {tgt} {
    set ::InsertionTarget $tgt;
    set ::CharByCodeInsertionTarget $tgt;
}
#Insert character entry stuff here
set CustomCEMEntries 0;
namespace eval ipaentry {
    variable IPAECColumnLabelList [list];
    variable IPAECRowLabelList [list];
    variable IPAEVColumnLabelList [list];
    variable IPAEVRowLabelList [list];

    set row 0;
    variable poa $row;
    incr row;
    variable vlstop $row;
    incr row;
    variable vdstop $row;
    incr row;
    variable vdimplosive $row;
    incr row;
    variable click $row;
    incr row;
    variable nasal  $row;
    incr row;
    variable trill $row;
    incr row;
    variable tap $row;
    incr row;
    variable vlfric $row;
    incr row;
    variable vdfric $row;
    incr row;
    variable vlaffric $row;
    incr row;
    variable vdaffric $row;
    incr row;
    variable approx $row;
    incr row; 
    variable latapprox $row;
    variable lastrow $row;

    set col 0;
    variable manner $col; 
    incr col;
    variable labial $col; 
    incr col;
    variable labiodental $col; 
    incr col;
    variable dental $col; 
    incr col;
    variable alveolar $col;
    incr col;
    variable postalveolar $col; 
    incr col;
    variable retroflex $col; 
    incr col;
    variable palatal $col; 
    incr col;
    variable velar $col; 
    incr col;
    variable uvular $col; 
    incr col;
    variable pharyngeal $col; 
    incr col;
    variable epiglottal $col; 
    incr col;
    variable glottal $col;
    variable lastcolumn $col;

    proc cfesh {} {
	variable vlfric
	variable postalveolar
	eval [list [format ".ipaec.r%dc%d" $vlfric $postalveolar] configure -text "\u0161"];
    }
    proc cresh {} {
	variable vlfric
	variable postalveolar
	eval [list [format ".ipaec.r%dc%d" $vlfric $postalveolar] configure -text "\u0283"];
    }

    proc cfeshr {} {
	variable vlfric
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vlfric $retroflex] configure -text "\u1E63"];
    }
    proc creshr {} {
	variable vlfric
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vlfric $retroflex] configure -text "\u0282"];
    }

    proc cryod {} {
	variable approx
	variable palatal
	eval [list [format ".ipaec.r%dc%d" $approx $palatal] configure -text "j"];
    }

    proc cfyod {} {
	variable approx
	variable palatal
	eval [list [format ".ipaec.r%dc%d" $approx $palatal] configure -text "y"];
    }

    proc crny {} {
	variable nasal
	variable palatal
	eval [list [format ".ipaec.r%dc%d" $nasal $palatal] configure -text "\u0272"];
    }
    proc cfny {} {
	variable nasal
	variable palatal
	eval [list [format ".ipaec.r%dc%d" $nasal $palatal] configure -text "\u00F1"];
    }

    proc crch {} {
	variable vlaffric
	variable postalveolar
	eval [list [format ".ipaec.r%dc%d" $vlaffric $postalveolar] configure -text "t\u0283"];
    }
    proc cfch {} {
	variable vlaffric
	variable postalveolar
	eval [list [format ".ipaec.r%dc%d" $vlaffric $postalveolar] configure -text "\u010D"];
    }

    proc crchr {} {
	variable vlaffric
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vlaffric $retroflex] configure -text "\u0288\u0282"];
    }

    proc cfchr {} {
	variable vlaffric
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vlaffric $retroflex] configure -text "\u1E6D\u1E63"];
    }


    proc crdj {} {
	eval [list [format ".ipaec.r%dc%d" $ipaentry::vdaffric $ipaentry::postalveolar] configure -text "d\u0292"];
    }

    proc cfdj {} {
	eval [list [format ".ipaec.r%dc%d" $ipaentry::vdaffric $ipaentry::postalveolar] configure -text "\u01C6"];
    }

    proc crdjr {} {
	eval [list [format ".ipaec.r%dc%d" $ipaentry::vdaffric $ipaentry::retroflex] configure -text "\u0256\u0290"];
    }

    proc cfdjr {} {
	eval [list [format ".ipaec.r%dc%d" $ipaentry::vdaffric $ipaentry::retroflex] configure -text "\u1E0D\u1E93"];
    }

    proc crzh {} {
	variable vdfric
	variable postalveolar
	eval [list [format ".ipaec.r%dc%d" $vdfric $postalveolar] configure -text "\u0292"];
    }
    proc cfzh {} {
	variable vdfric
	variable postalveolar
	eval [list [format ".ipaec.r%dc%d" $vdfric $postalveolar] configure -text "\u017E"];
    }

    proc crtr {} {
	variable vlstop
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vlstop $retroflex] configure -text "\u0288"];
    }
    proc cftr {} {
	variable vlstop
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vlstop $retroflex] configure -text "\u1E6D"];
    }

    proc crdr {} {
	variable vdstop
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vdstop $retroflex] configure -text "\u0256"];
    }
    proc cfdr {} {
	variable vdstop
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vdstop $retroflex] configure -text "\u1E0D"];
    }


    proc crnr {} {
	variable nasal
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $nasal $retroflex] configure -text "\u0273"];
    }
    proc cfnr {} {
	variable nasal
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $nasal $retroflex] configure -text "\u1E47"];
    }


    proc crzhr {} {
	variable vdfric
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vdfric $retroflex] configure -text "\u0290"];
    }
    proc cfzhr {} {
	variable vdfric
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $vdfric $retroflex] configure -text "\u1E93"];
    }

    proc crflapr {} {
	variable tap;
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $tap $retroflex] configure -text "\u027D"];
    }
    proc cfflapr {} {
	variable tap;
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $tap $retroflex] configure -text "\u1E5B"];
    }

    proc crlr {} {
	variable latapprox;
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $latapprox $retroflex] configure -text "\u026D"];
    }
    proc cflr {} {
	variable latapprox;
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $latapprox $retroflex] configure -text "\u1E37"];
    }

    proc crrr {} {
	variable approx;
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $approx $retroflex] configure -text "\u027B"];
    }
    proc cfrr {} {
	variable approx;
	variable retroflex
	eval [list [format ".ipaec.r%dc%d" $approx $retroflex] configure -text "\u1E5B"];
    }

    proc bhcham {w} {
	balloonhelp_for $w "The standard IPA character is shown by default. When the mouse\npointer is over the button, the \"North American\" variant is shown.\nRight click to insert the variant."
    }

    proc bhsp {w} {
	balloonhelp_for $w "The characters below and to the right are characters\nthat do not fit neatly into the chart."
    }

    proc UnpackConsonantColumnLabels {} {
	variable IPAECColumnLabelList;
	foreach l $IPAECColumnLabelList {
	    grid forget $l;
	}
    }

    proc PackConsonantColumnLabels {} {
	destroy .ipaec
	PopupIPAEntryC;
    }

    proc UnpackConsonantRowLabels {} {
	variable IPAECRowLabelList;
	foreach l $IPAECRowLabelList {
	    grid forget $l;
	}
    }

    proc PackConsonantRowLabels {} {
	destroy .ipaec;
	PopupIPAEntryC;
    }

    proc UnpackVowelColumnLabels {} {
	variable IPAEVColumnLabelList;
	foreach l $IPAEVColumnLabelList {
	    grid forget $l;
	}
    }

    proc PackVowelColumnLabels {} {
	destroy .ipaev;
	PopupIPAEntryV;
    }

    proc UnpackVowelRowLabels {} {
	variable IPAEVRowLabelList;
	foreach l $IPAEVRowLabelList {
	    grid forget $l;
	}
    }

    proc PackVowelRowLabels {} {
	destroy .ipaev;
	PopupIPAEntryV;
    }

    proc PopupIPAEntryC {} {
	variable poa
	variable vlstop 
	variable vdstop
	variable vdimplosive 
	variable click 
	variable nasal  
	variable trill 
	variable tap 
	variable vlfric 
	variable vdfric 
	variable vlaffric 
	variable vdaffric 
	variable approx 
	variable latapprox
	variable lastrow

	variable manner 
	variable labial 
	variable labiodental 
	variable dental 
	variable alveolar 
	variable postalveolar 
	variable retroflex 
	variable palatal 
	variable velar 
	variable uvular 
	variable pharyngeal 
	variable epiglottal 
	variable glottal
	variable lastcolumn

	variable IPAECColumnLabelList;
	variable IPAECRowLabelList;
	global DisplayConsonantChartColumnLabelsP
	global DisplayConsonantChartRowLabelsP

	set IPAECColumnLabelList [list];
	set IPAECRowLabelList [list];
	
	set xp 2;
	set yp 3;
	set spcolor 	\#c46276;
	toplevel .ipaec -borderwidth 3 -class CharEntry
	wm title .ipaec [_ "Consonant Symbols"]
	BindKeys .ipaec;
	set DownMsg [_ "Display IPA Consonant Chart"];
	set UpMsg   [_ "Remove IPA Consonant Chart"];
	bind .ipaec <Destroy> \
	    "set ::IPACIsDisplayedP 0;$::CEM entryconfigure $::ToggleIPACIndex -label \"$DownMsg\""
	bind .ipaec <Unmap> \
	    "set ::IPACIsDisplayedP 0;$::CEM entryconfigure $::ToggleIPACIndex -label \"$DownMsg\""
	bind .ipaec <Map> \
	    "set ::IPACIsDisplayedP 1;$::CEM entryconfigure $::ToggleIPACIndex -label \"$UpMsg\""
	set tmp [label [format ".ipaec.r%dc%d" $manner $poa] \
	  -text [_ "IPA\nConsonants"]  -padx 5 -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList $tmp;
	lappend IPAECRowLabelList $tmp;
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $labial] \
	    -text [_ "labial"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $labiodental] \
	    -text [_ "labio\ndental"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $dental] \
	    -text [_ "dental"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $alveolar] \
	    -text [_ "alveolar"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $postalveolar] \
	    -text [_ "post\nalveolar"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $retroflex] \
	    -text [_ "retro\nflex"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $palatal] \
	    -text [_ "palatal"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $velar] \
	    -text [_ "velar"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $uvular] \
	    -text [_ "uvular"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $pharyngeal] \
	    -text [_ "pharyn\ngeal"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $epiglottal] \
	    -text [_ "epi\nglottal"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAECColumnLabelList [label [format ".ipaec.r%dc%d" $manner $glottal] \
	    -text [_ "glottal"]   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]

	#Voiceless oral stops
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $vlstop $poa] \
	       -text [_ "voiceless stops"] -anchor w  -padx $xp -pady $yp \
	       -bg $::ColorSpecs(IPAHeadings,Background)]
	set cmd {$::InsertionTarget insert insert [list p]}
	set bn [format ".ipaec.r%dc%d"  $vlstop $labial]
	button $bn  -text "p" -padx $xp -pady $yp -command  $cmd \
	    -fg $::ColorSpecs(UserTextEntry,Foreground); #debug
	balloonhelpd_for $bn [_ "voiceless labial stop"]

	label [format ".ipaec.r%dc%d" $vlstop $labiodental] \
	    -text " " -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list t\u032A]}
	set bn [format ".ipaec.r%dc%d"  $vlstop $dental]
	button $bn  -text "t\u032A"   -padx $xp -pady $yp -command  $cmd
	balloonhelpd_for $bn [_ "voiceless dental stop"]

	set cmd {$::InsertionTarget insert insert [list t]}
	set bn [format ".ipaec.r%dc%d"  $vlstop $alveolar]
	button $bn -text "t"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless alveolar stop"]

	label [format ".ipaec.r%dc%d"  $vlstop $postalveolar] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u0288]}
	set bn [format ".ipaec.r%dc%d"  $vlstop $retroflex]
	button $bn  -text \u0288   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E6D]}
	bind $bn <Enter> {ipaentry::cftr}
	bind $bn <Leave> {ipaentry::crtr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiceless retroflex stop"]

	set cmd {$::InsertionTarget insert insert [list c]}
	set bn [format ".ipaec.r%dc%d"  $vlstop $palatal]
	button $bn -text "c"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless palatal stop"]

	set cmd {$::InsertionTarget insert insert [list k]}
	set bn [format ".ipaec.r%dc%d" $vlstop $velar]
	button $bn -text "k" -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless velar stop"]

	set cmd {$::InsertionTarget insert insert [list q]}
	set bn [format ".ipaec.r%dc%d" $vlstop $uvular]
	button $bn  -text "q"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless uvular stop"]

	label [format ".ipaec.r%dc%d" $vlstop $pharyngeal] -text " "   -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $vlstop $epiglottal] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u0294]}
	set bn [format ".ipaec.r%dc%d" $vlstop $glottal]
	button $bn -text \u0294   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless glottal stop"]

	#Voiced oral stops
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $vdstop $poa] \
	       -text [_ "voiced stops"] -anchor w   -padx $xp -pady $yp \
	       -bg $::ColorSpecs(IPAHeadings,Background)]
	set cmd {$::InsertionTarget insert insert [list b]}
	set bn [format ".ipaec.r%dc%d" $vdstop $labial]
	button $bn -text "b"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced bilabial stop"]

	label [format ".ipaec.r%dc%d" $vdstop $labiodental] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list d\u032A]}
	set bn [format ".ipaec.r%dc%d" $vdstop $dental]
	button $bn -text "d\u032A"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced dental stop"]

	set cmd {$::InsertionTarget insert insert [list d]}
	set bn [format ".ipaec.r%dc%d" $vdstop $alveolar]
	button  $bn -text "d"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced alveolar stop"]

	label [format ".ipaec.r%dc%d"  $vdstop $postalveolar] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u0256]}
	set bn [format ".ipaec.r%dc%d"  $vdstop $retroflex]
	button $bn  -text \u0256   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E0D]}
	bind $bn <Enter> {ipaentry::cfdr}
	bind $bn <Leave> {ipaentry::crdr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiced retroflex stop"]

	set cmd {$::InsertionTarget insert insert [list \u025F]}
	set bn [format ".ipaec.r%dc%d" $vdstop $palatal]
	button $bn -text \u025F   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced palatal stop"]


	set cmd {$::InsertionTarget insert insert [list g]}
	set bn [format ".ipaec.r%dc%d" $vdstop $velar]
	button $bn -text "g"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced velar stop"]

	set cmd {$::InsertionTarget insert insert [list \u0262]}
	set bn [format ".ipaec.r%dc%d" $vdstop $uvular]
	button $bn -text \u0262  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced uvular stop"]

	label [format ".ipaec.r%dc%d" $vdstop $pharyngeal] -text " "  -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u02A1]}
	set bn [format ".ipaec.r%dc%d" $vdstop $epiglottal]
	button $bn -text \u02A1  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced epiglottal stop"]

	label [format ".ipaec.r%dc%d" $vdstop $glottal] -text " "  -padx $xp -pady $yp

	#Nasals
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $nasal $poa] \
	       -text [_ "nasals"] -anchor w -padx $xp -pady $yp \
	       -bg $::ColorSpecs(IPAHeadings,Background)]

	set cmd {$::InsertionTarget insert insert [list m]}
	set bn [format ".ipaec.r%dc%d" $nasal $labial]
	button $bn -text "m" -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "bilabial nasal"]

	set cmd {$::InsertionTarget insert insert [list \u0271]}
	set bn [format ".ipaec.r%dc%d" $nasal $labiodental]
	button $bn -text \u0271   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "labiodental nasal"]

	set cmd {$::InsertionTarget insert insert [list n\u032A]}
	set bn [format ".ipaec.r%dc%d" $nasal $dental]
	button $bn -text n\u032A   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "dental nasal"]

	set cmd {$::InsertionTarget insert insert [list n]}
	set bn [format ".ipaec.r%dc%d" $nasal $alveolar]
	button $bn -text "n"      -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "alveolar nasal"]

	label [format ".ipaec.r%dc%d"  $nasal $postalveolar] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u0273]}
	set bn [format ".ipaec.r%dc%d"  $nasal $retroflex]
	button $bn  -text \u0273   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E47]}
	bind $bn <Enter> {ipaentry::cfnr}
	bind $bn <Leave> {ipaentry::crnr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "retroflex nasal"]

	set cmd {$::InsertionTarget insert insert [list \u0272]}
	set bn [format ".ipaec.r%dc%d" $nasal $palatal]
	button $bn  -text \u0272    -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u00F1]}
	bind $bn <Enter> {ipaentry::cfny}
	bind $bn <Leave> {ipaentry::crny}
	bhcham $bn;
	balloonhelpd_for $bn [_ "palatal nasal"]

	set cmd {$::InsertionTarget insert insert [list \u014B]}
	set bn [format ".ipaec.r%dc%d" $nasal $velar]
	button $bn -text \u014B    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "velar nasal"]

	set cmd {$::InsertionTarget insert insert [list \u0274]}
	set bn [format ".ipaec.r%dc%d" $nasal $uvular]
	button $bn -text \u0274    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "uvular nasal"]

	label [format ".ipaec.r%dc%d" $nasal $pharyngeal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $nasal $epiglottal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $nasal $glottal] -text " "    -padx $xp -pady $yp

	#Voiceless fricatives
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $vlfric $poa] \
	       -text [_ "voiceless fricatives"] -anchor w  -padx $xp -pady $yp \
	       -bg $::ColorSpecs(IPAHeadings,Background)]

	set cmd {$::InsertionTarget insert insert [list \u03C6]}
	set bn [format ".ipaec.r%dc%d" $vlfric $labial]
	button $bn -text \u03C6  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless bilabial fricative"]

	set cmd {$::InsertionTarget insert insert [list f]}
	set bn [format ".ipaec.r%dc%d" $vlfric $labiodental]
	button $bn  -text "f"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless labiodental fricative"]

	set cmd {$::InsertionTarget insert insert [list \u03B8]}
	set bn [format ".ipaec.r%dc%d" $vlfric $dental]
	button $bn -text \u03B8   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless dental fricative"]

	set cmd {$::InsertionTarget insert insert [list "s"]}
	set bn [format ".ipaec.r%dc%d" $vlfric $alveolar]
	button $bn -text "s"  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless alveolar fricative"]

	set cmd {$::InsertionTarget insert insert [list \u0283]}
	set bn [format ".ipaec.r%dc%d" $vlfric $postalveolar];
	button $bn -text \u0283   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u0161]}
	bind $bn <Enter> {ipaentry::cfesh}
	bind $bn <Leave> {ipaentry::cresh}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiceless postalveolar fricative"]

	set cmd {$::InsertionTarget insert insert [list \u0282]}
	set bn [format ".ipaec.r%dc%d" $vlfric $retroflex];
	button $bn -text \u0282   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E63]}
	bind $bn <Enter> {ipaentry::cfeshr}
	bind $bn <Leave> {ipaentry::creshr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiceless retroflex fricative"]

	set cmd {$::InsertionTarget insert insert [list \u0255]}
	set bn [format ".ipaec.r%dc%d" $vlfric $palatal]
	button $bn -text \u0255  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless palatal fricative"]

	set cmd {$::InsertionTarget insert insert [list "x"]}
	set bn [format ".ipaec.r%dc%d" $vlfric $velar]
	button $bn -text "x" -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless velar fricative"]

	set cmd {$::InsertionTarget insert insert [list "\u03C7"]}
	set bn [format ".ipaec.r%dc%d" $vlfric $uvular]
	button $bn -text "\u03C7"    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless uvular fricative"]

	set cmd {$::InsertionTarget insert insert [list \u0127]}
	set bn [format ".ipaec.r%dc%d" $vlfric $pharyngeal]
	button $bn -text \u0127 -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless pharyngeal fricative"]

	set cmd {$::InsertionTarget insert insert [list \u029C]}
	set bn [format ".ipaec.r%dc%d" $vlfric $epiglottal]
	button $bn -text \u029C   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless epiglottal fricative"]

	set cmd {$::InsertionTarget insert insert [list "h"]}
	set bn [format ".ipaec.r%dc%d" $vlfric $glottal]
	button $bn -text "h" -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless glottal fricative"]

	#Voiced fricatives
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $vdfric $poa] \
	       -text [_ "voiced fricatives"] -anchor w \
	       -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]

	set cmd {$::InsertionTarget insert insert [list \u03B2]}
	set bn [format ".ipaec.r%dc%d" $vdfric $labial]
	button $bn -text \u03B2  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced bilabial fricative"]

	set cmd {$::InsertionTarget insert insert [list v]}
	set bn [format ".ipaec.r%dc%d" $vdfric $labiodental]
	button $bn -text "v" -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced labiodental fricative"]

	set cmd {$::InsertionTarget insert insert [list "\u00F0"]}
	set bn [format ".ipaec.r%dc%d" $vdfric $dental]
	button $bn -text "\u00F0"    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced dental fricative"]

	set cmd {$::InsertionTarget insert insert [list "z"]}
	set bn [format ".ipaec.r%dc%d" $vdfric $alveolar]
	button $bn -text "z"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced alveolar fricative"]

	set cmd {$::InsertionTarget insert insert [list \u0292]}
	set bn [format ".ipaec.r%dc%d" $vdfric $postalveolar]
	button $bn -text \u0292  -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u017E]}
	bind $bn <Enter> {ipaentry::cfzh}
	bind $bn <Leave> {ipaentry::crzh}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiced postalveolar fricative"]

	set cmd {$::InsertionTarget insert insert [list \u0290]}
	set bn [format ".ipaec.r%dc%d" $vdfric $retroflex]
	button $bn -text \u0290   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E93]}
	bind $bn <Enter> {ipaentry::cfzhr}
	bind $bn <Leave> {ipaentry::crzhr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiced retroflex fricative"]

	set cmd {$::InsertionTarget insert insert [list \u029D]}
	set bn [format ".ipaec.r%dc%d" $vdfric $palatal]
	button $bn -text  \u029D  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced palatal fricative"]

	set cmd {$::InsertionTarget insert insert [list \u0263]}
	set bn [format ".ipaec.r%dc%d" $vdfric $velar]
	button $bn -text \u0263  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced velar fricative"]

	set cmd {$::InsertionTarget insert insert [list "\u0281"]}
	set bn [format ".ipaec.r%dc%d" $vdfric $uvular]
	button $bn -text "\u0281" -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced uvular fricative"]

	set cmd {$::InsertionTarget insert insert [list \u0295]}
	set bn [format ".ipaec.r%dc%d" $vdfric $pharyngeal]
	button $bn -text \u0295 -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced pharyngeal fricative"]

	set cmd {$::InsertionTarget insert insert [list \u02A2]}
	set bn [format ".ipaec.r%dc%d" $vdfric $epiglottal] 
	button $bn -text \u02A2    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced epiglottal fricative"]

	set cmd {$::InsertionTarget insert insert [list \u0266]}
	set bn [format ".ipaec.r%dc%d" $vdfric $glottal]
	button $bn -text \u0266    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced glottal fricative"]

	#trills
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $trill $poa] \
	       -text [_ "trills"] -anchor w   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	set cmd {$::InsertionTarget insert insert [list \u0299]}
	set bn [format ".ipaec.r%dc%d" $trill $labial]
	button  $bn -text \u0299      -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "bilabial trill"]

	label [format ".ipaec.r%dc%d" $trill $labiodental] -text " "   -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $trill $dental] -text " "   -padx $xp -pady $yp
	set cmd {$::InsertionTarget insert insert [list r]}
	set bn [format ".ipaec.r%dc%d" $trill $alveolar]
	button $bn -text "r"      -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "alveolar trill"]

	label [format ".ipaec.r%dc%d"  $trill $postalveolar] -text " "   -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d"  $trill $retroflex] -text " "   -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $trill $palatal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $trill $velar] -text " "    -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u0280]}
	set bn [format ".ipaec.r%dc%d" $trill $uvular]
	button $bn -text \u0280    -padx $xp -pady $yp
	balloonhelpd_for $bn [_ "uvular trill"]

	label [format ".ipaec.r%dc%d" $trill $pharyngeal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $trill $epiglottal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $trill $glottal] -text " "    -padx $xp -pady $yp

	#taps and flaps
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $tap $poa] \
	       -text [_ "taps/flaps"] -anchor w   -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	label [format ".ipaec.r%dc%d" $tap $labial] -text " "      -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $tap $labiodental] -text " "   -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $tap $dental] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u027E]}
	set bn [format ".ipaec.r%dc%d" $tap $alveolar]
	button $bn -text \u027E -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "uvular trill"]

	label [format ".ipaec.r%dc%d"  $tap $postalveolar] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u027D]}
	set bn [format ".ipaec.r%dc%d"  $tap $retroflex];
	button $bn -text \u027D   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E5B]}
	bind $bn <Enter> {ipaentry::cfflapr}
	bind $bn <Leave> {ipaentry::crflapr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "retroflex tap"]

	label [format ".ipaec.r%dc%d" $tap $palatal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $tap $velar] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $tap $uvular] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $tap $pharyngeal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $tap $epiglottal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $tap $glottal] -text " "    -padx $xp -pady $yp

	#approximants
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $approx $poa] \
	       -text [_ "approximants"] -anchor w \
	       -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	label [format ".ipaec.r%dc%d" $approx $labial] -text " "      -padx $xp -pady $yp
	set cmd {$::InsertionTarget insert insert [list \u028B]}
	set bn [format ".ipaec.r%dc%d" $approx $labiodental]
	button $bn -text \u028B   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "labiodental approximant"]

	label [format ".ipaec.r%dc%d" $approx $dental] -text " "   -padx $xp -pady $yp
	set cmd {$::InsertionTarget insert insert [list \u0279]}
	set bn [format ".ipaec.r%dc%d" $approx $alveolar]
	button $bn -text  \u0279 -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "alveolar approximant"]

	label [format ".ipaec.r%dc%d"  $approx $postalveolar] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u027B]}
	set bn [format ".ipaec.r%dc%d"  $approx $retroflex]
	button $bn -text \u027B   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E5B]}
	bind $bn <Enter> {ipaentry::cfrr}
	bind $bn <Leave> {ipaentry::crrr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "retroflex approximant"]

	set cmd {$::InsertionTarget insert insert [list "j"]}
	set bn [format ".ipaec.r%dc%d" $approx $palatal]
	button $bn -text "j"    -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list "y"]}
	bind $bn <Enter> {ipaentry::cfyod}
	bind $bn <Leave> {ipaentry::cryod}
	bhcham $bn;
	balloonhelpd_for $bn [_ "palatal approximant"]

	set cmd {$::InsertionTarget insert insert [list \u0270]}
	set bn [format ".ipaec.r%dc%d" $approx $velar]
	button $bn  -text \u0270    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "velar approximant"]

	label [format ".ipaec.r%dc%d" $approx $uvular] -text " "  \
	    -padx $xp -pady $yp -bg $spcolor
	label [format ".ipaec.r%dc%d" $approx $pharyngeal] -text " "  \
	    -padx $xp -pady $yp -bg $spcolor

	#label [format ".ipaec.r%dc%d" $approx $epiglottal] -text " "  \
	    -padx $xp -pady $yp -bg $spcolor
	#label [format ".ipaec.r%dc%d" $approx $glottal] -text " "   \
	    -padx $xp -pady $yp -bg $spcolor

	#lateral approximants
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $latapprox $poa] \
	       -text [_ "lateral approximants"] -anchor w \
	       -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	label [format ".ipaec.r%dc%d" $latapprox $labial] -text " "      -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $latapprox $labiodental] -text " "   -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $latapprox $dental] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list "l"]}
	set bn [format ".ipaec.r%dc%d" $latapprox $alveolar]
	button $bn -text "l"    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "alveolar lateral approximant"]

	label [format ".ipaec.r%dc%d"  $latapprox $postalveolar] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u026D]}
	set bn [format ".ipaec.r%dc%d"  $latapprox $retroflex]
	button $bn  -text \u026D   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E37]}
	bind $bn <Enter> {ipaentry::cflr}
	bind $bn <Leave> {ipaentry::crlr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "retroflex lateral approximant"]

	set cmd {$::InsertionTarget insert insert [list \u028E]}
	set bn [format ".ipaec.r%dc%d" $latapprox $palatal]
	button $bn -text \u028E    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "palatal lateral approximant"]

	set cmd {$::InsertionTarget insert insert [list \u029F]}
	set bn [format ".ipaec.r%dc%d" $latapprox $velar]
	button $bn -text \u029F    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "velar lateral approximant"]

	label [format ".ipaec.r%dc%d" $latapprox $uvular] -text " "  \
	    -padx $xp -pady $yp -bg $spcolor
	#label [format ".ipaec.r%dc%d" $latapprox $pharyngeal] -text " "  \
	    -padx $xp -pady $yp -bg $spcolor
	#label [format ".ipaec.r%dc%d" $latapprox $epiglottal] -text " "  \
	    -padx $xp -pady $yp
	#label [format ".ipaec.r%dc%d" $latapprox $glottal] -text " "  \
	    -padx $xp -pady $yp

	#voiced implosives
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $vdimplosive $poa] \
	       -text [_ "voiced implosives"] -anchor w \
	       -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]

	set cmd {$::InsertionTarget insert insert [list \u0253]}
	set bn [format ".ipaec.r%dc%d" $vdimplosive $labial]
	button $bn -text \u0253 -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced bilabial implosive"]

	label [format ".ipaec.r%dc%d" $vdimplosive $labiodental] -text " " \
	    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $vdimplosive $dental] -text " "  \
	    -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u0257]}
	set bn [format ".ipaec.r%dc%d" $vdimplosive $alveolar]
	button $bn -text \u0257   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced alveolar implosive"]

	label [format ".ipaec.r%dc%d"  $vdimplosive $postalveolar] -text " " \
	    -padx $xp -pady $yp

	#Not officially recognized by IPA yet.
	set cmd {$::InsertionTarget insert insert [list \u1D91]}
	set bn [format ".ipaec.r%dc%d"  $vdimplosive $retroflex]
	button $bn -text \u1D91  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced retroflex implosive"]

	set cmd {$::InsertionTarget insert insert [list \u0284]}
	set bn [format ".ipaec.r%dc%d" $vdimplosive $palatal]
	button $bn -text \u0284  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced palatal implosive"]

	set cmd {$::InsertionTarget insert insert [list \u0260]}
	set bn [format ".ipaec.r%dc%d" $vdimplosive $velar]
	button $bn -text \u0260 -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced velar implosive"]

	set cmd {$::InsertionTarget insert insert [list \u029B]}
	set bn [format ".ipaec.r%dc%d" $vdimplosive $uvular]
	button $bn -text \u029B    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced uvular implosive"]

	label [format ".ipaec.r%dc%d" $vdimplosive $pharyngeal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $vdimplosive $epiglottal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $vdimplosive $glottal] -text " "    -padx $xp -pady $yp

	#clicks
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $click $poa] \
	       -text [_ "clicks"] -anchor w  -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]

	set cmd {$::InsertionTarget insert insert [list \u0298]}
	set bn [format ".ipaec.r%dc%d" $click $labial]
	button $bn -text \u0298      -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "labial click"]

	label [format ".ipaec.r%dc%d" $click $labiodental] -text " "   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list \u01C0]}
	set bn [format ".ipaec.r%dc%d" $click $dental]
	button $bn -text \u01C0   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "dental click"]

	set cmd {$::InsertionTarget insert insert [list \u01C2]}
	set bn [format ".ipaec.r%dc%d" $click $alveolar]
	button $bn -text \u01C2   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "alveolar click"]

	set cmd {$::InsertionTarget insert insert [list \u01C3]}
	set bn [format ".ipaec.r%dc%d"  $click $postalveolar]
	button $bn -text \u01C3 \
	    -padx $xp -pady $yp  -command $cmd
	balloonhelpd_for $bn [_ "postalveolar click"]

	label [format ".ipaec.r%dc%d"  $click $retroflex] -text " "   -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $click $palatal] -text " "  -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $click $velar] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $click $uvular] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $click $pharyngeal] -text " "    -padx $xp -pady $yp
	label [format ".ipaec.r%dc%d" $click $epiglottal] -text " "    -padx $xp -pady $yp
	set cmd {BackDelete $::InsertionTarget}
	set bn [format ".ipaec.r%dc%d" $click $glottal];
	button $bn -text [_ "Delete"]  -padx $xp -pady $yp -command $cmd -background "\#E0D0FF";
	set DeleteButton $bn;


	#Voiceless affricates
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $vlaffric $poa] \
	       -text [_ "voiceless affricates"] -anchor w -padx $xp -pady $yp \
	       -bg $::ColorSpecs(IPAHeadings,Background)]

	set cmd {$::InsertionTarget insert insert [list p\u03C6]}
	set bn [format ".ipaec.r%dc%d" $vlaffric $labial]
	button $bn -text "p\u03C6"      -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless bilabial affricate"]

	label [format ".ipaec.r%dc%d" $vlaffric $labiodental] -text "" -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list t\u03B8]}
	set bn [format ".ipaec.r%dc%d" $vlaffric $dental]
	button $bn -text "t\u03B8"   -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless dental affricate"]

	set cmd {$::InsertionTarget insert insert [list "ts"]}
	set bn [format ".ipaec.r%dc%d" $vlaffric $alveolar]
	button $bn -text "ts"    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless alveolar affricate"]

	set cmd {$::InsertionTarget insert insert [list t\u0283]}
	set bn [format ".ipaec.r%dc%d"  $vlaffric $postalveolar]
	button $bn -text "t\u0283"   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u010D]}
	bind $bn <Enter> {ipaentry::cfch}
	bind $bn <Leave> {ipaentry::crch}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiceless postalveolar affricate"]

	set cmd {$::InsertionTarget insert insert [list \u0288\u0282]}
	set bn [format ".ipaec.r%dc%d"  $vlaffric $retroflex]
	button $bn -text \u0288\u0282   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E6D\u1E63]}
	bind $bn <Enter> {ipaentry::cfchr}
	bind $bn <Leave> {ipaentry::crchr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiceless retroflex affricate"]

	set cmd {$::InsertionTarget insert insert [list t\u0255]}
	set bn [format ".ipaec.r%dc%d" $vlaffric $palatal]
	button $bn -text t\u0255 -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless palatal affricate"]

	set cmd {$::InsertionTarget insert insert [list "kx"]}
	set bn [format ".ipaec.r%dc%d" $vlaffric $velar]
	button $bn -text "kx"    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless velar affricate"]

	set cmd {$::InsertionTarget insert insert [list "q\u03C7"]}
	set bn [format ".ipaec.r%dc%d" $vlaffric $uvular]
	button $bn -text "q\u03C7"    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiceless uvular affricate"]

	label [format ".ipaec.r%dc%d" $vlaffric $pharyngeal] -text "" \
	    -padx $xp -pady $yp -bg $spcolor
	label [format ".ipaec.r%dc%d" $vlaffric $epiglottal] -text "" \
	    -padx $xp -pady $yp -bg $spcolor
	label [format ".ipaec.r%dc%d" $vlaffric $glottal] -text "" \
	    -padx $xp -pady $yp -bg $spcolor

	#Voiced affricate
	lappend IPAECRowLabelList [label [format ".ipaec.r%dc%d" $vdaffric $poa] \
	       -text [_ "voiced affricates"] -anchor w -padx $xp -pady $yp \
	       -bg $::ColorSpecs(IPAHeadings,Background)]

	set cmd {$::InsertionTarget insert insert [list b\u03B2]}
	set bn [format ".ipaec.r%dc%d" $vdaffric $labial]
	button $bn -text b\u03B2      -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced bilabial affricate"]

	label [format ".ipaec.r%dc%d" $vdaffric $labiodental] -text ""   -padx $xp -pady $yp

	set cmd {$::InsertionTarget insert insert [list "d\u00F0"]}
	set bn [format ".ipaec.r%dc%d" $vdaffric $dental]
	button $bn -text "d\u00F0"    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced dental affricate"]

	set cmd {$::InsertionTarget insert insert [list "dz"]}
	set bn [format ".ipaec.r%dc%d" $vdaffric $alveolar]
	button  $bn -text "dz"  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced alveolar affricate"]

	set cmd {$::InsertionTarget insert insert [list d\u0292]}
	set bn [format ".ipaec.r%dc%d" $vdaffric $postalveolar]
	button $bn -text d\u0292  -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u01C6]}
	bind $bn <Enter> {ipaentry::cfdj}
	bind $bn <Leave> {ipaentry::crdj}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiced postalveolar affricate"]

	set cmd {$::InsertionTarget insert insert [list \u0256\u0290]}
	set bn [format ".ipaec.r%dc%d" $vdaffric $retroflex]
	button $bn -text \u0256\u0290   -padx $xp -pady $yp -command $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u1E0D\u1E93]}
	bind $bn <Enter> {ipaentry::cfdjr}
	bind $bn <Leave> {ipaentry::crdjr}
	bhcham $bn;
	balloonhelpd_for $bn [_ "voiced retroflex affricate"]

	set cmd {$::InsertionTarget insert insert [list d\u029D]}
	set bn [format ".ipaec.r%dc%d" $vdaffric $palatal]
	button $bn -text  d\u029D  -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced palatal affricate"]

	set cmd {$::InsertionTarget insert insert [list g\u0263]}
	set bn [format ".ipaec.r%dc%d" $vdaffric $velar]
	button $bn -text g\u0263    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced velar affricate"]

	set cmd {$::InsertionTarget insert insert [list "\u0262\u0281"]}
	set bn [format ".ipaec.r%dc%d" $vdaffric $uvular]
	button $bn -text "\u0262\u0281"    -padx $xp -pady $yp -command $cmd
	balloonhelpd_for $bn [_ "voiced uvular affricate"]

	label [format ".ipaec.r%dc%d" $vdaffric $pharyngeal] -text "" \
	    -padx $xp -pady $yp -bg $spcolor
	#label [format ".ipaec.r%dc%d" $vdaffric $epiglottal] -text "" \
	    -padx $xp -pady $yp -bg $spcolor
	#label [format ".ipaec.r%dc%d" $vdaffric $glottal] -text ""   \
	    -padx $xp -pady $yp -bg $spcolor

	set cmd {$::InsertionTarget insert insert [list "\u028D"]}
	button .ipaec.r13c10 -text \u028D -padx $xp -pady $yp -command $cmd
	balloonhelpd_for .ipaec.r13c10  [_ "Voiceless labio-velar fricative"]
	set cmd {$::InsertionTarget insert insert [list "\u0265"]}
	button .ipaec.r11c11 -text \u0265 -padx $xp -pady $yp -command $cmd
	balloonhelpd_for .ipaec.r11c11  [_ "Voiced labio-palatal approximant"]
	set cmd {$::InsertionTarget insert insert [list "\u0077"]}
	button .ipaec.r11c12 -text \u0077 -padx $xp -pady $yp -command $cmd
	balloonhelpd_for .ipaec.r11c12  [_ "Voiced labio-velar approximant"]
	set cmd {$::InsertionTarget insert insert [list "\u01C1"]}
	button .ipaec.r12c11 -text \u01C1 -padx $xp -pady $yp -command $cmd
	set cmd {$::InsertionTarget insert insert [list "\u02A0"]}
	button .ipaec.r12c12 -text \u02A0 -padx $xp -pady $yp -command $cmd
	set cmd {$::InsertionTarget insert insert [list "\u026C"]}
	button .ipaec.r13c11 -text \u026C -padx $xp -pady $yp -command $cmd
	set cmd {$::InsertionTarget insert insert [list "\u026E"]}
	button .ipaec.r13c12 -text \u026E -padx $xp -pady $yp -command $cmd

	for {set row 0} {$row <= $lastrow} {incr row} {
	    set line [list];
	    for {set col 0} {$col <= $lastcolumn} {incr col} {
		set cell [format ".ipaec.r%dc%d" $row $col]
		lappend line $cell
		if {[string equal [winfo class $cell] "Label"]} {
		    balloonhelp_for $cell  [_ "Left click to insert standard characters.\nRight click to insert alternative characters."]
		}
	    }
	    eval grid $line -sticky news;
	}
	balloonhelpd_for .ipaec.r12c11  [_ "Lateral click"]
	balloonhelpd_for .ipaec.r12c12  [_ "Voiceless uvular implosive"]
	balloonhelpd_for .ipaec.r13c11  [_ "Voiceless lateral fricative"]
	balloonhelpd_for .ipaec.r13c12  [_ "Voiced lateral fricative"]
	bhsp [format ".ipaec.r%dc%d" $latapprox $uvular]
	bhsp [format ".ipaec.r%dc%d" $approx $uvular]
	bhsp [format ".ipaec.r%dc%d" $approx $pharyngeal]
	bhsp [format ".ipaec.r%dc%d" $vlaffric $pharyngeal]
	bhsp [format ".ipaec.r%dc%d" $vlaffric $epiglottal]
	bhsp [format ".ipaec.r%dc%d" $vlaffric $glottal]
	bhsp [format ".ipaec.r%dc%d" $vdaffric $pharyngeal]
	balloonhelpd_for $DeleteButton [_ "No, there isn't a glottal click  called \"Delete\".\nThis is so that you can correct mistakes while using\nthe mouse, without having to go back to the keyboard."]
	after idle {
	    update idletasks
	    raise .ipaec;
	}
	if {!$DisplayConsonantChartColumnLabelsP} {UnpackConsonantColumnLabels}
	if {!$DisplayConsonantChartRowLabelsP} {UnpackConsonantRowLabels}
    }

    set row 0;
    variable frontness $row;
    incr row;
    variable close $row;
    incr row;
    variable closeclosemid $row;
    incr row;
    variable closemid $row;
    incr row;
    variable midmid $row;
    incr row;
    variable openmid $row;
    incr row;
    variable openopenmid $row;
    incr row;
    variable open $row;
    variable lastvrow $row;

    set col 0;
    variable height $col; 
    incr col;
    variable fronturd $col;
    incr col;
    variable frontrnd $col;
    incr col;
    variable centurd $col;
    incr col;
    variable centrnd $col;
    incr col;
    variable backurd $col;
    incr col;
    variable backrnd $col;
    variable lastvcolumn $col;

    proc vry {} {
	variable close
	variable frontrnd
	eval [list [format ".ipaev.r%dc%d" $close $frontrnd] configure -text "y"];
	}
    
    proc vfy {} {
	variable close
	variable frontrnd
	eval [list [format ".ipaev.r%dc%d" $close $frontrnd] configure -text "\u00FC"];
	}

    proc vrphi {} {
	variable frontrnd;
	variable closemid
	eval [list [format ".ipaev.r%dc%d" $closemid $frontrnd] configure -text "\u00F8"];
    }
    
    proc vfphi {} {
	variable frontrnd;
	variable closemid
	eval [list [format ".ipaev.r%dc%d" $closemid $frontrnd] configure -text "o\u0308"];
    }

    proc vroe {} {
	variable frontrnd;
	variable openmid
	eval [list [format ".ipaev.r%dc%d" $openmid $frontrnd] configure -text "\u0153"];
    }

    proc vfoe {} {
	variable frontrnd;
	variable openmid
	eval [list [format ".ipaev.r%dc%d" $openmid $frontrnd] configure -text "\u0254\0308"];
    }


    proc PopupIPAEntryV {} {
	variable frontness
	variable close
	variable closeclosemid
	variable closemid
	variable openmid
	variable midmid
	variable openopenmid
	variable open
	variable lastvrow

	variable height
	variable fronturd
	variable frontrnd
	variable centurd
	variable centrnd
	variable backurd
	variable backrnd
	variable lastvcolumn

	variable IPAEVColumnLabelList
	variable IPAEVRowLabelList
	global DisplayVowelChartColumnLabelsP
	global DisplayVowelChartRowLabelsP

	set xp 2;
	set yp 3;
	toplevel .ipaev -borderwidth 3
	wm title .ipaev [_ "Vowel Symbols"]
	BindKeys .ipaev;
	set DownMsg [_ "Display IPA Vowel Chart"];
	set UpMsg   [_ "Remove IPA Vowel Chart"];
	bind .ipaev <Destroy> \
	    "set ::IPAVIsDisplayedP 0;$::CEM entryconfigure $::ToggleIPAVIndex -label \"$DownMsg\""
	bind .ipaev <Unmap> \
	    "set ::IPAVIsDisplayedP 0;$::CEM entryconfigure $::ToggleIPAVIndex -label \"$DownMsg\""
	bind .ipaev <Map> \
	    "set ::IPAVIsDisplayedP 1;$::CEM entryconfigure $::ToggleIPAVIndex -label \"$UpMsg\""

	#Column headers
	set tmp [label [format ".ipaev.r%dc%d" $height $frontness] \
	     -text [_ "IPA\nVowels"]  -padx $xp -pady $yp -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAEVColumnLabelList $tmp;
	lappend IPAEVRowLabelList $tmp;
	lappend IPAEVColumnLabelList [label [format ".ipaev.r%dc%d" $height $fronturd] \
	     -text [_ "Front\nUnrounded"]  -padx $xp -pady $yp \
     	     -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAEVColumnLabelList [label [format ".ipaev.r%dc%d" $height $frontrnd] \
	     -text [_ "Front\nRounded"]  -padx $xp -pady $yp \
  	     -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAEVColumnLabelList [label [format ".ipaev.r%dc%d" $height $centurd] \
	     -text [_ "Central\nUnrounded"]  -padx $xp -pady $yp \
	     -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAEVColumnLabelList [label [format ".ipaev.r%dc%d" $height $centrnd] \
	     -text [_ "Central\nRounded"]  -padx $xp -pady $yp \
   	     -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAEVColumnLabelList [label [format ".ipaev.r%dc%d" $height $backurd] \
	     -text [_ "Back\nUnrounded"]  -padx $xp -pady $yp \
  	     -bg $::ColorSpecs(IPAHeadings,Background)]
	lappend IPAEVColumnLabelList [label [format ".ipaev.r%dc%d" $height $backrnd] \
	     -text [_ "Back\nRounded"]  -padx $xp -pady $yp \
  	     -bg $::ColorSpecs(IPAHeadings,Background)]


	#Close vowels
	lappend IPAEVRowLabelList [label [format ".ipaev.r%dc%d" $close $frontness] \
	   -text [_ "Close"] -anchor w  -padx $xp -pady $yp \
  	   -bg $::ColorSpecs(IPAHeadings,Background)]
	set cmd {$::InsertionTarget insert insert [list i]}
	button [format ".ipaev.r%dc%d"  $close $fronturd] \
	     -text "i"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list y]}
	set bn  [format ".ipaev.r%dc%d"  $close $frontrnd]
	button $bn -text "y"   -padx $xp -pady $yp -command  $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u00FC]}
	bind $bn <Enter> {ipaentry::vfy}
	bind $bn <Leave> {ipaentry::vry}
	bhcham $bn;

	set cmd {$::InsertionTarget insert insert [list \u0268]}
	button [format ".ipaev.r%dc%d"  $close $centurd] \
	     -text "\u0268"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list \u0289]}
	button [format ".ipaev.r%dc%d"  $close $centrnd] \
	     -text "\u0289"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list \u0268]}
	button [format ".ipaev.r%dc%d"  $close $backurd] \
	     -text "\u0268"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list u]}
	button [format ".ipaev.r%dc%d"  $close $backrnd] \
	     -text "u"   -padx $xp -pady $yp -command  $cmd

	# Close close mid vowels
	lappend IPAEVRowLabelList [label [format ".ipaev.r%dc%d" $closeclosemid $frontness] \
	     -text [_ "Close Close Mid"] -anchor w  -padx $xp -pady $yp \
	  -bg $::ColorSpecs(IPAHeadings,Background)]
	set cmd {$::InsertionTarget insert insert [list I]}
	button [format ".ipaev.r%dc%d"  $closeclosemid $fronturd] \
	     -text "I"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list Y]}
	button [format ".ipaev.r%dc%d"  $closeclosemid $frontrnd] \
	     -text "Y"   -padx $xp -pady $yp -command  $cmd
	label [format ".ipaev.r%dc%d"  $closeclosemid $centurd] \
	     -text ""   -padx $xp -pady $yp
	set cmd {$::InsertionTarget insert insert [list \u028A]}
	button [format ".ipaev.r%dc%d"  $closeclosemid $centrnd] \
	     -text "\u028A"   -padx $xp -pady $yp -command  $cmd
	label [format ".ipaev.r%dc%d"  $closeclosemid $backurd] \
	     -text ""   -padx $xp -pady $yp
	label [format ".ipaev.r%dc%d"  $closeclosemid $backrnd] \
	     -text ""   -padx $xp -pady $yp

	# Close mid vowels
	lappend IPAEVRowLabelList [label [format ".ipaev.r%dc%d" $closemid $frontness] \
	     -text [_ "Close Mid"] -anchor w  -padx $xp -pady $yp \
	  -bg $::ColorSpecs(IPAHeadings,Background)]
	set cmd {$::InsertionTarget insert insert [list e]}
	button [format ".ipaev.r%dc%d"  $closemid $fronturd] \
	     -text "e"   -padx $xp -pady $yp -command  $cmd

	set cmd {$::InsertionTarget insert insert [list \u00F8]}
	set bn [format ".ipaev.r%dc%d"  $closemid $frontrnd]
	button $bn -text "\u00F8"   -padx $xp -pady $yp -command  $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list o\u0308]}
	bind $bn <Enter> {ipaentry::vfphi}
	bind $bn <Leave> {ipaentry::vrphi}
	bhcham $bn;

	set cmd {$::InsertionTarget insert insert [list \u0258]}
	button [format ".ipaev.r%dc%d"  $closemid $centurd] \
	     -text "\u0258"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list \u0275]}
	button [format ".ipaev.r%dc%d"  $closemid $centrnd] \
	     -text "\u0275"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list \u0264]}
	button [format ".ipaev.r%dc%d"  $closemid $backurd] \
	     -text "\u0264"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list o]}
	button [format ".ipaev.r%dc%d"  $closemid $backrnd] \
	     -text "o"   -padx $xp -pady $yp -command  $cmd

	# Mid mid vowels
	lappend IPAEVRowLabelList [label [format ".ipaev.r%dc%d" $midmid $frontness] \
	     -text [_ "Mid Mid"] -anchor w  -padx $xp -pady $yp \
	  -bg $::ColorSpecs(IPAHeadings,Background)]
	label [format ".ipaev.r%dc%d"  $midmid $fronturd] \
	     -text ""   -padx $xp -pady $yp
	label [format ".ipaev.r%dc%d"  $midmid $frontrnd] \
	     -text ""   -padx $xp -pady $yp
	set cmd {$::InsertionTarget insert insert [list \u0259]}
	button [format ".ipaev.r%dc%d"  $midmid $centurd] \
	     -text "\u0259"   -padx $xp -pady $yp -command  $cmd
	set cmd {BackDelete $::InsertionTarget}
	set bn [format ".ipaev.r%dc%d" $midmid $centrnd];
	button $bn -text [_ "Delete"]  \
	     -padx $xp -pady $yp -command $cmd -background "\#E0D0FF";
	set DeleteButton $bn;
	label [format ".ipaev.r%dc%d"  $midmid $backurd] -text ""   -padx $xp -pady $yp
	label [format ".ipaev.r%dc%d"  $midmid $backrnd] -text ""   -padx $xp -pady $yp

	# Open mid vowels
	lappend IPAEVRowLabelList [label [format ".ipaev.r%dc%d" $openmid $frontness] \
	     -text [_ "Open Mid"] -anchor w  -padx $xp -pady $yp \
  	     -bg $::ColorSpecs(IPAHeadings,Background)]
	set cmd {$::InsertionTarget insert insert [list \u025B]}
	button [format ".ipaev.r%dc%d"  $openmid $fronturd] \
	     -text \u025B   -padx $xp -pady $yp -command  $cmd

	set cmd {$::InsertionTarget insert insert [list \u0153]}
	set bn [format ".ipaev.r%dc%d"  $openmid $frontrnd]
	button $bn -text "\u0153"   -padx $xp -pady $yp -command  $cmd
	bind $bn <<B3>> {$::InsertionTarget insert insert [list \u0254\u0308]}
	bind $bn <Enter> {ipaentry::vfoe}
	bind $bn <Leave> {ipaentry::vroe}
	bhcham $bn;

	set cmd {$::InsertionTarget insert insert [list \u025C]}
	button [format ".ipaev.r%dc%d"  $openmid $centurd] \
	     -text "\u025C"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list \u025E]}
	button [format ".ipaev.r%dc%d"  $openmid $centrnd] \
	     -text "\u025E"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list \u028C]}
	button [format ".ipaev.r%dc%d"  $openmid $backurd] \
	     -text "\u028C"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list \u0254]}
	button [format ".ipaev.r%dc%d"  $openmid $backrnd] \
	     -text "\u0254"   -padx $xp -pady $yp -command  $cmd

	# Open open mid vowels
	lappend IPAEVRowLabelList [label [format ".ipaev.r%dc%d" $openopenmid $frontness] \
	     -text [_ "Open Open Mid"] -anchor w  -padx $xp -pady $yp \
  	     -bg $::ColorSpecs(IPAHeadings,Background)]
	set cmd {$::InsertionTarget insert insert [list \u00E6]}
	button [format ".ipaev.r%dc%d"  $openopenmid $fronturd] \
	     -text \u00E6   -padx $xp -pady $yp -command  $cmd
	label [format ".ipaev.r%dc%d"  $openopenmid $frontrnd] -text ""   -padx $xp -pady $yp
	set cmd {$::InsertionTarget insert insert [list \u0250]}
	button [format ".ipaev.r%dc%d"  $openopenmid $centurd] \
	     -text "\u0250"   -padx $xp -pady $yp -command  $cmd
	label [format ".ipaev.r%dc%d"  $openopenmid $centrnd] \
		     -text ""   -padx $xp -pady $yp
	label [format ".ipaev.r%dc%d"  $openopenmid $backurd] \
	     -text ""   -padx $xp -pady $yp
	label [format ".ipaev.r%dc%d"  $openopenmid $backrnd] \
	     -text ""   -padx $xp -pady $yp

	# Open  vowels
	lappend IPAEVRowLabelList [label [format ".ipaev.r%dc%d" $open $frontness] \
	   -text [_ "Open"] -anchor w  -padx $xp -pady $yp \
	   -bg $::ColorSpecs(IPAHeadings,Background)]
	set cmd {$::InsertionTarget insert insert [list a]}
	button [format ".ipaev.r%dc%d"  $open $fronturd] \
	     -text "a"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list \u0276]}
	button [format ".ipaev.r%dc%d"  $open $frontrnd] \
	     -text "\u0276"   -padx $xp -pady $yp -command  $cmd
	label [format ".ipaev.r%dc%d"  $open $centurd] \
	     -text ""   -padx $xp -pady $yp
	label [format ".ipaev.r%dc%d"  $open $centrnd] \
	     -text ""   -padx $xp -pady $yp
	set cmd {$::InsertionTarget insert insert [list \u0251]}
	button [format ".ipaev.r%dc%d"  $open $backurd] \
	     -text "\u0251"   -padx $xp -pady $yp -command  $cmd
	set cmd {$::InsertionTarget insert insert [list \u0252]}
	button [format ".ipaev.r%dc%d"  $open  $backrnd] \
	     -text "\u0252"   -padx $xp -pady $yp -command  $cmd

	for {set row 0} {$row <= $lastvrow} {incr row} {
	    set line [list];
	    for {set col 0} {$col <= $lastvcolumn} {incr col} {
		set cell [format ".ipaev.r%dc%d" $row $col]
		lappend line $cell
		balloonhelp_for $cell  \
		    [_ "Left click to insert standard characters.\n\
Right click to insert alternative characters."]
	    }
	    eval grid $line -sticky news;
	}
	balloonhelpd_for $DeleteButton [_ "No, there isn't a mid-mid central rounded vowel called \"Delete\".\nThis is so that you can correct mistakes while using\nthe mouse, without having to go back to the keyboard."]
	if {!$DisplayVowelChartColumnLabelsP} {UnpackVowelColumnLabels}
	if {!$DisplayVowelChartRowLabelsP} {UnpackVowelRowLabels}
    }

#Rearrange into functional groups, e.g. tone marks, vowel quality marks,
#etc. Separate IPA and non-IPA.

    proc PopupIPAEntryD {} {
	set PerRow 10;
	set BaseChar "o";
	set xp 2;
	set yp 3;
	toplevel .ipaed -borderwidth 3
	wm title .ipaed [_ "Phonetic Diacritics"]
	BindKeys .ipaed
	set DownMsg [_ "Display IPA Diacritic Chart"];
	set UpMsg   [_ "Remove IPA Diacritic Chart"];
	bind .ipaed <Destroy> \
	    "set ::IPADIsDisplayedP 0;$::CEM entryconfigure $::ToggleIPADIndex -label \"$DownMsg\""
	bind .ipaed <Unmap> \
	    "set ::IPADIsDisplayedP 0;$::CEM entryconfigure $::ToggleIPADIndex -label \"$DownMsg\""
	bind .ipaed <Map> \
	    "set ::IPADIsDisplayedP 1;$::CEM entryconfigure $::ToggleIPADIndex -label \"$UpMsg\""
	array set Diacritics [list\
				  0  {"\u02D0" "long"}\
				  1  {"\u02D1" "half long"}\
				  2  {"\u0303" "nasalized"}\
				  3  {"\u0324" "breathy voice"}\
				  4  {"\u0325" "voiceless"}\
				  5  {"\u032C" "voiced"}\
				  6  {"\u031A" "unreleased"}\
				  7  {"\u031C" "lower/open variety of vowel"}\
				  8  {"\u031D" "raised/closed variety of vowel"}\
				  9  {"\u031E" "lower/open variety of vowel"}\
				  10  {"\u031F" "advanced/fronted"}\
				  11  {"\u0320" "retracted/backed"}\
				  12  {"\u0321" "palatalized"}\
				  13  {"\u0322" "retroflex"}\
				  14  {"\u0323" "closer variety of vowel/retroflex"}\
				  15  {"\u0329" "syllabic"}\
				  16  {"\u032A" "dental"}\
				  17  {"\u032B" "labialized"}\
				  18  {"\u0330" "creaky voice"}\
				  19  {"\u0334" "velarized/pharyngealized"}\
				  20  {"\u0346" "dentolabial"}\
				  21  {"\u0347" "alveolar"}\
				  22  {"\u0348" "strong articulation"}\
				  23  {"\u0349" "weak articulation"}\
				  24  {"\u02CA" "high tone"}\
				  25  {"\u02CB" "low tone"}\
				  26  {"\u0302" "falling tone"}\
				  27  {"\u0304" "long"}\
				  28  {"\u0306" "short"}\
				  29  {"\u0308" "non-canonical backness"}\
				  30  {"\u02CC" "secondary stress"}\
				  31  {"\u02C8" "primary stress/downstep"}\
				  32  {"\u02B8" "palatalized"}\
				  33  {"\u02DE" "rhotacized"}\
				  34  {"\u0328" "nasalized"}\
				  35  {"\u034A" "denasal"}\
				  36  {"\u034B" "nasal escape"}\
				  37  {"\u034C" "velopharyngeal friction"}\
				  38  {"\u034D" "labial spreading"}\
				  39  {"\u034E" "whistled"}
				 ];

	set DiacriticCnt [llength [array names Diacritics]];
	set Rows [expr ceil(double($DiacriticCnt)/double($PerRow))]
	set Total [expr $PerRow * $Rows]
	#Generate buttons with blank labels for empty padding slots
	for {set k 0} {$k < $Total} {incr k} {
	    set row [expr $k/$PerRow]
	    set col [expr $k%$PerRow]
	    if {[info exist Diacritics($k)]} {
		button .ipaed.r${row}c${col} -text "$BaseChar[lindex $Diacritics($k) 0]"\
		    -padx $xp -pady $yp\
		    -command "\$::InsertionTarget insert insert  [lindex $Diacritics($k) 0]"
		balloonhelpd_for .ipaed.r${row}c${col} [_ [lindex $Diacritics($k) 1]]  
	    } else {
		label .ipaed.r${row}c${col} -text "";
	    }
	    set LastRow $row;
	    set LastCol $col;
	}
	#Lay the buttons and labels out in a grid.
	for {set row 0} {$row <= $LastRow} {incr row} {
	    set line [list];
	    for {set col 0} {$col <= $LastCol} {incr col} {
		set cell [format ".ipaed.r%dc%d" $row $col]
		lappend line $cell
	    }
	    eval grid $line -sticky news;
	}
    }

#Accented letters
    proc PopupIPAEntryA {} {
	set xp 3;
	set yp 3;
	toplevel .ipaea -borderwidth 4
	wm title .ipaea [_ "Accented Letters"]
	BindKeys .ipaea
	bind .ipaea <Destroy> {set ::IPAAIsDisplayedP 0}
	set DownMsg [_ "Display Accented Letter Chart"]
	set UpMsg   [_ "Remove Accented Letter Chart"]
	set AccentedLetters [list \
				 {"\u00E0" "a with grave" "\u00C0"}\
				 {"\u00E1" "a with acute" "\u00C1"}\
				 {"\u00E2" "a with circumflex" "\u00C2"}\
				 {"\u00E3" "a with tilde" "\u00C3"}\
				 {"\u00E4" "a with diaresis" "\u00C4"}\
				 {"\u00E5" "a with ring above" "\u00C5"}\
				 {"\u0101" "a with macron" "\u0100"}\
				 {"\u0103" "a with breve" "\u0102"}\
				 {"\u0105" "a with ogonek" "\u0104"}\
				 {"\u01CE" "a with caron" "A\u030C"}\
				 {"\u2C65" "a with stroke" "\u023A"}\
				 {"\u0227" "a with dot above" "\u0226"}\
				 {"\u1EA1" "a with dot below" "\u1EA0"}\
				 {"\u0201" "a with double grave" "\u0200"}\
				 {"\u0203" "a with inverted breve" "\u0202"}\
				 {"\u01DF" "a with diaresis and macron" "\u01DE"}\
				 {"\u01E1" "a with dot above and macron" "\u01E0"}\
				 {"\u01FB" "a with ring above and acute" "\u01FA"}\
				 {"\u01E3" "ash with macron" "\u01E2"}\
				 {"\u01FD" "ash with acute" "\u01FC"}\
				 {"\u0180" "b with stroke" "\u0243"}\
				 {"\u0183" "b with topbar" "\u0182"}\
				 {"\u1E03" "b with dot above" "\u1E02"}\
				 {"\u1E05" "b with dot below" "\u1E04"}\
				 {"\u1E07" "b with line below" "\u1E06"}\
				 {"\u0188" "c with hook" "\u0187"}\
				 {"\u0107" "c with acute" "\u0106"}\
				 {"\u0109" "c with circumflex" "\u0108"}\
				 {"\u010B" "c with dot above" "\u010A"}\
				 {"\u00E7" "c with cedilla" "\u00C7"}\
				 {"\u1E09" "c with cedilla and acute" "\u1E08"}\
				 {"\u010D" "c with caron" "\u010C"}\
				 {"\u023C" "c with stroke" "\u023B"}\
				 {"\u018C" "d with topbar" "\u018B"}\
				 {"\u010F" "d with caron" "\u010E"}\
				 {"\u0111" "d with stroke" "\u0110"}\
				 {"\u0221" "d with curl" "\uFFFD"}\
				 {"\u1E0B" "d with dot above" "\u1E0A"}\
				 {"\u1E0D" "d with dot below" "\u1E0C"}\
				 {"\u1E0F" "d with line below" "\u1E0E"}\
				 {"\u00E8" "e with grave" "\u00C8"}\
				 {"\u00E9" "e with acute" "\u00C9"}\
				 {"\u00EA" "e with circumflex" "\u00CA"}\
				 {"e\u0303" "e with tilde" "E\u0303"}\
				 {"\u00EB" "e with diaresis" "\u00CB"}\
				 {"\u0113" "e with macron" "\u0112"}\
				 {"\u0115" "e with breve" "\u0114"}\
				 {"\u0117" "e with dot above" "\u0116"}\
				 {"\u1EB9" "e with dot below" "\u1EB8"}\
				 {"\u0119" "e with ogonek" "\u0118"}\
				 {"\u011B" "e with caron" "\u011A"}\
				 {"\u0247" "e with stroke" "\u0246"}\
				 {"\u0229" "e with cedilla" "\u0228"}\
				 {"\u0205" "e with double grave" "\u0204"}\
				 {"\u0207" "e with inverted breve" "\u0206"}\
				 {"\u01F5" "g with acute" "\u01F4"}\
				 {"\u011D" "g with circumflex" "\u011C"}\
				 {"\u011F" "g with breve" "\u011E"}\
				 {"\u0121" "g with dot above" "\u0000"}\
				 {"\u0123" "g with cedilla" "\u0122"}\
				 {"\u01E5" "g with stroke" "\u01E4"}\
				 {"\u01E7" "g with caron" "\u01E6"}\
				 {"\u021F" "h with caron" "\u021E"}\
				 {"\u0125" "h with circumflex" "\u0124"}\
				 {"\u0127" "h with stroke" "\u0126"}\
				 {"\u1E25" "h with dot below" "\u1E24"}\
				 {"\u00EC" "i with grave" "\u00CC"}\
				 {"\u00ED" "i with acute" "\u00CD"}\
				 {"\u00EE" "i with circumflex" "\u00CE"}\
				 {"\u00EF" "i with diaresis" "\u00CF"}\
				 {"\u0129" "i with tilde" "\u0128"}\
				 {"\u012B" "i with macron" "\u012A"}\
				 {"\u012D" "i with breve" "\u012C"}\
				 {"\u012F" "i with ogonek" "\u012E"}\
				 {"\u01D0" "i with caron" "\u01CF"}\
				 {"\u0209" "i with double grave" "\u0208"}\
				 {"\u020B" "i with inverted breve" "\u020A"}\
				 {"\u1ECB" "i with dot below" "\u1ECA"}\
				 {"\u0131" "i without dot" "\u0049"}\
				 {"\u0135" "j with circumflex" "\u0134"}\
				 {"\u01F0" "j with caron" "J\u030C"}\
				 {"\u0249" "j with stroke" "J\u0248"}\
				 {"\u01E9" "k with caron" "\u01E8"}\
				 {"\u0199" "k with hook" "\u0198"}\
				 {"\u0137" "k with cedilla " "\u0136"}\
				 {"\u1E33" "k with dot below" "\u1E32"}\
				 {"\u013A" "l with acute" "\u0139"}\
				 {"\u013C" "l with cedilla" "\u013B"}\
				 {"\u013E" "l with caron" "\u013D"}\
				 {"\u0140" "l with middle dot" "\u013F"}\
				 {"\u0142" "l with stroke" "\u0141"}\
				 {"\u019A" "l with bar" "\u023D"}\
				 {"\u0234" "l with curl" "\uFFFD"}\
				 {"\u1E37" "l with dot below" "\u1E36"}\
				 {"\u1E39" "l with dot below and macron" "\u01E38"}\
				 {"\u1E3B" "l with line below" "\u1E3A"}\
				 {"\u019B" "lambda with stroke" "\uFFFD"}\
				 {"\u1E3F" "m with acute" "\u1E3E"}\
				 {"\u1E43" "m with dot below" "\u1E42"}\
				 {"\u1E41" "m with dot above" "\u1E40"}\
				 {"\u0271" "m with hook" "\uFFFD"}\
				 {"\u0235" "n with a curl" "\uFFFD"}\
				 {"\u0146" "n with cedilla" "\u0145"}\
				 {"\u0148" "n with caron" "\u0147"}\
				 {"\u01F9" "n with grave" "\u01F8"}\
				 {"\u00F1" "n with tilde" "\u00D1"}\
				 {"\u019E" "n with long right leg" "\u0220"}\
				 {"\u0144" "n with acute" "\u0143"}\
				 {"\u1E45" "n with dot above" "\u1E44"}\
				 {"\u1E47" "n with dot below" "\u1E46"}\
				 {"\u1E49" "n with line below" "\u1E48"}\
				 {"\u00F2" "o with grave" "\u00D2"}\
				 {"\u00F3" "o with acute" "\uD3"}\
				 {"\u00F4" "o with circumflex" "\u00D4"}\
				 {"\u00F5" "o with tilde" "\u00D5"}\
				 {"\u00F6" "o with diaresis" "\u00D6"}\
				 {"\u00F8" "o with stroke" "\u00D8"}\
				 {"\u014D" "o with macron" "\u014C"}\
				 {"\u014F" "o with breve" "\u014E"}\
				 {"\u0151" "o with double acute" "\u0150"}\
				 {"\u01A1" "o with horn" "\u01A0"}\
				 {"\u01EB" "o with ogonek" "\u01EA"}\
				 {"\u01ED" "o with ogonek and macron" "\u01EC"}\
				 {"\u01D2" "o with caron" "\u01D1"}\
				 {"\u022B" "o with diaresis and macron" "\u022A"}\
				 {"\u022D" "o with tilde and macron" "\u022C"}\
				 {"\u022F" "o with dot above" "\u022E"}\
				 {"\u01FF" "o with stroke and acute" "\u01FE"}\
				 {"\u020D" "o with double grave" "\u020C"}\
				 {"\u020F" "o with inverted breve" "\u020E"}\
				 {"\u0231" "o with dot above and macron" "\u0230"}\
				 {"\u1ECD" "o with dot below" "\u1ECC"}\
				 {"\u0223" "ou" "\u0222"}\
			         {"\u01A5" "p with hook" "\u01A4"}\
				 {"\u0155" "r with acute" "\u0154"}\
				 {"\u0157" "r with cedilla" "\u0156"}\
				 {"\u0159" "r with caron" "\u0158"}\
				 {"\u024D" "r with stroke" "\u024C"}\
				 {"\u0211" "r with double grave" "\u0210"}\
				 {"\u0213" "r with inverted breve" "\u0214"}\
				 {"\u1E5B" "r with dot below" "\u1E5A"}\
				 {"\u027F" "reversed r with fishhook" "\uFFFD"}\
				 {"\u015B" "s with acute" "\u015A"}\
				 {"\u015D" "s with circumflex" "\u015C"}\
				 {"\u015F" "s with cedilla" "\u015E"}\
				 {"\u0161" "s with caron" "\u0160"}\
				 {"\u0219" "s with comma below" "\u0218"}\
				 {"\u1E63" "s with dot below" "\u1E62"}\
				 {"\u021B" "t with comma below" "\u021A"}\
				 {"\u01AB" "t with palatal hook" "\uFFFD"}\
				 {"\u01AD" "t with hook" "\u01AC"}\
				 {"\u0163" "t wwith cedilla" "\u0162"}\
				 {"\u0165" "t with caron" "\u0164"}\
				 {"\u0167" "t with stroke" "\u0166"}\
				 {"\u2C66" "t with diagonal stroke" "\u023E"}\
				 {"\u00F9" "u with grave" "\u00D9"}\
				 {"\u00FA" "u with acute" "\u00DA"}\
				 {"\u00FB" "u with circumflex" "\u00DB"}\
				 {"\u00FC" "u with diaresis" "\u00DC"}\
				 {"\u01D4" "u with caron" "\u01D3"}\
				 {"\u01D6" "u with diaresis and macron" "\u01D5"}\
				 {"\u01D8" "u with diaresis and acute" "\u01D7"}\
				 {"\u01DA" "u with diaresis and caron" "\u01D9"}\
				 {"\u01DC" "u with diaresis and grave" "\u01DB"}\
				 {"\u01B0" "u with horn" "\u01AF"}\
				 {"\u0169" "u with tilde" "\u0168"}\
				 {"\u016B" "u with macron" "\u016A"}\
				 {"\u016D" "u with breve" "\u016C"}\
				 {"\u016F" "u with ring above" "\u016E"}\
				 {"\u0171" "u with double acute" "\u0170"}\
				 {"\u0173" "u with ogonek" "\u0172"}\
				 {"\u0215" "u with double grave" "\u0214"}\
				 {"\u0217" "u with inverted breve" "\u0216"}\
				 {"\u1EE5" "u with dot below" "\u1EE4"}\
				 {"\u2C74" "v with curl" "\uFFFD"}\
				 {"\u0175" "w with circumflex" "\u0174"}\
				 {"\u01B4" "y with hook" "\u01B3"}\
				 {"\u00FD" "y with acute" "\u00DD"}\
				 {"\u00FF" "y with diaresis" "\u0278"}\
				 {"\u0233" "y with macron" "\u0232"}\
				 {"\u0177" "y with circumflex" "\u0176"}\
				 {"\u024F" "y with stroke" "\u024E"}\
				 {"\u017A" "z with acute" "\u0179"}\
				 {"\u017C" "z with dot above" "\u017B"}\
				 {"\u017E" "z with caron" "\u017D"}\
				 {"\u01B6" "z with stroke" "\u01B5"}\
				 {"\u0225" "z with hook" "\u0224"}\
				 {"\u0236" "t with curl" "\uFFFD"}\
				 {"\u01EF" "zh with caron" "\u01EE"}]

	if {$::UseNoGlyphsP} {
	    lappend AccentedLetters {"\u1D8D" "x with palatal hook" "X\u0321"}\
				 {"\u1D6F" "m with middle tilde" "\uFFFD"}
	}
	
	#Sort by gloss
	set AccentedLetters [lsort -index 1 $AccentedLetters]

	set AccentedLetterCnt [llength $AccentedLetters];
	if {[info exists ::AccentedLettersPerRow]} {
	    set PerRow $::AccentedLettersPerRow
	} else {
	    set PerRow [expr int(ceil(sqrt($AccentedLetterCnt)))]
	}
	set Rows [expr ceil(double($AccentedLetterCnt)/double($PerRow))]
	set Total [expr $PerRow * $Rows]
	#Generate buttons with blank labels for empty padding slots
	for {set k 0} {$k < $Total} {incr k} {
	    set row [expr $k/$PerRow]
	    set col [expr $k%$PerRow]
	    if {$k < $AccentedLetterCnt} {
		button .ipaea.r${row}c${col} -text "[lindex [lindex $AccentedLetters $k] 0]"\
		    -padx $xp -pady $yp\
		    -command "\$::InsertionTarget insert insert  [lindex [lindex $AccentedLetters $k] 0]"
		balloonhelpd_for .ipaea.r${row}c${col} [_ [lindex [lindex $AccentedLetters $k] 1] ]  
		set cap [lindex [lindex $AccentedLetters $k] 2]
		bind .ipaea.r${row}c${col} <<B3>> "\$::InsertionTarget insert insert $cap"
	    } else {
		label .ipaea.r${row}c${col} -text "";
	    }
	}
	set LastRow $row;
	set LastCol $col;
	set LastCell .ipaea.r${row}c${col}
	destroy $LastCell
	button $LastCell -image $::LeftArrowImage -command {BackDelete $::InsertionTarget}
	balloonhelpd_for $LastCell  [_ "Delete, so that you can correct mistakes while using\nthe mouse, without having to go back to the keyboard."]
	
	#Lay the buttons and labels out in a grid.
	for {set row 0} {$row <= $LastRow} {incr row} {
	    set line [list];
	    for {set col 0} {$col <= $LastCol} {incr col} {
		set cell [format ".ipaea.r%dc%d" $row $col]
		lappend line $cell
	    }
	    eval grid $line -sticky news;
	}
    }

}

proc SetupImages {} {
    global LeftArrowImage

    set LeftArrowImage [image create photo -data {
	R0lGODlhFAAUAIAAAAAAAL+/vyH5BAEAAAEALAAAAAAUABQAAAIqjI+py43gGIAxTVrNxXVz
	54WTIopWGZ7oRq7X4o4wm2lvbX+ZjGv9/ysAADs=
    }]
}

proc BackDelete {w} {
    if {[string equal [winfo class $w] "Text"]} {
	set Insert [$w index insert]
	set FirstHalf  [$w get 1.0 $Insert-1chars]
	set SecondHalf [string trimright [$w get insert end]]
	$w delete 1.0 end;
	$w insert 1.0 $FirstHalf$SecondHalf;
    } else {
	set delind [expr [$w index insert] -1];
	$w delete $delind;
    }
}

proc PopupCharEntryByCode {} {
    toplevel .charent -borderwidth 4 -relief raised
    BindKeys .charent;
    wm title .charent "Entry By Codepoint";
    after idle {
	update idletasks
	set xmax [winfo screenwidth .charent]
	set ymax [winfo screenheight .charent]
	set x0 [expr 1 * ($xmax -[winfo reqwidth .charent])/3];
	set y0 [expr 1 * ($ymax -[winfo reqheight .charent])/3];
	wm geometry .charent "+$x0+$y0";
    }
    label .charent.title -text [_ "Insert Character by Numerical Code"]
    frame .charent.ef;
    entry .charent.ef.ent -width 8 -font MainFont -relief flat -bg \#E0EEFF;
    label .charent.ef.prefix -text "0x" -font MainFont
    pack .charent.ef.prefix -side left  -expand 1 -fill x -anchor e
    pack .charent.ef.ent    -side right -expand 1 -fill x -anchor w
    button .charent.d  -text [_ "Dismiss"] -command {destroy .charent}
    button .charent.x  -text [_ "Delete"] -command [list BackDelete $::CharByCodeInsertionTarget]
    button .charent.i  -text [_ "Insert"] -command {InsertUnicode}
    pack .charent.title -side top
    pack .charent.ef -side top
    pack .charent.d -side left -expand 1 -fill both
    pack .charent.x -side left -expand 1 -fill both
    pack .charent.i -side right -expand 1 -fill both
    focus .charent.ef.ent;
    set DownMsg [_ "Display Widget for Entering Characters by Unicode Code"];
    set UpMsg   [_ "Remove Widget for Entering Characters by Unicode Code"];
    bind .charent <Destroy> \
	"set ::CharEntryByCodeIsDisplayedP 0;$::CEM entryconfigure $::ToggleCharEntryByCodeIndex -label \"$DownMsg\""
    bind .charent <Unmap> \
	"set ::CharEntryByCodeIsDisplayedP 0;$::CEM entryconfigure $::ToggleCharEntryByCodeIndex -label \"$DownMsg\""
    bind .charent <Map> \
	"set ::CharEntryByCodeIsDisplayedP 1;$::CEM entryconfigure $::ToggleCharEntryByCodeIndex -label \"$UpMsg\""
    bind .charent.ef.ent <Destroy> {set ::InsertionTarget $::XLINP}
    bind .charent.ef.ent <Return> {InsertUnicode}
    bind .charent.ef.ent <Control-k> {.charent.ef.ent delete 0 end;break}


    set charentbh  [_ "Enter a character by its Unicode code, as a sequence of four hexadecimal digits.\nFor example, you may specify \'\u0298\', the International Phonetic Alphabet\nsymbol for the bilabial click, as 0298, and the Chinese character \'\u4E39\' \"egg\" as 4E39.\nDo not type the prefix 0x shown to the left of the entry box.\nIt is entered for you automatically. Return inserts the current character.\nControl-k erases the current character code."]
    balloonhelp_for .charent $charentbh
    balloonhelp_for .charent.title $charentbh
    balloonhelp_for .charent.ef.ent $charentbh
    balloonhelp_for .charent.i $charentbh
    balloonhelp_for .charent.d $charentbh
    balloonhelp_for .charent.x $charentbh
    set ::CharEntryByCodeIsDisplayedP 1;
}

#Get entry from popup, validate it, and insert it into the insertion target.
proc InsertUnicode {} {
    ClearMessageWindow;
    set str [.charent.ef.ent get];
    if {[string length $str] == 0} {
	ShowMessage [_ "The empty string is not a valid Unicode codepoint"]
	return ;
    }
    #Validate
    set BadValueP 0;
    #Make sure all digits are hex and that prefix is appropriate.
    if {[regexp {[[:xdigit:]]{4,4}} $str] == 0} {
	ShowMessage [_ "$str is not a well-formed hexadecimal Unicode value"]
	return;
    }
    set str [format "0x%s" $str]
    if {[scan $str "%x" num] < 1} {
	ShowMessage [format [_ "Ill-formed code %s"] $str]
	return;
    }

    #Reject illegal codepoints.
    if {$num  >  65535} {
	ShowMessage [_ "Tcl/Tk may not support codepoints outside the BMP (Plane 0)."]
    }
    set BadValueP 0;
    set BadRanges [list\
  	       0x0750 0x077F\
	       0x07C0 0x08FF\
	       0x1380 0x139F\
	       0x18B0 0x18FF\
	       0x1980 0x19DF\
	       0x1A00 0x1CFF\
	       0x1D80 0x1DFF\
	       0x2C00 0x2E7F\
	       0x2FE0 0x2FEF\
	       0x31C0 0x31EF\
	       0x9FB0 0x9FFF\
	       0xA4D0 0xABFF\
	       0xD7B0 0xD7FF\
	       0xD800 0xDBFF\
	       0xDC00 0xDFFF\
	       0xFE10 0xFE1F\
	       0x10140 0x102FF\
	       0x104B0 0x107FF\
	       0x10840 0x1CFFF\
	       0x1D200 0x1D2FF\
	       0x1D360 0x1D3FF\
	       0x1D800 0x1FFFF\
	       0x2A6E0 0x2F7FF\
	       0x2FAB0 0x2FFFF\
	       0xE0080 0xE00FF\
	       0xE01F0 0xEFFFF\
	       0xFFFFE 0xFFFFF];

    for {set k 0} {$k < [llength $BadRanges]} {incr k} {
	if { ($num >= [lindex $BadRanges [expr 2 * $k]]) &&\
		 ($num <= [lindex $BadRanges [expr (2 * $k) + 1]])} { 
	    set BadValueP 1;
	    break; 
	}
    }
    if {$BadValueP} {
	ShowMessage [format [_ "%s is not a valid Unicode codepoint"] $str]
	return;
    }
    if {$num >= 1114110} {
	ShowMessage [_ "Warning: codepoints above 0x10FFFD have not been assigned as of version 4.0."]
    }

    #Insert
    $::CharByCodeInsertionTarget insert insert [format "%c" $num]
}


proc IndicateIPACUp {} {
	$::CEM entryconfigure $::ToggleIPACIndex \
	    -label [_ "Remove IPA Consonant Chart"];
	set ::IPACIsDisplayedP 1;
}

proc IndicateIPACDown {} {
	$::CEM entryconfigure $::ToggleIPACIndex \
	    -label [_ "Display IPA Consonant Chart"];
	set ::IPACIsDisplayedP 0;
}

proc ToggleIPAC {} {
    global IPACIsDisplayedP;
    global m;

    if { $IPACIsDisplayedP == 0} {
	if {[winfo exists .ipaec]} {
	    wm deiconify .ipaec;
	} else { 
	    ipaentry::PopupIPAEntryC;
	}
	IndicateIPACUp;
    } else {
	wm iconify .ipaec;
	IndicateIPACDown;
    }
}

proc IndicateIPAVUp {} {
	$::CEM entryconfigure $::ToggleIPAVIndex \
	    -label [_ "Remove IPA Vowel Chart"];
	set ::IPAVIsDisplayedP 1;
}

proc IndicateIPAVDown {} {
	$::CEM entryconfigure $::ToggleIPAVIndex \
	    -label [_ "Display IPA Vowel Chart"];
	set ::IPAVIsDisplayedP 0;
}

proc ToggleIPAV {} {
    global IPAVIsDisplayedP;
    global m;

    if { $IPAVIsDisplayedP == 0} {
	if {[winfo exists .ipaev]} {
	    wm deiconify .ipaev;
	} else { 
	    ipaentry::PopupIPAEntryV;
	}
	$::CEM entryconfigure $::ToggleIPAVIndex -label [_ "Remove IPA Vowel Chart"];
	set IPAVIsDisplayedP 1;
    } else {
	wm iconify .ipaev;
	$::CEM entryconfigure $::ToggleIPAVIndex -label [_ "Display IPA Vowel Chart"];
	set IPAVIsDisplayedP 0;
    }
}

proc IndicateIPADUp {} {
	$::CEM entryconfigure $::ToggleIPADIndex \
	    -label [_ "Remove IPA Diacritic Chart"];
	set ::IPADIsDisplayedP 1;
}

proc IndicateIPADDown {} {
	$::CEM entryconfigure $::ToggleIPADIndex \
	    -label [_ "Display IPA Diacritic Chart"];
	set ::IPADIsDisplayedP 0;
}

proc ToggleIPAD {} {
    global IPADIsDisplayedP;
    global m;

    if { $IPADIsDisplayedP == 0} {
	if {[winfo exists .ipaed]} {
	    wm deiconify .ipaed;
	} else { 
	    ipaentry::PopupIPAEntryD;
	}
	IndicateIPADUp
    } else {
	wm iconify .ipaed;
	IndicateIPADDown;
    }
}

proc IndicateIPAAUp {} {
	$::CEM entryconfigure $::ToggleIPAAIndex \
	    -label [_ "Remove Accented Letter Chart"];
	set ::IPAAIsDisplayedP 1;
}

proc IndicateIPAADown {} {
	$::CEM entryconfigure $::ToggleIPAAIndex \
	    -label [_ "Display Accented Letter Chart"];
	set ::IPAAIsDisplayedP 0;
}

proc ToggleIPAA {} {
    global m;

    if { $::IPAAIsDisplayedP == 0} {
	if {[winfo exists .ipaea]} {
	    wm deiconify .ipaea;
	} else { 
	    ipaentry::PopupIPAEntryA;
	}
	IndicateIPAAUp;
    } else {
	wm iconify .ipaea;
	IndicateIPAADown;
    }
}


proc IndicateCharEntryByCodeUp {} {
	$::CEM entryconfigure $::ToggleCharEntryByCodeIndex \
	    -label [_ "Remove Widget for Entering Characters by Unicode Code"];
	set ::CharEntryByCodeIsDisplayedP 1;
}

proc IndicateCharEntryByCodeDown {} {
	$::CEM entryconfigure $::ToggleCharEntryByCodeIndex \
	    -label [_ "Display Widget for Entering Characters by Unicode Code"];
	set ::CharEntryByCodeIsDisplayedP 0;
}

proc ToggleCharEntryByCode {} {
    global CharEntryByCodeIsDisplayedP;
    global m;

    if { $CharEntryByCodeIsDisplayedP == 0} {
	if {[winfo exists .charent]} {
	    wm deiconify .charent;
	    raise .charent;
	} else { 
	    PopupCharEntryByCode;
	}
	IndicateCharEntryByCodeUp;
    } else {
	wm iconify .charent;
	IndicateCharEntryByCodeDown;
    }
}

proc SetAccentedLetterColumns {s} {
    if {$s > 1} {
	set ::AccentedLettersPerRow $s;
    }
}

proc SetDisplayConsonantChartColumnLabelsP {b} {
    global DisplayConsonantChartColumnLabelsP
    set DisplayConsonantChartColumnLabelsP [Boolean $b];
}

proc SetDisplayConsonantChartRowLabelsP {b} {
    global DisplayConsonantChartRowLabelsP
    set DisplayConsonantChartRowLabelsP [Boolean $b];
}

proc SetDisplayVowelChartColumnLabelsP {b} {
    global DisplayVowelChartColumnLabelsP
    set DisplayVowelChartColumnLabelsP [Boolean $b]
}

proc SetDisplayVowelChartRowLabelsP {b} {
    global DisplayVowelChartRowLabelsP
    set DisplayVowelChartRowLabelsP [Boolean $b]
}

proc ControlDisplayConsonantChartColumnLabels {} {
    if {$::DisplayConsonantChartColumnLabelsP} {
	ipaentry::PackConsonantColumnLabels;
    } else {
	ipaentry::UnpackConsonantColumnLabels;
    }
}

proc ControlDisplayConsonantChartRowLabels {} {
    if {$::DisplayConsonantChartRowLabelsP} {
	ipaentry::PackConsonantRowLabels;
    } else {
	ipaentry::UnpackConsonantRowLabels;
    }
}

proc ControlDisplayVowelChartColumnLabels {} {
    if {$::DisplayVowelChartColumnLabelsP} {
	ipaentry::PackVowelColumnLabels;
    } else {
	ipaentry::UnpackVowelColumnLabels;
    }
}

proc ControlDisplayVowelChartRowLabels {} {
    if {$::DisplayVowelChartRowLabelsP} {
	ipaentry::PackVowelRowLabels;
    } else {
	ipaentry::UnpackVowelRowLabels;
    }
}

####################################################

#File     -> LineList	LoadCustomCharacterChart
#LineList -> CDEF	DefineCustomCharacterChart
#CDEF     -> Popup      PopupSpecialPalette

proc ReadCustomCharacterChartPopup {args} {
    if {[llength $args]} {
	set dl [LoadCustomCharacterChart [lindex $args 0]]
    } else {
	set dl [LoadCustomCharacterChart];
    }
    PopupSpecialPalette [DefineCustomCharacterChart $dl]
}

proc DefineCustomCharacterChartPopup {ll} {
    PopupSpecialPalette [DefineCustomCharacterChart $ll];
}

#Returns a linelist.
proc LoadCustomCharacterChart {args} {
    if {[llength $args]} {
	set fn [lindex $args 0]
    } else {
	set fn [tk_getOpenFile -title [_ "Load Custom Character Chart"]];
	if {[string equal $fn ""]} {
	    ShowMessage [_ "File selection cancelled."];
	    return ;
	}
    }
    if { [catch {open $fn "r"} fhd ] != 0} {
	ShowMessage [format [_ "Unable to open character chart definition file %s."] $fn];
	return ;
    }
    set LineCnt 0
    while { [gets $fhd line] > 0} {
	lappend Lines $line;
	incr LineCnt
    }
    close $fhd;
    if {$LineCnt < 1} {
	ShowMessage [_ "File %s is empty" $fn]
	return "";
    }
    ShowMessage [format [_ "Loaded custom character chart definition from %s."] $fn]
    return $Lines;
}


#This procedure takes a list of lines defining a custom
#character entry chart, which may have been read from a
#standalone file or may be an instant list in an
#init file, and generates an internal character chart
#definition, which it stores. It does not actualy
#create a display
proc DefineCustomCharacterChart {LineList {fn NONE}} {
    #The first line is special. It contains meta-information:
    #the title, the desired number of buttons per row,
    #and the proposed font family and size. Only the
    #title is obligatory.
    set line [lindex $LineList 0]
    set flds [split $line "|"]
    set FieldCnt [llength $flds];
    set Title [lindex $flds 0];
    #The remaining lines contain pairs of code sequences and glosses.
    set cd [list];
    set LineList [lrange $LineList 1 end]
    foreach line $LineList {
	set f [split $line "|"];
	set gloss [lindex $f 1];
	set c [string trim [lindex $f 0] \"];
	set cf [split $c];
	set str "\{\"";
	foreach n $cf {
	    append str [format "%s" $n]
	}
	append str "\""
	append str [format " \"%s\"\}" $gloss]
	lappend cd $str;
    }
    set info [list [join $cd]]
    if {$FieldCnt > 1} {
	lappend info [lindex $flds 1];	# Columns
    }
    if {$FieldCnt > 2} {
	lappend info [lindex $flds 2];	# Font family
    }
    if {$FieldCnt > 3} {
	lappend info [lindex $flds 3];	# Font size
    }
    if {![string equal $fn NONE]} {	# File name
	lappend info $fn;
    }
    set ::SpecialCharacterPalette($Title) $info;
    $::CEM add command -label $Title -command "PopupSpecialPalette $Title";
    incr ::CustomCEMEntries;
    return $Title;
}

#Creates a popup chart from a stored definition.
set SccCnt 0;
proc PopupSpecialPalette {Title} {
    set name [format ".scc%d" $::SccCnt] 
    if {[info exists ::SpecialCharacterPalette($Title,WidgetName)]} {
	set w $::SpecialCharacterPalette($Title,WidgetName);
	if {[winfo exists $w]} {
	    wm deiconify $w;
	    raise $w;
	    return ;
	}
    }
    set cdefs [lindex $::SpecialCharacterPalette($Title) 0];
    set Items [llength $::SpecialCharacterPalette($Title)];
    if {$Items > 1} {
	set PerRow [lindex $::SpecialCharacterPalette($Title) 1];
    } else {
	set PerRow 5;
    }
    if {$Items > 2} {
	set FontFamily [lindex $::SpecialCharacterPalette($Title) 2];
    } else {
	set FontFamily $::FontInfo(CharacterEntryFont,family)
    }
    if {$Items > 3} {
	set FontSize [lindex $::SpecialCharacterPalette($Title) 3];
    } else {
	set FontSize $::FontInfo(CharacterEntryFont,size);
    }
    incr ::SccCnt;
    set xp 3;
    set yp 3;
    toplevel $name -borderwidth 4
    set ::SpecialCharacterPalette($Title,WidgetName) $name; 
    wm title $name $Title
    iwidgets::scrolledframe $name.sf -vscrollmode dynamic -hscrollmode dynamic \
	-height 180 -width 250
    set tf [$name.sf childsite]
    pack $name.sf -expand 1 -fill both
    BindKeys $name
    set msg [_ "Left click to insert the lower-case character.\nRight click to insert the upper-case character."]
    balloonhelp_for $name.sf $msg
    bind [$name.sf component vertsb] <<B3>> \
	"ScrollbarMoveBigIncrement [$name.sf component vertsb] 0.20 %x %y"
    set CdefCnt [llength $cdefs];
    set Rows [expr int(ceil(double($CdefCnt)/double($PerRow)))]
    set Total [expr $PerRow * $Rows]
    if {$CdefCnt == $Total} {
	incr Rows;
	set Total [expr $PerRow * $Rows]
    }
    set fontname [string trimleft $name .]Font
    font create $fontname -family $FontFamily -size $FontSize
    for {set k 0} {$k < $Total} {incr k} {
	set row [expr $k/$PerRow]
	set col [expr $k%$PerRow]
	if {$k < $CdefCnt} {
	    set csalt "";
	    set entry [lindex $cdefs $k];
	    set chstr [lindex $entry 0];
	    set gloss [lindex $entry 1];
	    set chstrParts [split $chstr ":"];
	    if {[llength $chstrParts] > 1} {
		set cs    [lindex $chstrParts 0];
		set csalt [lindex $chstrParts 1];
	    } else {
		set cs [lindex $chstr 0];
	    }
	    button $tf.r${row}c${col} -text $cs\
		-padx $xp -pady $yp -font $fontname\
		-command "\$::InsertionTarget insert insert $cs"
	    if {![string equal $csalt ""]} {
		bind $tf.r${row}c${col} <<B3>> "\$::InsertionTarget insert insert $csalt"
	    }
	    balloonhelpd_for $tf.r${row}c${col} $gloss;
	} else {
	    set ln $tf.r${row}c${col};
	    #Generate buttons with blank labels for empty padding slots
	    label $ln  -text "";
	}
    }
    set LastRow $row;
    set LastCol $col;
    set LastCell $tf.r${row}c${col}
    destroy $LastCell
    button $LastCell -image $::LeftArrowImage -command {BackDelete $::InsertionTarget}
    balloonhelpd_for $LastCell  [_ "Delete, so that you can correct mistakes while using\nthe mouse, without having to go back to the keyboard."]
    
    #Lay the buttons and labels out in a grid.
    for {set row 0} {$row <= $LastRow} {incr row} {
	set line [list];
	for {set col 0} {$col <= $LastCol} {incr col} {
	    set cell [format "%s.r%dc%d" $tf $row $col]
	    lappend line $cell
	}
	eval grid $line -sticky news;
    }
    return $name;
}

#End of character insertion code

namespace eval search {
    variable SearchOffset;
    variable SrchWCnt 0;
    variable GotoCharWCnt 0;
    namespace export GetSearchTarget

    proc SearchInitialize {w} {
	variable SearchOffset;
	set SearchOffset($w,f) 1.0;
	set SearchOffset($w,b) 1.0;
    }
    
    proc SearchText {w re forwardp cw} {
	variable SearchOffset;
	if {$forwardp} {
	    set df "-forward";
	    set Offset $SearchOffset($w,f);
	} else {
	    set df "-backward";
	    set Offset $SearchOffset($w,b);
	}
	set CurrentLine [lindex [split $Offset "."] 0]
	set CurrentChar [lindex [split $Offset "."] 1]
	set ind [$w search $df -regexp -count MatchLen $re $CurrentLine.$CurrentChar]
	if {$ind == ""} {
	    ShowMessage [_ "No match found"];
	    return 
	} else {
	    $w tag delete match;
	    $w tag configure "match" -background $::ColorSpecs(Match,Background);
	    $w tag add match $ind "$ind + $MatchLen chars"
	    $w see $ind
	    set MatchBegin [$w index match.first]
	    set MatchEnd   [$w index match.last-1chars];
	    set MatchBeginChar [CountChars [$w get 1.0 $MatchBegin-1chars]];
	    set MatchEndChar [expr $MatchBeginChar + [CountChars [$w get $MatchBegin $MatchEnd]]];
	    set msg [format [_ "Matched \[%s-%s\]  \[%s-%s\]"] $MatchBegin $MatchEnd $MatchBeginChar $MatchEndChar];
	    ShowMessage $msg
	    $cw.ri configure -text $msg;
	    set SearchOffset($w,f) "$ind + $MatchLen chars + 1 chars"
	    set SearchOffset($w,b) "$ind -1 chars"
	    return
	}
    }

    proc SearchResults {} {
	if {[string length [$::TXTOUT get 1.0 end]] <= 1 } {
	    ShowMessage [_ "There is no result text to search."]
	    return
	}
	PopupSearch $::TXTOUT  [_ "Search Results"];
    }

    proc SearchInput {} {
	if {[string length [$::TXTINP get 1.0 end]] <= 1 } {
	    ShowMessage [_ "There is no input text to search."]
	    return
	}
	PopupSearch $::TXTINP [_ "Search Input"];
    }

    proc SearchXLLHS {} {
	if {[string length [$::XLINP get 1.0 end]] <= 1 } {
	    ShowMessage [_ "There is no transliteration LHS to search."]
	    return
	}
	PopupSearch $::XLINP [_ "Search Transliteration LHS"];
    }

    proc SearchXLRHS {} {
	if {[string length [$::XLOUT get 1.0 end]] <= 1 } {
	    ShowMessage [_ "There is no transliteration RHS to search."]
	    return
	}
	PopupSearch $::XLOUT [_ "Search Transliteration RHS"];
    }

    proc GotoCharInput {} {
	if {[string length [$::TXTINP get 1.0 end]] <= 1 } {
	    ShowMessage [_ "There is no input text to search."]
	    return
	}
	PopupGotoChar $::TXTINP [_ "Go to Character in Input"];
    }

    proc GotoCharResults {} {
	if {[string length [$::TXTOUT get 1.0 end]] <= 1 } {
	    ShowMessage [_ "There is no result text to search."]
	    return
	}
	PopupGotoChar $::TXTOUT [_ "Go to Character in Results"];
    }

    proc GotoCharXLLHS {} {
	if {[string length [$::XLINP get 1.0 end]] <= 1 } {
	    ShowMessage [_ "There is no transliteration LHS to search."]
	    return
	}
	PopupGotoChar $::XLINP [_ "Go to Character in Transliteration LHS"];
    }

    proc GotoCharXLRHS {} {
	if {[string length [$::XLOUT get 1.0 end]] <= 1 } {
	    ShowMessage [_ "There is no transliteration RHS to search."]
	    return
	}
	PopupGotoChar $::XLOUT [_ "Go to Character in Transliteration RHS"];
    }

    proc PopupSearch {w title} {
	global MainFont;
	variable SrchWCnt;
	incr SrchWCnt;

	set pn [format ".srch%d" $SrchWCnt];
	if {[winfo exists $pn]} {
	    raise $pn
	    return
	}
	toplevel $pn -borderwidth 4 -relief raised \
	    -bg $::ColorSpecs(PopupWidgetDefault,Background);
	wm title $pn $title;
	label $pn.ri  -bg $::ColorSpecs(PopupWidgetDefault,Background);
	entry $pn.ent -width 40 -font MainFont   \
	    -bg $::ColorSpecs(PopupWidgetDefault,Background);
	button $pn.f  -text [_ "Search Forward"] \
	    -command "search::SearchForEntry $w 1 $pn"  \
	    -bg $::ColorSpecs(PopupWidgetButton,Background);
	button $pn.b  -text [_ "Search Backward"] \
	    -command "search::SearchForEntry $w 0 $pn" \
	    -bg $::ColorSpecs(PopupWidgetButton,Background);
	button $pn.d  -text [_ "Dismiss"] \
	    -command "destroy $pn" \
	    -bg $::ColorSpecs(PopupWidgetButton,Background);
	pack $pn.ri -side top
	pack $pn.ent -side top
	pack $pn.f -side right
	pack $pn.d -side left
	pack $pn.b -side right
	raise $pn;
	set AfterIdleCmd [format {update idletasks
	    set TextLeftEdge [winfo rootx %s];
	    dmsg "TextLeftEdge=$TextLeftEdge"
	    set TextRightEdge [expr $TextLeftEdge +  [winfo width %s]];
	    dmsg "TextRightEdge=$TextRightEdge"
	    set PopupWidth [winfo width %s];
	    dmsg "PopupWidth=$PopupWidth"
	    set x [expr int($TextRightEdge - ((2.0/3.0) * $PopupWidth))]
	    dmsg "x=$x"
	    set TextTopEdge  [winfo rooty %s];
	    dmsg "TextTopEdge=$TextTopEdge"
	    set TextMidHeight [expr $TextTopEdge + ([winfo height %s]/2.0)]
	    dmsg "TextMidHeight=$TextMidHeight"
	    set PopupHeight [winfo height %s];
	    dmsg "PopupHeight=$PopupHeight"
	    set y [expr int ($TextMidHeight -((2.0/3.0) * $PopupHeight))]
	    dmsg "y=$y"
	    wm geometry %s "+$x+$y"
	} $w $w $pn $w $pn $w $pn];
	after idle $AfterIdleCmd;
	BindKeys $pn;
	bind $pn <FocusIn> "SetInsertionTargets $pn.ent"
	bind $pn.ent <FocusIn> "SetInsertionTargets $pn.ent"
	bind $pn.ent <Return> "search::SearchForEntry $w 1 $pn"
    }

    proc SearchForEntry {w d pn} {
	set tgt [$pn.ent get];
	SearchText $w $tgt $d $pn;
    }

    proc PopupGotoChar {w title} {
	global MainFont;
	variable GotoCharWCnt;
	incr GotoCharWCnt;

	set pn [format ".gotochar%s" [regsub -all "\\." $w "-"]]
	if {[winfo exists $pn]} {
	    raise $pn
	    return
	}
	toplevel $pn -borderwidth 4 -relief raised \
	    -bg $::ColorSpecs(PopupWidgetDefault,Background);
	wm title $pn "";
	label $pn.title -text $title \
	    -bg $::ColorSpecs(PopupWidgetDefault,Background);
	entry $pn.ent -width 40 -font MainFont \
    	    -bg $::ColorSpecs(PopupWidgetDefault,Background);
	bind $pn <FocusIn> "SetInsertionTargets $pn.ent"
	button $pn.f  -text [_ "Go"] \
	    -command "search::GotoCharInText $w $pn" \
	    -bg $::ColorSpecs(PopupWidgetButton,Background);
	button $pn.d  -text [_ "Dismiss"] \
	    -command "destroy $pn" \
	    -bg $::ColorSpecs(PopupWidgetButton,Background);
	pack $pn.title -side top
	pack $pn.ent -side top
	pack $pn.f -side right
	pack $pn.d -side left
	raise $pn;
	set AfterIdleCmd [format {update idletasks
	    set TextLeftEdge [winfo rootx %s];
	    dmsg "TextLeftEdge=$TextLeftEdge"
	    set TextRightEdge [expr $TextLeftEdge +  [winfo width %s]];
	    dmsg "TextRightEdge=$TextRightEdge"
	    set PopupWidth [winfo width %s];
	    dmsg "PopupWidth=$PopupWidth"
	    set x [expr int($TextRightEdge - ((2.0/3.0) * $PopupWidth))]
	    dmsg "x=$x"
	    set TextTopEdge  [winfo rooty %s];
	    dmsg "TextTopEdge=$TextTopEdge"
	    set TextMidHeight [expr $TextTopEdge + ([winfo height %s]/2.0)]
	    dmsg "TextMidHeight=$TextMidHeight"
	    set PopupHeight [winfo height %s];
	    dmsg "PopupHeight=$PopupHeight"
	    set y [expr int ($TextMidHeight -((2.0/3.0) * $PopupHeight))]
	    dmsg "y=$y"
	    wm geometry %s "+$x+$y"
	} $w $w $pn $w $pn $w $pn];
	after idle $AfterIdleCmd;
	BindKeys $pn;
	bind $pn.ent <FocusIn> "SetInsertionTargets $pn.ent"
	bind $pn.ent <Return>  "search::GotoCharInText $w $pn"
    }

    #tw = text window in which to locate character
    #cw = calling window - the name of the popup from which to get the character position.
    proc GotoCharInText {tw cw} {
	#Handle both tcl line.char format indices and pure character offsets.
	set CharPos [$cw.ent get]
	$tw tag delete goto;
	$tw tag configure "goto" -background $::ColorSpecs(Match,Background);
	if {[scan $CharPos %d.%d line char] >= 2} {
	    $tw tag add goto $CharPos "$CharPos + 1 chars"
	    $tw see $CharPos;
	} else {
	    $tw tag add goto 1.0+${CharPos}chars 1.1+${CharPos}chars
	    $tw see 1.0+${CharPos}chars
	}
    }
}


proc CountChars {t} {
    return  [llength [split $t ""]];
}

proc PackMessageRegion {} {
    pack $::MSG -side left -expand 1 -fill both;
}

proc PackProgressBar {} {
    pack $::PBAR -side right -expand 1 -fill x;
}


proc RecordXLGenTime {} {
    set ::XLGenTime [clock seconds]
}

proc SetBoth {a b} {
    .cf.sbarleft  set $a $b;
    .cf.sbarright set $a $b;
}

proc YViewBoth {args} {
    set ac [llength $args];
    set a  [lindex $args 0]
    set b  [lindex $args 1]
    if {$ac == 3} {
	set c  [lindex $args 2]
	$::XLINP yview $a $b $c;
	$::XLOUT yview $a $b $c;
    } else {
	$::XLINP yview $a $b;
	$::XLOUT yview $a $b;
    }
    UpdateEntryLineNumbers;
}

proc InsertSelection {} {
    if {[catch {set sel [selection get -selection CLIPBOARD -type UTF8_STRING]}]} {
	ShowMessage [_ "The clipboard is empty."];
	return;
    } else {
	$::InsertionTarget insert insert $sel;
    }
}

# The following sets up a progress bar.

proc SetProgressbarOptions {} {
    global ColorSpecs

    option add *Progress.undoneForeground black  widgetDefault
    option add *Progress.undoneBackground $ColorSpecs(Message,Background)  widgetDefault
    option add *Progress.doneForeground   white  widgetDefault
    option add *Progress.doneBackground   \#EFAC49 widgetDefault
    option add *Progress.borderWidth      3      widgetDefault
    option add *Progress$.relief           sunken widgetDefault
}

namespace eval dkfprogress {
    namespace export Progress SetProgress

    proc Progress {w wid args} {
	uplevel 1 [list frame $w -class Progress] $args

	foreach {val} {
	    undoneForeground doneForeground
	    undoneBackground doneBackground
	} {
	    set class [string toupper [string index $val 0]\
		    ][string range $val 1 end]
	    set $val [option get $w $val $class]
	}

	set varname [namespace current]::progressPercent($w)
	frame $w.l -borderwidth 0 -background $undoneBackground -width $wid
	label $w.l.l -textvariable $varname -borderwidth 0 \
		-foreground $undoneForeground -background $undoneBackground
	$w.l configure -height [expr {int([winfo reqheight $w.l.l]+2)}]
	frame $w.l.fill -background $doneBackground
	label $w.l.fill.l -textvariable $varname -borderwidth 0 \
		-foreground $doneForeground -background $doneBackground

	bind $w.l <Configure> [namespace code [list ProgressConf $w "%w"]]

	pack $w.l -fill both -expand 1
	place $w.l.l -relx 0.5 -rely 0.5 -anchor center
	place $w.l.fill -x 0 -y 0 -relheight 1 -relwidth 0
	place $w.l.fill.l -x 0 -rely 0.5 -anchor center

	SetProgress $w 0
	return $w
    }

    proc ProgressConf {w width} {
	place conf $w.l.fill.l -x [expr {int($width/2)}]
    }

    proc ColorSequence n {
	if {$n<0} {set n 0} elseif {$n>100} {set n 100}
	set green 0; 
	set red 0;
	set blue 0;
	set green  [expr int(2.55 *(100 - $n))];
	set red [expr  int(2.55 * $n)];
	return [format "#%02x%02x%02x" $red $green $blue];
    }

    proc SetProgress {win value {range 100}} {
	if {![winfo exists $win]} {return}
	set progress [expr {int(100*$value)/int($range)}]
	set relwidth [expr {double($value)/double($range)}]
	variable progressPercent
	place conf $win.l.fill -relwidth $relwidth
	$win.l.fill configure -bg [ColorSequence $progress];
	set progressPercent($win) "${progress}%"
    }
}
namespace import dkfprogress::Progress dkfprogress::SetProgress


# Compare in the first instance on string length.
proc ReCompare {a b} {
    set lenA [string length $a];
    set lenB [string length $b];
    if {$lenA != $lenB} {
	return [expr $lenB - $lenA]
    }  else {
	return [string compare $a $b];
    }
}

set TransliterationInProgressP 0;
set ::AbortedP 0;

proc SendToOutputWindow {s} {
    $::TXTOUT insert end $s;
}

proc WriteOut {s} {
    puts -nonewline $s;
}

proc Transliterate {} {
    if { ($::XLWhat != 1) && ($::ParseChoice == 0) } {
	ShowMessage [_ "You must specify a parser in order to transliterate selectively."]
	return;
    }
    if {!$::BatchModeP} {ClearMessageWindow}
    if { ($::XLWhat == 1) && ($::ParseChoice != 0) } {
	ShowMessage [_ "Warning: you are using a parser but transliterating the entire input."]
    }
    set ::AbortP 0;
    set ::AbortedP 0;
    set ::TransliterationInProgressP 0;
    if {$::BatchModeP} {
	set map [GetTransliterationFromXL]
	set RawSrcTxt $::InputText;
    } else {
	set map [GetTransliterationFromWindows]
	set RawSrcTxt [string trim [$::TXTINP get 1.0 end]]
    }

    if {$::WhichProcedure > 0} {
	set map [list]
    } else {
	if {[llength $map] == 0} {
	    ShowMessage [_ "No transliteration is defined."]
	    return
	}
    }

    set SrcTxtLines [split $RawSrcTxt "\n"];
    set TextChars [CountChars $RawSrcTxt];
    if {$TextChars == 0} {
	ShowMessage [_ "There is no text to transliterate."]
	return
    }
    set TextLines [llength $SrcTxtLines];

    if {$::BatchModeP} {
	set Emit WriteOut;
	fconfigure stdout -encoding utf-8
    } else {
	set Emit SendToOutputWindow;
	$::TXTOUT configure -state normal
	$::TXTOUT delete 1.0 end;
	pack forget $::MSG;
	PackProgressBar
	PackMessageRegion;
    }

    #Set up output delimiters.
    set LOD "";
    set ROD "";

    switch -exact --  $::OutputDelimiterChoice {
	1 {
	    set LOD "~{";
	    set ROD "~}";
	}
	2 {
	    switch -exact -- $::OutputPairDelimiters {
		1 {set LOD "\u2329";set ROD "\u232A"}
		2 {set LOD "(";set ROD ")"}
		3 {set LOD "\[";set ROD "\]"}
		4 {set LOD "{";set ROD "}"}
		5 {set LOD "\u00AB";set ROD "\u00BB"}
		6 {set LOD "<";set ROD ">"}
		7 {set LOD "\u2018";set ROD "\u2019"}
		8 {set LOD "\u201C";set ROD "\u201D"}
	    }
	}
	3 {
	    set LOD $::OutputUnpairedDelimiter
	    set ROD $::OutputUnpairedDelimiter
	}
	4 {
	    set LOD [format "<%s>"  $::OutputXMLTag];
	    set ROD [format "</%s>" $::OutputXMLTag];
	}
	5 {
	    set LOD $::OutputPairedStringOpen;
	    set ROD $::OutputPairedStringClose;
	}
    }

    #Set up duplicate delimiters
    set LDD "";
    set RDD "";
    if {$::DuplicateSelection != 0} {
	if {$::DelimitDuplicateP} {
	    switch -exact --  $::DuplicateDelimiterChoice {
		1 {
		    set LDD "~{";
		    set RDD "~}";
		}
		2 {
		    switch -exact -- $::DupPairedDelimiter {
			1 {set LDD "\u2329";set RDD "\u232A"}
			2 {set LDD "(";set RDD ")"}
			3 {set LDD "\[";set RDD "\]"}
			4 {set LDD "{";set RDD "}"}
			5 {set LDD "\u00AB";set RDD "\u00BB"}
			6 {set LDD "<";set RDD ">"}
			7 {set LDD "\u2018";set RDD "\u2019"}
			8 {set LDD "\u201C";set RDD "\u201D"}
		    }
		}
		3 {
		    set LDD $::DuplicateUnpairedDelimiter
		    set RDD $::DuplicateUnpairedDelimiter
		}
		4 {
		    set LDD [format "<%s>"  $::DuplicateXMLTag];
		    set RDD [format "</%s>" $::DuplicateXMLTag];
		}
		5 {
		    set LDD $::DuplicatePairedStringOpen;
		    set RDD $::DuplicatePairedStringClose;
		}
	    }
	}
    }
    set ::TransliterationInProgressP 1;
    set ::StartTime [clock seconds]

    if {$::ParseChoice > 0} {
	set Parser $::Parsers($::ParseChoice);
    }
    #Parse the input into domains if required
    if {($::XLWhat > 0) && ($::ParseChoice > 0)} {
	if {$::VerboseP} {
	    ShowMessage [_ "Parsing input into domains..."]
	}
	set SrcTxtDomains [$Parser $RawSrcTxt $LDD $RDD $::PairedStringOpen $::PairedStringClose];
	set Domains [llength $SrcTxtDomains];
	set TwoPercent [expr int(double($Domains)/50.0)]
	set SrcList $SrcTxtDomains;
	set SrcCnt $Domains;
	if {$::VerboseP} {
	    ShowMessage [_ "Transliterating..."]
	}
    } else {
	set SrcList $SrcTxtLines;
	set SrcCnt $TextLines;
	set TwoPercent [expr int(double($TextLines)/50.0)]
    }
    #Even-numbered domains (including the first, number 0) are matrix.
    #Odd-numbered domains (including the second, number 1) are embedded.
    if {$TwoPercent == 0} {set TwoPercent 1} 

    #Now do the actual transliteration
    if {$::MapLongestFirstP} {
	set ResultList [TransliterateByLength $SrcList $::XLWhat $TwoPercent $SrcCnt];
    } else {
	set ResultList [TransliterateStraight $SrcList $::XLWhat $TwoPercent $SrcCnt $map];
    }

    #Now emit the results, inserting output delimiters as we go.
    set dcnt 0;
    if {$::XLWhat > 1} {
	#We're distinguishing between embedded and matrix domains.
	foreach d $ResultList {
	    if {$dcnt % 2 == 0} {
		# Matrix domain
		if {($dcnt > 0) && ($::OutputDelimiterChoice == 3)} {
		    $Emit $LOD
		}
		$Emit $d;
	    } else {
		# Embedded domain
		$Emit $LOD;		
		$Emit $d;	
		if {$::OutputDelimiterChoice != 3} {
		    $Emit $ROD
		}
	    }
	    incr dcnt;
	}
    } else {
	#We're not distinguishing between embedded and matrix domains.
	foreach d $ResultList {
	    $Emit $d;
	}
    }

    set ::ResultGenTime [clock seconds]
    if {$::AbortedP} {ClearTransliteratedText}
    set ElapsedTime [expr $::ResultGenTime - $::StartTime];
    if {$::XLWhat > 1} {
	set Q [format [_ "Rules %d   Lines: %s   Domains: %s   Input Characters %s"] \
		   $::XLRulesInWindows \
		   [DelimitNumber $TextLines] \
		   [DelimitNumber $SrcCnt] \
		   [DelimitNumber $TextChars]];
    } else {
	set Q [format [_ "Rules %d   Lines: %s   Input Characters %s"] \
		   $::XLRulesInWindows \
		   [DelimitNumber $TextLines] \
		   [DelimitNumber $TextChars]];
    }
    if {$::AbortedP} {
	set T [format [_ "  Elapsed Time: %d seconds (Aborted)"]  $ElapsedTime];
    } else {
	set T [format [_ "  Elapsed Time: %d seconds"]  $ElapsedTime];
    }
    if {$::VerboseP} {
	ShowMessage [format "%s %s" $Q $T]
    }
    if {!$::BatchModeP} {
	$::TXTOUT configure -state disabled
	pack forget $::PBAR
    }
    set ::TransliterationInProgressP 0;
    FlushJournal;
}

proc TransliterateByLength {SrcTxt What TwoPercent TextLines} {
    #Break into submaps corresponding to each length
    array set SubMaps {};
    set MaxLength 0;
    set CurrentLine 0;
    set ResultList [list];
    foreach k [array names ::XL] {
	set n [string length $k];
	if {$n > $MaxLength} {set MaxLength $n}
	lappend SubMaps($n) $k;
	lappend SubMaps($n) $::XL($k);
    }
    if {$What == 2} {set Residue 0} else {set Residue 1}
    foreach l $SrcTxt {
	if {$::AbortP} {
	    ShowMessage [_ "Transliteration Aborted"];
	    SetProgress $::PBAR 0;
	    set ::AbortedP 1;
	    return ;
	}
	if { ($What > 1)  && ($CurrentLine % 2 == $Residue) } {
	    lappend ResultList $l;
	} else {
	    set Result $l;
	    for {set k $MaxLength} {$k > 0} {incr k -1} {
		if {[info exists SubMaps($k)]} {
		    if {$::FoldCaseP} {
			set Result [string map -nocase $SubMaps($k) $Result];
		    } else {
			set Result [string map $SubMaps($k) $Result];
		    }
		}
	    }
	    if {$What == 1} {append Result "\n"}
	    lappend ResultList $Result;
	}
	incr CurrentLine;
	if {$CurrentLine % $TwoPercent == 0} {
	    set FractionCompleted [expr double($CurrentLine)/double($TextLines)];
	    set ET [expr [clock seconds] - $::StartTime];
	    set ETR [expr int(((1.0 / $FractionCompleted) - 1.0) * $ET)]
	    if {!$::BatchModeP} {
		ShowMessage [format \
		     [_ "Time Elapsed: %3d seconds. Estimated time remaining: %3d seconds."] \
			     $ET $ETR]
		SetProgress $::PBAR $CurrentLine $TextLines;
		update;
	    }
	}
    }
    return $ResultList;
}

#Even-numbered domains (including the first, number 0) are matrix.
#Odd-numbered domains (including the second, number 1) are embedded.
proc TransliterateStraight {SrcTxt What TwoPercent TextLines Map} {
    set CurrentLine 0;
    set ResultList [list];
    if {$What == 2} {set Residue 0} else {set Residue 1}
    if {$::WhichProcedure == 1} {
	set extcl [CreateExternCommandForExec]
    }
    foreach l $SrcTxt {
	if {$::AbortP} {
	    ShowMessage [_ "Transliteration Aborted"];
	    SetProgress $::PBAR 0;
	    set ::AbortedP 1;
	    return ;
	}
	if { ($What > 1)  && ($CurrentLine % 2 == $Residue) } {
	    lappend ResultList $l;
	} else {
	    if {$::WhichProcedure == 1} {
		set Result [TransliterateExternal $l $extcl]
	    } elseif {$::WhichProcedure == 2} {
		set Result [interp eval safeinterp $::PluginName [list $l]]
	    } elseif {$::FoldCaseP} {
		set Result [string map -nocase $Map $l];
	    } else {
		set Result [string map $Map $l];
	    }
	    if {$What == 1 } {append Result "\n"}
	    lappend ResultList $Result;
	}
	incr CurrentLine;
	if {$CurrentLine % $TwoPercent == 0} {
	    set FractionCompleted [expr double($CurrentLine)/double($TextLines)];
	    set ET [expr [clock seconds] - $::StartTime];
	    set ETR [expr int(((1.0 / $FractionCompleted) - 1.0) * $ET)]
	    if {!$::BatchModeP} {
		ShowMessage [format \
		  [_ "Time Elapsed: %3d seconds. Estimated time remaining: %3d seconds."] $ET $ETR]
		SetProgress $::PBAR [expr int(100.0 * $FractionCompleted)]
		update;
	    }
	}
    }
    return $ResultList;
}

proc TransliterateExternal {s cl} {
    if {[catch {open $::ExternTempFile "w"} ExternTempHandle ] != 0} {
	ShowMessage [format [_ "Unable to open temp file %s."] $ExternTempFile];
	return;
    }
    puts $ExternTempHandle $s
    close $ExternTempHandle
    catch {eval $cl} res
    return $res
}

#Creates the actual command to be executed as a child process
#from the information provided by the user, to which i/o setup
#needs to be added and the correct list format arranged.
proc CreateExternCommandForExec {} {
    set cl [list exec]
    foreach c $::ExternalCommandList {
	lappend cl $c
    }
    lappend cl "2>"
    lappend cl "/dev/null"
    lappend cl "<"
    lappend cl $::ExternTempFile
    return $cl
}

proc ExplainInsertCharacterByCode {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(InsertCharacterByCode)]} {
	if {[winfo exists $PopupList(InsertCharacterByCode)]} {
	    destroy $PopupList(InsertCharacterByCode);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "Insert Character By Code"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(InsertCharacterByCode) $po;
    AppendToTextDisplay $po [_ ""];
    AppendToTextDisplay $po [_ ""];
}

proc ExplainUnicode {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(Unicode)]} {
	if {[winfo exists $PopupList(Unicode)]} {
	    destroy $PopupList(Unicode);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "Unicode"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(Unicode) $po;
    AppendToTextDisplay $po [_ ""];
    AppendToTextDisplay $po [_ "The Unicode Standard is maintained by the "];
    AppendLinkToTextDisplay $po [_ "Unicode Consortium"] {ShowWebPage http://www.unicode.org};
    AppendToTextDisplay $po [_ "Much information is available at the Consortium web site, including the standard itself."];
}

proc ShowParameters {} {
    global FoldCaseP;
    global MapLongestFirstP;
#Doesn't do anything yet.
}

#Return a list of strings, each consisting of a matrix or embedded language string.
#The even numbered strings should all be matrix strings (possibly of zero length).
#The odd-numbered strings should all be embedded strings (again possibly of zero length).
#A text with no embedded text will be returned as a list containing a single string.
#A text with no matrix text will be returned as a list containing two strings,
#the first, of zero length, the second containing the entire text.

# FSA States
# 0 reading matrix
# 1 previous character was tilde not preceded by another tilde, before that, matrix
# 2 reading embedded
# 3 previous character was tilde not preceded by another tilde, before that, embedded

set GD "~"
set LD "\{"
set RD "\}"

#The last two arguments are dummies for compatibility with another function
proc HZParse {SrcText LDD RDD ot ct} {
    set CharCnt [string length $SrcText];
    set i 0;
    set HZCloses 0;
    set State 0;
    set StringList [list];
    set MatrixText "";
    set EmbeddedText "";
    while {$i < $CharCnt} {
	set c [string index $SrcText $i]
	incr i;
	switch -exact -- $c {
	    "~" {
		switch -exact -- $State {
		    0 {set State 1}
		    1 {
			set State 0;
			append MatrixText $c
		    }
		    2 {set State 3}
		    3 {
			append EmbeddedText $c;
		    }
		    default {GDNeverError}
		}
	    }
	    "\{"  {
		switch -exact -- $State {
		    0 {append MatrixText $c}
		    1 {
			set State 2;
		    }
		    2 {append MatrixText $c}
		    default {GDLDWithinError}
		}
	    }
	    "\}" {
		switch -exact -- $State {
		    0 {append MatrixText $c}
		    1 {GDRDWithinError}
		    2 {append MatrixText $c}
		    3 {
			incr HZCloses;
			set State 0;
			if {$::DuplicateSelection == 1} {
			    set PureEmbeddedText $EmbeddedText;
			    append MatrixText \
				[format "%s%s%s " $LDD $PureEmbeddedText $RDD];
			}
			lappend StringList $MatrixText;
			set MatrixText "";
			if {$::DuplicateSelection == 2} {
			    set PureEmbeddedText $EmbeddedText;
			    append MatrixText \
				[format " %s%s%s" $LDD $PureEmbeddedText $RDD];
			}
			lappend StringList $EmbeddedText
			set EmbeddedText "";
		    }
		    default {GDNeverError}
		}
	    }
	    "\n" {
		switch -exact -- $State {
		    0 {append MatrixText $c}
		    1 {
			set State 0;
		    }
		    3 {
			set State 2;
		    }
		    default {append EmbeddedText $c}
		}
	    }
	    default {
		switch -exact -- $State {
		    0 {append MatrixText $c}
		    2 {append EmbeddedText $c}
		    default {SingletonGDError}
		}
	    }
	}
    } 
    lappend StringList $MatrixText;
    if {$HZCloses == 0} {
	ShowMessage [_ "Warning: no embedded regions were found in the input text."]
    }
    return $StringList;
}


proc GDLDWithinError {} {
    error "HZ parser: $::GD$::LD within HZ domain";
}

proc GDRDWithinError {} {
    error "HZ parser: $::GD$::RD within HZ domain";
}

proc SingletonGDError {} {
    error "HZ parser: Invalid input: singleton $::GD"
}

proc GDNeverError {} {
    error "HZ parser: $::GD in state other than 0-3 should not occur"
}


#FSA States
# 0 reading matrix
# 1 read left delimiter - reading embedded

#The last two arguments are dummies for compatibility with another function
proc PairedSingleParse {SrcText LDD RDD ot ct} {
    set CharCnt [string length $SrcText];
    set i 0;
    set Closes 0;
    set State 0;
    set StringList [list];
    set MatrixText "";
    set EmbeddedText "";
    switch -exact -- $::PairedDelimiter {
	1 {set LD "\u2329";set RD "\u232A"}
	2 {set LD "(";set RD ")"}
	3 {set LD "\[";set RD "\]"}
	4 {set LD "{";set RD "}"}
	5 {set LD "\u00AB";set RD "\u00BB"}
	6 {set LD "<";set RD ">"}
	7 {set LD "\u2018";set RD "\u2019"}
	8 {set LD "\u201C";set RD "\u201D"}
    }

    while {$i < $CharCnt} {
	set c [string index $SrcText $i]
	incr i;
	if {$c == $LD} {
	    if {$State == 0} {
		set State 1
	    } else {
		set emsg [format [_ "Parse error at character %d: a left delimiter was\nencountered before the previous domain was terminated.\n"] $i]
		PopupErrorInfo $emsg;
		ShowMessage [format [_ "Parse error at character %d"] $i]
		if {$::StrictParseP} {
		    return "";
		} else {
		    incr ::ParseErrorCnt
		    continue ;
		}
	    }
	} elseif {$c == $RD} {
	    if {$State == 0} {
		set emsg [format [_ "Parse error at character %d: a right delimiter was\nencountered without a matching left delimiter.\n"] $i]
		PopupErrorInfo $emsg;
		ShowMessage [format [_ "Parse error at character %d"] $i]
		if {$::StrictParseP} {
		    return "";
		} else {
		    incr ::ParseErrorCnt
		    continue ;
		}
	    } else {
		incr Closes;
		set State 0;
		if {$::DuplicateSelection == 1} {
		    set PureEmbeddedText $EmbeddedText;
		    append MatrixText \
			[format "%s%s%s " $LDD $PureEmbeddedText $RDD];
		}
		lappend StringList $MatrixText;
		set MatrixText "";
		if {$::DuplicateSelection == 2} {
		    set PureEmbeddedText $EmbeddedText;
		    append MatrixText \
			[format " %s%s%s" $LDD $PureEmbeddedText $RDD];
		}
		lappend StringList $EmbeddedText
		set EmbeddedText "";
	    }
	} else {
	    if {$State == 0} {
		append MatrixText $c
	    } else {
		append EmbeddedText $c
	    }
	}
    }
    lappend StringList $MatrixText;
    if {$Closes == 0} {
	ShowMessage [_ "Warning: no embedded regions were found in the input text."]
    }
    return $StringList;
}

#The last two arguments are dummies for compatibility with another function
proc XMLParse {SrcTxt LDD RDD ot ct} {
    set OpenTag  [format "<%s>"  $::XMLTag];
    set CloseTag [format "</%s>" $::XMLTag];
    return [PairedStringParse $SrcTxt $LDD $RDD $OpenTag $CloseTag];
}

proc PairedStringParse {SrcTxt LDD RDD OpenTag CloseTag} {
    set Closes 0;
    set StringList [list];
    set MatrixText "";
    set EmbeddedText "";
    set OpenTagLength [string length $OpenTag];
    set CloseTagLength [string length $CloseTag];

    set s $SrcTxt;
    while {[string length $s] > 0} {
	set oloc [string first $OpenTag $s];
	if {$oloc < 0} {
	    append Matrix $s;
	    lappend StringList $Matrix;
	    return $StringList;
	}
	append Matrix [string range $s 0 [expr $oloc-1]];
	set s [string range $s [expr $oloc + $OpenTagLength] end];
	set cloc [string first $CloseTag $s];
	if {$cloc < 0} {
	    ShowMessage "Parse error: unterminated domain.";
	    return [list];
	}
	incr Closes;
	set Embedded [string range $s 0 [expr $cloc -1]];
	if {$::DuplicateSelection == 1} {
	    append Matrix [format "%s%s%s " $LDD $Embedded $RDD]
	}
	lappend StringList $Matrix;
	lappend StringList $Embedded;
	set Matrix "";
	if {$::DuplicateSelection == 2} {
	    append Matrix \
		[format " %s%s%s" $LDD $Embedded $RDD];
	}
	set Embedded "";
	set s [string range $s [expr $cloc + $CloseTagLength] end];
    }
    lappend StringList $Matrix
    if {$Closes == 0} {
	ShowMessage "Warning: no embedded regions were found in the input text."
    }
    return $StringList;
}

#The last two arguments are dummies for compatibility with another function
proc UnpairedSingleParse {SrcText LDD RDD ot ct} {
    set CharCnt [string length $SrcText];
    set i 0;
    set Closes 0;
    set State 0;
    set StringList [list];
    set MatrixText "";
    set EmbeddedText "";
    set Delimiter $::UnpairedDelimiter;
    while {$i < $CharCnt} {
	set c [string index $SrcText $i]
	incr i;
	if {$c == $Delimiter} {
	    if {$State == 0} {
		lappend StringList $MatrixText;
		set MatrixText "";
		set State 1;
	    } else {
		if {$::DuplicateSelection == 1} {
		    lset StringList end [append \
			     [lindex $StringList end] [format "%s%s%s " $LDD $EmbeddedText $RDD]]
		}
		lappend StringList $EmbeddedText;
		set EmbeddedText "";
		if {$::DuplicateSelection == 2} {
		    set PureEmbeddedText $EmbeddedText;
		    append MatrixText \
			[format " %s%s%s" $LDD $PureEmbeddedText $RDD];
		}
		set State 0;
		incr Closes;
	    }
	} else {
	    if {$State == 0} {
		append MatrixText $c;
	    } else {
		append EmbeddedText $c;
	    }
	}
    } 
    lappend StringList $MatrixText;
    if {$State == 1} {
	ShowMessage [_ "Warning: parse error: unpaired delimiter: domain not closed."]
    }
    if {$Closes == 0} {
	ShowMessage [_ "Warning: no embedded regions were found in the input text."]
    }
    return $StringList;
}

proc PopupParseDialog {} {

    toplevel .par
    wm title .par [_ "Parse Settings"]
    BindKeys .par

    option add *par*Entry.foreground black 100
    frame .par.top
    frame .par.top.xlw 
    label .par.top.xlw.lab -text "What Should be Transliterated?" -anchor w 
    radiobutton .par.top.xlw.rdb1 -variable XLWhat -value 1  -text [_ "Everything"]  -anchor w
    radiobutton .par.top.xlw.rdb2 -variable XLWhat -value 2  -text [_ "Embedded"]  -anchor w
    radiobutton .par.top.xlw.rdb3 -variable XLWhat -value 3 -text [_ "Matrix"]  -anchor w
    bind .par.top.xlw.lab  <<B3>> ExplainWhatShouldBeTransliterated;
    bind .par.top.xlw.rdb1 <<B3>> ExplainWhatShouldBeTransliterated;
    bind .par.top.xlw.rdb2 <<B3>> ExplainWhatShouldBeTransliterated;
    bind .par.top.xlw.rdb3 <<B3>> ExplainWhatShouldBeTransliterated;

    balloonhelpd_for .par.top.xlw.rdb1 \
	[_ "Transliterate the entire input.\nIgnore any markup into domains."]
    balloonhelpd_for .par.top.xlw.rdb2 \
	[_ "Parse the input into the segments enclosed by delimiters\n(\"embedded\") and the segments between them (\"matrix\").\nTransliterate only the segments enclosed by delimiters."]
    balloonhelpd_for .par.top.xlw.rdb3 \ [_ "Parse the input into the segments enclosed by delimiters\n(\"embedded\") and the segments between them (\"matrix\").\nTransliterate only the segments not enclosed by delimiters."]

    pack .par.top.xlw.lab -side top -expand 1 -fill both -anchor w
    pack .par.top.xlw.rdb1 -side left -expand 1 -fill both
    pack .par.top.xlw.rdb2 -side left -expand 1 -fill both
    pack .par.top.xlw.rdb3 -side left -expand 1 -fill both

    pack  .par.top.xlw -side left  -expand 1 -fill both -padx 8

    frame .par.parse  -relief ridge -border 2 
    frame .par.parse.top 
    frame .par.parse.p1 
    frame .par.parse.p2
    frame .par.parse.p3
    frame .par.parse.p4
    frame .par.parse.p5
    label .par.parse.top.lab -text [_ "How Should the Text Be Parsed?"] -anchor nw
    frame .par.parse.top.sep -width 2 -relief sunken 
    bind .par.parse.top.lab <<B3>> ExplainNeedForParsing;

    #Strict or relaxed? (meaning continue after parse errors)
    frame .par.parse.top.str
    label       .par.parse.top.str.lab  -text [_ "Be Strict?"] -anchor w
    radiobutton .par.parse.top.str.rdb0 -variable StrictParseP -value 0 -text [_ "no"]
    radiobutton .par.parse.top.str.rdb1 -variable StrictParseP -value 1 -text [_ "yes"]
    pack .par.parse.top.str.lab .par.parse.top.str.rdb0 .par.parse.top.str.rdb1 \
	-anchor w  -side left -expand 0
    
    balloonhelpd_for .par.parse.top.str.rdb0 [_ "Recover from parse errors."]
    balloonhelpd_for .par.parse.top.str.rdb1 [_ "Abort on parse error."]

    bind .par.parse.top.str.lab <<B3>> ExplainStrictParse;
    bind .par.parse.top.str.rdb0 <<B3>> ExplainStrictParse;
    bind .par.parse.top.str.rdb1 <<B3>> ExplainStrictParse;
    
    pack  .par.parse.top.lab -side left  -expand 1 -fill both -padx 8 -anchor nw
    pack  .par.parse.top.sep -side left  -expand 1 -fill y -anchor w  -padx 8
    pack  .par.parse.top.str -side right  -expand 1 -fill both -padx 8

    radiobutton .par.parse.p1.rdb -variable ParseChoice -value 1 -text "HZ"  -anchor w
    pack .par.parse.p1.rdb -side left -expand 1 -fill both
    bind .par.parse.p1.rdb <<B3>> ExplainHZEscapes;

    radiobutton .par.parse.p2.rdb -variable ParseChoice -value 2 -text [_ "Paired Character  "] \
	-anchor w
    radiobutton .par.parse.p2.rb1 -variable PairedDelimiter  -value 1 -text "\u2329 \u232A"  -anchor w
    radiobutton .par.parse.p2.rb2 -variable PairedDelimiter  -value 2 -text "( )"  -anchor w
    radiobutton .par.parse.p2.rb3 -variable PairedDelimiter  -value 3 -text "\[ \]"  -anchor w
    radiobutton .par.parse.p2.rb4 -variable PairedDelimiter  -value 4 -text "{ }"  -anchor w
    radiobutton .par.parse.p2.rb5 -variable PairedDelimiter  -value 5 -text "\u00AB \u00BB"  -anchor w
    radiobutton .par.parse.p2.rb6 -variable PairedDelimiter  -value 6 -text "< >"  -anchor w
    radiobutton .par.parse.p2.rb7 -variable PairedDelimiter  -value 7 -text "\u2018 \u2019"  -anchor w
    radiobutton .par.parse.p2.rb8 -variable PairedDelimiter  -value 8 -text "\u201C \u201D"  -anchor w

    balloonhelpd_for .par.parse.p2.rb1\
	[_ "0x2329 Left-Pointing Angle Bracket\n0x232A Right-Pointing Angle Bracket"]
    balloonhelpd_for .par.parse.p2.rb2\
	[_ "0x0028 Left Parenthesis\n0x0029 Right Parenthesis"]
    balloonhelpd_for .par.parse.p2.rb3\
	[_ "0x005B Left Square Bracket\n0x005D Right Square Bracket"]
    balloonhelpd_for .par.parse.p2.rb4\
	[_ "0x007B Left Curly Bracket\n0x007D Right Curly Bracket"]
    balloonhelpd_for .par.parse.p2.rb5\
	[_ "0x00AB Left-Pointing Double Angle Quotation Mark\n0x00BB Right-Pointing Double Angle Quotation Mark"]
    balloonhelpd_for .par.parse.p2.rb6\
	[_ "0x003C Less-Than Sign\n0x003E Greater-Than Sign"]
    balloonhelpd_for .par.parse.p2.rb7\
	[_ "0x2018 Left Single Quotation Mark\n0x2019 Right Single Quotation Mark"]
    balloonhelpd_for .par.parse.p2.rb8\
	[_ "0x201C Left Double Quotation Mark\n0x201D Right Double Quotation Mark"]

    bind .par.parse.p2 <<B3>> ExplainPairedDelimiters

    pack .par.parse.p2.rdb -side left -expand 1 -fill both -padx 0
    pack .par.parse.p2.rb1 -side left -expand 1 -fill both -padx 0
    pack .par.parse.p2.rb2 -side left -expand 1 -fill both -padx 5
    pack .par.parse.p2.rb3 -side left -expand 1 -fill both -padx 5
    pack .par.parse.p2.rb4 -side left -expand 1 -fill both -padx 5
    pack .par.parse.p2.rb5 -side left -expand 1 -fill both -padx 5
    pack .par.parse.p2.rb6 -side left -expand 1 -fill both -padx 5
    pack .par.parse.p2.rb7 -side left -expand 1 -fill both -padx 5
    pack .par.parse.p2.rb8 -side left -expand 1 -fill both -padx 5

    radiobutton .par.parse.p5.rdb -variable ParseChoice  -value 5 \
	-anchor w -text [_ "Paired String   "]
    label .par.parse.p5.olab -text [_ "Open: "] -anchor w
    entry .par.parse.p5.oent -width 6
    label .par.parse.p5.clab -text [_ "  Close: "] -anchor w
    entry .par.parse.p5.cent -width 6
    .par.parse.p5.oent insert insert $::PairedStringOpen
    .par.parse.p5.cent insert insert $::PairedStringClose
    .par.parse.p5.oent configure -state disabled
    .par.parse.p5.cent configure -state disabled
    frame .par.parse.p5.sep2 -width 3

    pack .par.parse.p5.rdb  -side left -expand 0 -fill none -padx 0
    pack .par.parse.p5.olab  -side left -expand 0 -fill none  -anchor w  -padx 0
    pack .par.parse.p5.oent  -side left -expand 1 -fill both -padx 0
    pack .par.parse.p5.clab -side left -expand 0 -fill none -anchor w -padx 0
    pack .par.parse.p5.cent -side left -expand 1 -fill both -padx 0
    pack .par.parse.p5.sep2  -side left -expand 1 -fill both

    radiobutton .par.parse.p3.rdb -variable ParseChoice -value 3 \
	-text [_ "Separator Character"]  -anchor w
    entry .par.parse.p3.ent -width 2
    pack .par.parse.p3.rdb  -side left -expand 0 -fill both
    pack .par.parse.p3.ent  -side left -expand 1 -fill none -anchor w
    .par.parse.p3.ent insert insert $::UnpairedDelimiter
    bind .par.parse.p3.ent  <Control-v> InsertSelection;
    bind .par.parse.p3.rdb <<B3>> ExplainSeparatorCharacter

    radiobutton .par.parse.p4.rdb -variable ParseChoice -value 4 \
	-text [_ "XML with tag   "]  -anchor w
    entry .par.parse.p4.ent -width 12
    pack .par.parse.p4.rdb  -side left -expand 0 -fill both
    pack .par.parse.p4.ent  -side left -expand 1 -fill none -anchor w
    .par.parse.p4.ent insert insert $::XMLTag
    bind .par.parse.p4.ent  <Control-v> InsertSelection;

    bind .par.parse.p4.rdb <<B3>> ExplainXML

    frame .par.parse.sep -height 6
    pack .par.parse.top -side top -expand 1 -fill both -pady 0
    pack .par.parse.sep -side top -expand 1 -fill both -pady 0
    pack .par.parse.p1  -side top -expand 1 -fill both
    pack .par.parse.p2  -side top -expand 1 -fill both -anchor w
    pack .par.parse.p5  -side top -expand 1 -fill both
    pack .par.parse.p3  -side top -expand 1 -fill both
    pack .par.parse.p4  -side top -expand 1 -fill both

    .par.parse.top.str.rdb0 configure -state disabled
    .par.parse.top.str.rdb1 configure -state disabled
    .par.parse.p2.rb1 configure -state disabled
    .par.parse.p2.rb2 configure -state disabled
    .par.parse.p2.rb3 configure -state disabled
    .par.parse.p2.rb4 configure -state disabled
    .par.parse.p2.rb5 configure -state disabled
    .par.parse.p2.rb6 configure -state disabled
    .par.parse.p2.rb7 configure -state disabled
    .par.parse.p2.rb8 configure -state disabled
    .par.parse.p1.rdb  configure -state disabled
    .par.parse.p2.rdb  configure -state disabled
    .par.parse.p3.rdb  configure -state disabled
    .par.parse.p4.rdb  configure -state disabled
    .par.parse.p5.rdb  configure -state disabled
    .par.parse.p5.oent configure -state disabled
    .par.parse.p5.cent configure -state disabled
    .par.parse.p3.ent configure -state disabled
    .par.parse.p4.ent configure -state disabled

    .par.parse.top.str.rdb${::StrictParseP} invoke
    .par.top.xlw.rdb${::XLWhat} invoke
    if {$::ParseChoice > 0} {
	.par.parse.p${::ParseChoice}.rdb invoke
    }

    frame .par.newdel -relief ridge -border 2 
    frame .par.newdel.top 
    frame .par.newdel.p0 
    frame .par.newdel.p1 
    frame .par.newdel.p2
    frame .par.newdel.p3
    frame .par.newdel.p4
    frame .par.newdel.p5

    label .par.newdel.top.lab -text [_ "How Should the Output be Delimited?"] -anchor nw
    frame .par.newdel.top.sep -width 2 -relief sunken 
    bind .par.newdel.top.lab <<B3>> ExplainOutputDelimiters;
    pack  .par.newdel.top.lab -side left  -expand 1 -fill both -padx 8 -anchor nw
    pack  .par.newdel.top.sep -side left  -expand 1 -fill y -anchor w  -padx 8

    radiobutton .par.newdel.p0.rdb -variable OutputDelimiterChoice -value 0 -text [_ "Not At All"] \
	-anchor w
    pack .par.newdel.p0.rdb -side left -expand 1 -fill both
    radiobutton .par.newdel.p1.rdb -variable OutputDelimiterChoice -value 1 -text "HZ"  -anchor w
    pack .par.newdel.p1.rdb -side left -expand 1 -fill both
    bind .par.newdel.p1.rdb <<B3>> ExplainHZEscapes;

    radiobutton .par.newdel.p2.rdb -variable OutputDelimiterChoice -value 2 -text [_ "Paired Character  "] \
	-anchor w
    radiobutton .par.newdel.p2.rb1 -variable OutputPairDelimiters  -value 1 -text "\u2329 \u232A" \
	-anchor w
    radiobutton .par.newdel.p2.rb2 -variable OutputPairDelimiters  -value 2 -text "( )"  -anchor w
    radiobutton .par.newdel.p2.rb3 -variable OutputPairDelimiters  -value 3 -text "\[ \]"  -anchor w
    radiobutton .par.newdel.p2.rb4 -variable OutputPairDelimiters  -value 4 -text "{ }"  -anchor w
    radiobutton .par.newdel.p2.rb5 -variable OutputPairDelimiters  -value 5 -text "\u00AB \u00BB" \
	-anchor w
    radiobutton .par.newdel.p2.rb6 -variable OutputPairDelimiters  -value 6 -text "< >"  -anchor w
    radiobutton .par.newdel.p2.rb7 -variable OutputPairDelimiters  -value 7 -text "\u2018 \u2019" \
	-anchor w
    radiobutton .par.newdel.p2.rb8 -variable OutputPairDelimiters  -value 8 -text "\u201C \u201D" \
	-anchor w

    balloonhelpd_for .par.newdel.p2.rb1\
	[_ "0x2329 Left-Pointing Angle Bracket\n0x232A Right-Pointing Angle Bracket"]
    balloonhelpd_for .par.newdel.p2.rb2\
	[_ "0x0028 Left Parenthesis\n0x0029 Right Parenthesis"]
    balloonhelpd_for .par.newdel.p2.rb3\
	[_ "0x005B Left Square Bracket\n0x005D Right Square Bracket"]
    balloonhelpd_for .par.newdel.p2.rb4\
	[_ "0x007B Left Curly Bracket\n0x007D Right Curly Bracket"]
    balloonhelpd_for .par.newdel.p2.rb5\
	[_ "0x00AB Left-Pointing Double Angle Quotation Mark\n0x00BB Right-Pointing Double Angle Quotation Mark"]
    balloonhelpd_for .par.newdel.p2.rb6\
	[_ "0x003C Less-Than Sign\n0x003E Greater-Than Sign"]
    balloonhelpd_for .par.newdel.p2.rb7\
	[_ "0x2018 Left Single Quotation Mark\n0x2019 Right Single Quotation Mark"]
    balloonhelpd_for .par.newdel.p2.rb8\
	[_ "0x201C Left Double Quotation Mark\n0x201D Right Double Quotation Mark"]

    bind .par.newdel.p2 <<B3>> ExplainPairedDelimiters

    pack .par.newdel.p2.rdb -side left -expand 1 -fill both -padx 0
    pack .par.newdel.p2.rb1 -side left -expand 1 -fill both -padx 0
    pack .par.newdel.p2.rb2 -side left -expand 1 -fill both -padx 5
    pack .par.newdel.p2.rb3 -side left -expand 1 -fill both -padx 5
    pack .par.newdel.p2.rb4 -side left -expand 1 -fill both -padx 5
    pack .par.newdel.p2.rb5 -side left -expand 1 -fill both -padx 5
    pack .par.newdel.p2.rb6 -side left -expand 1 -fill both -padx 5
    pack .par.newdel.p2.rb7 -side left -expand 1 -fill both -padx 5
    pack .par.newdel.p2.rb8 -side left -expand 1 -fill both -padx 5

    radiobutton .par.newdel.p5.rdb -variable OutputDelimiterChoice  -value 5 \
	-anchor w -text [_ "Paired String   "]
    label .par.newdel.p5.olab -text [_ "Open: "] -anchor w
    entry .par.newdel.p5.oent -width 6
    label .par.newdel.p5.clab -text [_ "  Close: "] -anchor w
    entry .par.newdel.p5.cent -width 6
    .par.newdel.p5.oent insert insert $::PairedStringOpen
    .par.newdel.p5.cent insert insert $::PairedStringClose
    .par.newdel.p5.oent configure -state disabled
    .par.newdel.p5.cent configure -state disabled
    frame .par.newdel.p5.sep2 -width 3

    pack .par.newdel.p5.rdb  -side left -expand 0 -fill none -padx 0
    pack .par.newdel.p5.olab  -side left -expand 0 -fill none  -anchor w  -padx 0
    pack .par.newdel.p5.oent  -side left -expand 1 -fill both -padx 0
    pack .par.newdel.p5.clab -side left -expand 0 -fill none -anchor w -padx 0
    pack .par.newdel.p5.cent -side left -expand 1 -fill both -padx 0
    pack .par.newdel.p5.sep2  -side left -expand 1 -fill both

    radiobutton .par.newdel.p3.rdb -variable OutputDelimiterChoice -value 3 \
	-text [_ "Separator Character"]  -anchor w
    entry .par.newdel.p3.ent -width 2
    pack .par.newdel.p3.rdb  -side left -expand 0 -fill both
    pack .par.newdel.p3.ent  -side left -expand 1 -fill none -anchor w
    .par.newdel.p3.ent insert insert $::UnpairedDelimiter
    bind .par.newdel.p3.ent  <Control-v> InsertSelection;
    bind .par.newdel.p3.rdb <<B3>> ExplainSeparatorCharacter

    radiobutton .par.newdel.p4.rdb -variable OutputDelimiterChoice -value 4 \
	-text [_ "XML with tag   "]  -anchor w
    entry .par.newdel.p4.ent -width 12
    pack .par.newdel.p4.rdb  -side left -expand 0 -fill both
    pack .par.newdel.p4.ent  -side left -expand 1 -fill none -anchor w
    .par.newdel.p4.ent insert insert $::XMLTag
    bind .par.newdel.p4.ent  <Control-v> InsertSelection;

    bind .par.newdel.p4.rdb <<B3>> ExplainXML

    frame .par.newdel.sep -height 6
    pack .par.newdel.top -side top -expand 1 -fill both -pady 0
    pack .par.newdel.sep -side top -expand 1 -fill both -pady 0
    pack .par.newdel.p0  -side top -expand 1 -fill both -pady 0
    pack .par.newdel.p1  -side top -expand 1 -fill both
    pack .par.newdel.p2  -side top -expand 1 -fill both -anchor w
    pack .par.newdel.p5  -side top -expand 1 -fill both
    pack .par.newdel.p3  -side top -expand 1 -fill both
    pack .par.newdel.p4  -side top -expand 1 -fill both

    frame .par.cont
    button .par.cont.d  -text [_ "Dismiss"] -command ToggleParseDialog \
	-bg $::ColorSpecs(Button,bg)
    button .par.cont.s  -text [_ "Set"] -command {SetDelims}  -bg $::ColorSpecs(Button,bg)
    pack .par.cont.d .par.cont.s -side left -expand 1 -fill both -pady 2 -padx 25
    bind .par <Destroy> "set ::ParseDialogIsDisplayedP 0"
    

    balloonhelpd_for .par.cont.s [_ "Store the unpaired delimiter, XML tag, or paired\nstring without dismissing the dialog. (All other\nsettings are stored as soon as they are made.)"]
    balloonhelpd_for .par.cont.d [_ "Dismiss the dialog without storing the new\nunpaired delimiter,  XML tag, or paired string. (All other\nsettings are stored as soon as they are made.)"]

    frame .par.sep1  -height 3 -borderwidth 1 -relief ridge
    frame .par.sep2  -height 3 -borderwidth 1 -relief ridge
    frame .par.sep3  -height 3 -borderwidth 1 -relief ridge

    pack .par.top     -side top -expand 1 -fill both  -pady 5 -padx 3
    pack .par.sep1    -side top -expand 1 -fill x     -pady 3
    pack .par.parse   -side top -expand 1 -fill both  -pady 5 -padx 3
    pack .par.newdel   -side top -expand 1 -fill both  -pady 5 -padx 3
    pack .par.sep2    -side top -expand 1 -fill x     -pady 3
    pack .par.cont    -side top -expand 1 -fill both  -pady 5 -padx 3

    trace add variable ::XLWhat  write ToggleRetainDelimitersAndParseActive
    trace add variable ::ParseChoice  write ControlParseChoiceActive
    trace add variable ::OutputDelimiterChoice  write ControlOutputDelimiterChoiceActive

    .par.newdel.p2.rb1 configure -state disabled
    .par.newdel.p2.rb2 configure -state disabled
    .par.newdel.p2.rb3 configure -state disabled
    .par.newdel.p2.rb4 configure -state disabled
    .par.newdel.p2.rb5 configure -state disabled
    .par.newdel.p2.rb6 configure -state disabled
    .par.newdel.p2.rb7 configure -state disabled
    .par.newdel.p2.rb8 configure -state disabled
    .par.newdel.p1.rdb  configure -state disabled
    .par.newdel.p2.rdb  configure -state disabled
    .par.newdel.p3.rdb  configure -state disabled
    .par.newdel.p4.rdb  configure -state disabled
    .par.newdel.p5.rdb  configure -state disabled
    .par.newdel.p5.oent configure -state disabled
    .par.newdel.p5.cent configure -state disabled
    .par.newdel.p3.ent configure -state disabled
    .par.newdel.p4.ent configure -state disabled

    .par.top.xlw.rdb${::XLWhat} invoke
    if {$::OutputDelimiterChoice > 0} {
	.par.newdel.p${::OutputDelimiterChoice}.rdb invoke
    }
    ControlParseChoiceActive x y z
    ControlOutputDelimiterChoiceActive x y z
}

#This handles both the input and output delimiters
proc SetDelims {} {
    if {[winfo exists .par] == 0} return;
    set ::XMLTag [.par.parse.p4.ent get]
    set ::UnpairedDelimiter  [.par.parse.p3.ent get]
    set ::PairedStringOpen   [.par.parse.p5.oent get]
    set ::PairedStringClose  [.par.parse.p5.cent get]
    set ::OutputXMLTag [.par.newdel.p4.ent get]
    set ::OutputUnpairedDelimiter  [.par.newdel.p3.ent get]
    set ::OutputPairedStringOpen   [.par.newdel.p5.oent get]
    set ::OutputPairedStringClose  [.par.newdel.p5.cent get]
}

proc ToggleRetainDelimitersAndParseActive {n e o} {
    if {[winfo exists .par] == 0} return;
    if {$::XLWhat == 1} {
	.par.parse.top.str.rdb0  configure -state disabled
	.par.parse.top.str.rdb1  configure -state disabled
	.par.parse.p1.rdb configure -state disabled
	.par.parse.p2.rdb configure -state disabled
	.par.parse.p3.rdb configure -state disabled
	.par.parse.p4.rdb configure -state disabled
	.par.parse.p5.rdb configure -state disabled
	.par.parse.p2.rb1 configure -state disabled
	.par.parse.p2.rb2 configure -state disabled
	.par.parse.p2.rb3 configure -state disabled
	.par.parse.p2.rb4 configure -state disabled
	.par.parse.p3.ent configure -state disabled
	.par.parse.p4.ent configure -state disabled
	.par.parse.p5.rdb  configure -state   disabled
	.par.parse.p5.oent configure -state   disabled
	.par.parse.p5.cent configure -state   disabled
	.par.newdel.p1.rdb configure -state disabled
	.par.newdel.p2.rdb configure -state disabled
	.par.newdel.p3.rdb configure -state disabled
	.par.newdel.p4.rdb configure -state disabled
	.par.newdel.p5.rdb configure -state disabled
    } else {
	.par.parse.top.str.rdb0  configure -state normal
	.par.parse.top.str.rdb1  configure -state normal
	.par.parse.p1.rdb configure -state normal
	.par.parse.p2.rdb configure -state normal
	.par.parse.p3.rdb configure -state normal
	.par.parse.p4.rdb configure -state normal
	.par.parse.p5.rdb configure -state normal
	.par.newdel.p1.rdb configure -state normal
	.par.newdel.p2.rdb configure -state normal
	.par.newdel.p3.rdb configure -state normal
	.par.newdel.p4.rdb configure -state normal
	.par.newdel.p5.rdb configure -state normal
    }
}

proc ControlParseChoiceActive {n e o} {
#Disable everything, then enable whatever is appropriate. 
    .par.parse.p2.rb1 configure -state disabled
    .par.parse.p2.rb2 configure -state  disabled
    .par.parse.p2.rb3 configure -state  disabled
    .par.parse.p2.rb4 configure -state  disabled
    .par.parse.p2.rb5 configure -state  disabled
    .par.parse.p2.rb6 configure -state  disabled
    .par.parse.p2.rb7 configure -state  disabled
    .par.parse.p2.rb8 configure -state  disabled
    .par.parse.p3.ent configure -state  disabled
    .par.parse.p4.ent configure -state disabled
    .par.parse.p5.oent configure -state   disabled
    .par.parse.p5.cent configure -state   disabled
    switch $::ParseChoice {
	2 {
	    .par.parse.p2.rb1 configure -state normal
	    .par.parse.p2.rb2 configure -state normal
	    .par.parse.p2.rb3 configure -state normal
	    .par.parse.p2.rb4 configure -state normal
	    .par.parse.p2.rb5 configure -state normal
	    .par.parse.p2.rb6 configure -state normal
	    .par.parse.p2.rb7 configure -state normal
	    .par.parse.p2.rb8 configure -state normal
	}
	3 {
	    .par.parse.p3.ent configure -state normal
	}
	4 {
	    .par.parse.p4.ent configure -state normal
	}
	5 {
	    .par.parse.p5.oent  configure -state normal
	    .par.parse.p5.cent configure -state normal
	}
    }
}

proc ControlOutputDelimiterChoiceActive {n e o} {
#Disable everything, then enable whatever is appropriate. 
    .par.newdel.p2.rb1 configure -state disabled
    .par.newdel.p2.rb2 configure -state  disabled
    .par.newdel.p2.rb3 configure -state  disabled
    .par.newdel.p2.rb4 configure -state  disabled
    .par.newdel.p2.rb5 configure -state  disabled
    .par.newdel.p2.rb6 configure -state  disabled
    .par.newdel.p2.rb7 configure -state  disabled
    .par.newdel.p2.rb8 configure -state  disabled
    .par.newdel.p3.ent configure -state  disabled
    .par.newdel.p4.ent configure -state disabled
    .par.newdel.p5.oent configure -state   disabled
    .par.newdel.p5.cent configure -state   disabled
    switch $::OutputDelimiterChoice {
	2 {
	    .par.newdel.p2.rb1 configure -state normal
	    .par.newdel.p2.rb2 configure -state normal
	    .par.newdel.p2.rb3 configure -state normal
	    .par.newdel.p2.rb4 configure -state normal
	    .par.newdel.p2.rb5 configure -state normal
	    .par.newdel.p2.rb6 configure -state normal
	    .par.newdel.p2.rb7 configure -state normal
	    .par.newdel.p2.rb8 configure -state normal
	}
	3 {
	    .par.newdel.p3.ent configure -state normal
	}
	4 {
	    .par.newdel.p4.ent configure -state normal
	}
	5 {
	    .par.newdel.p5.oent  configure -state normal
	    .par.newdel.p5.cent configure -state normal
	}
    }
}


proc PopupErrorInfo {msg} {
    global MSG;
    if {$::TkPresentP} {
	if {[winfo exists .err] == 0} {
	    toplevel .err
	    wm title .err [_ "Errors"]
	    text .err.txt -width 60 -height 10 -wrap word -background \#FFC0D0
	    pack .err.txt -expand 1 -fill both
	}
	.err.txt insert end $msg
	raise .err
	update
    }
}

#option add *dup*background coral 100
proc PopupDuplicationDialog {} {
    toplevel .dup
    wm title .dup [_ "Duplication Settings"]
    BindKeys .dup

    frame .dup.top
    frame .dup.top.xlw
    label .dup.top.xlw.lab -text "Duplicate Embedded Region?" -anchor w
    radiobutton .dup.top.xlw.rdb0 -variable DuplicateSelection -value 0  \
	-text [_ "No"]  -anchor w
    radiobutton .dup.top.xlw.rdb1 -variable DuplicateSelection -value 1  \
	-text [_ "Before Transliteration"]  -anchor w
    radiobutton .dup.top.xlw.rdb2 -variable DuplicateSelection -value 2 \
	-text [_ "After Transliteration"]  -anchor w

    bind .dup <Destroy> [list $::m.misc entryconfigure $::ToggleDuplicationDialogIndex \
			     -label [_ "Display Duplication Dialog"] ]

    balloonhelpd_for .dup.top.xlw.rdb0 \
	[_ "Do not duplicate the embedded region."]
    balloonhelpd_for .dup.top.xlw.rdb1 \
	[_ "Place a duplicate of the embedded region before the transliteration."];
    balloonhelpd_for .dup.top.xlw.rdb2 \
	[_ "Place a duplicate of the embedded region ater  the transliteration."];

    pack .dup.top.xlw.lab -side top -expand 1 -fill both -anchor w
    pack .dup.top.xlw.rdb0 -side left -expand 0 -fill both
    pack .dup.top.xlw.rdb1 -side left -expand 0 -fill both
    pack .dup.top.xlw.rdb2 -side left -expand 1 -fill both

#Surround the duplicate with delimiters?
    frame       .dup.top.rd
    label       .dup.top.rd.lab  -text [_ "Surround the Duplicate with Delimiters?"] -anchor w
    frame       .dup.top.rd.rbf
    radiobutton .dup.top.rd.rbf.rdb0 -variable ::DelimitDuplicateP -value 0 -text [_ "no"]
    radiobutton .dup.top.rd.rbf.rdb1 -variable ::DelimitDuplicateP -value 1 -text [_ "yes"]

    pack .dup.top.rd.rbf.rdb0 .dup.top.rd.rbf.rdb1  -anchor w  -side left -expand 0
    pack .dup.top.rd.lab -side top -expand y -fill x  -anchor w
    pack .dup.top.rd.rbf -side bottom -expand y -anchor w
    .dup.top.rd.rbf.rdb0 configure -state disabled
    .dup.top.rd.rbf.rdb1 configure -state disabled

    balloonhelpd_for .dup.top.rd.rbf.rdb0  [_ "Do not surround the duplicate with delimiters"]
    balloonhelpd_for .dup.top.rd.rbf.rdb1  [_ "Surround the duplicate with delimiters."]

    frame .dup.top.sep -width 2 -relief sunken -background \#444466
    pack  .dup.top.xlw -side left  -expand 1 -fill both -padx 8
    pack  .dup.top.sep -side left  -expand 1 -fill y -anchor w  -padx 8
    pack  .dup.top.rd  -side right -expand 1 -fill both -padx 8

    frame .dup.delim
    label .dup.delim.lab -text [_ "What Delimiters Should be Used?"] -anchor w
    frame .dup.delim.p1
    frame .dup.delim.p2
    frame .dup.delim.p3
    frame .dup.delim.p4
    frame .dup.delim.p5

    radiobutton .dup.delim.p1.rdb -variable DuplicateDelimiterChoice -value 1 -text "HZ"  -anchor w
    pack .dup.delim.p1.rdb -side left -expand 1 -fill both
    bind .dup.delim.p1.rdb <<B3>> ExplainHZEscapes;

    radiobutton .dup.delim.p2.rdb -variable DuplicateDelimiterChoice -value 2 -text [_ "Paired Character"]  -anchor w
    radiobutton .dup.delim.p2.rb1 -variable DupPairedDelimiter  -value 1 -text "\u2329 \u232A"  -anchor w
    radiobutton .dup.delim.p2.rb2 -variable DupPairedDelimiter  -value 2 -text "( )"  -anchor w
    radiobutton .dup.delim.p2.rb3 -variable DupPairedDelimiter  -value 3 -text "\[ \]"  -anchor w
    radiobutton .dup.delim.p2.rb4 -variable DupPairedDelimiter  -value 4 -text "{ }"  -anchor w
    radiobutton .dup.delim.p2.rb5 -variable DupPairedDelimiter  -value 5 -text "\u00AB \u00BB"  -anchor w
    radiobutton .dup.delim.p2.rb6 -variable DupPairedDelimiter  -value 6 -text "< >"  -anchor w
    radiobutton .dup.delim.p2.rb7 -variable DupPairedDelimiter  -value 7 -text "\u2018 \u2019"  -anchor w
    radiobutton .dup.delim.p2.rb8 -variable DupPairedDelimiter  -value 8 -text "\u201C \u201D"  -anchor w

    balloonhelpd_for .dup.delim.p2.rb1\
	[_ "0x2329 Left-Pointing Angle Bracket\n0x232A Right-Pointing Angle Bracket"]
    balloonhelpd_for .dup.delim.p2.rb2\
	[_ "0x0028 Left Parenthesis\n0x0029 Right Parenthesis"]
    balloonhelpd_for .dup.delim.p2.rb3\
	[_ "0x005B Left Square Bracket\n0x005D Right Square Bracket"]
    balloonhelpd_for .dup.delim.p2.rb4\
	[_ "0x007B Left Curly Bracket\n0x007D Right Curly Bracket"]
    balloonhelpd_for .dup.delim.p2.rb5\
	[_ "0x00AB Left-Pointing Double Angle Quotation Mark\n0x00BB Right-Pointing Double Angle Quotation Mark"]
    balloonhelpd_for .dup.delim.p2.rb6\
	[_ "0x003C Less-Than Sign\n0x003E Greater-Than Sign"]
    balloonhelpd_for .dup.delim.p2.rb7\
	[_ "0x2018 Left Single Quotation Mark\n0x2019 Right Single Quotation Mark"]
    balloonhelpd_for .dup.delim.p2.rb8\
	[_ "0x201C Left Double Quotation Mark\n0x201D Right Double Quotation Mark"]

    .dup.delim.p2.rb1 configure -state disabled
    .dup.delim.p2.rb2 configure -state disabled
    .dup.delim.p2.rb3 configure -state disabled
    .dup.delim.p2.rb4 configure -state disabled
    .dup.delim.p2.rb5 configure -state disabled
    .dup.delim.p2.rb6 configure -state disabled
    .dup.delim.p2.rb7 configure -state disabled
    .dup.delim.p2.rb8 configure -state disabled

    pack .dup.delim.p2.rdb -side left -expand 1 -fill both
    pack .dup.delim.p2.rb1 -side left -expand 1 -fill both -padx 5
    pack .dup.delim.p2.rb2 -side left -expand 1 -fill both -padx 5
    pack .dup.delim.p2.rb3 -side left -expand 1 -fill both -padx 5
    pack .dup.delim.p2.rb4 -side left -expand 1 -fill both -padx 5
    pack .dup.delim.p2.rb5 -side left -expand 1 -fill both -padx 5
    pack .dup.delim.p2.rb6 -side left -expand 1 -fill both -padx 5
    pack .dup.delim.p2.rb7 -side left -expand 1 -fill both -padx 5
    pack .dup.delim.p2.rb8 -side left -expand 1 -fill both -padx 5

    radiobutton .dup.delim.p3.rdb -variable DuplicateDelimiterChoice -value 3 -text [_ "Separator Character"]  -anchor w
    entry .dup.delim.p3.ent -width 2
    .dup.delim.p3.ent insert insert $::DuplicateUnpairedDelimiter
    pack .dup.delim.p3.rdb  -side left -expand 0 -fill both
    pack .dup.delim.p3.ent  -side left -expand 1 -fill none -anchor w
    .dup.delim.p3.ent configure -state disabled
    bind .dup.delim.p3.ent  <Control-v> InsertSelection;
    bind .dup.delim.p3.rdb <<B3>> ExplainSeparatorCharacter;

    radiobutton .dup.delim.p4.rdb -variable DuplicateDelimiterChoice -value 4 -text [_ "XML with tag   "]  -anchor w
    entry .dup.delim.p4.ent -width 12
    .dup.delim.p4.ent insert insert $::DuplicateXMLTag
    pack .dup.delim.p4.rdb  -side left -expand 0 -fill both
    pack .dup.delim.p4.ent  -side left -expand 1 -fill none -anchor w
    .dup.delim.p4.ent configure -state disabled
    bind .dup.delim.p4.ent  <Control-v> InsertSelection;
    bind .dup.delim.p4.rdb <<B3>> ExplainXML;

    radiobutton .dup.delim.p5.rdb -variable DuplicateDelimiterChoice  -value 5 \
	-anchor w -state disabled -text [_ "Paired String   "]
    label .dup.delim.p5.olab -text [_ "Open: "] -anchor w
    entry .dup.delim.p5.oent -width 6
    label .dup.delim.p5.clab -text [_ "  Close: "] -anchor w
    entry .dup.delim.p5.cent -width 6
    .dup.delim.p5.oent insert insert $::DuplicatePairedStringOpen
    .dup.delim.p5.cent insert insert $::DuplicatePairedStringClose
    .dup.delim.p5.oent configure -state disabled
    .dup.delim.p5.cent configure -state disabled
    frame .dup.delim.p5.sep2 -width 3

    pack .dup.delim.p5.rdb  -side left -expand 0 -fill none -padx 0
    pack .dup.delim.p5.olab  -side left -expand 0 -fill none  -anchor w  -padx 0
    pack .dup.delim.p5.oent  -side left -expand 1 -fill both -padx 0
    pack .dup.delim.p5.clab -side left -expand 0 -fill none -anchor w -padx 0
    pack .dup.delim.p5.cent -side left -expand 1 -fill both -padx 0
    pack .dup.delim.p5.sep2  -side left -expand 1 -fill both

    .dup.delim.p1.rdb configure -state disabled
    .dup.delim.p2.rdb configure -state disabled
    .dup.delim.p3.rdb configure -state disabled
    .dup.delim.p4.rdb configure -state disabled
    .dup.delim.p5.rdb configure -state disabled

    pack .dup.delim.lab -side top -expand 1 -fill both -pady 0
    pack .dup.delim.p1  -side top -expand 1 -fill both -pady 0
    pack .dup.delim.p2  -side top -expand 1 -fill both -anchor w
    pack .dup.delim.p5  -side top -expand 1 -fill both
    pack .dup.delim.p3  -side top -expand 1 -fill both
    pack .dup.delim.p4  -side top -expand 1 -fill both

    .dup.delim.p1.rdb configure -state disabled
    .dup.delim.p2.rdb configure -state disabled
    .dup.delim.p3.rdb configure -state disabled
    .dup.delim.p4.rdb configure -state disabled
    .dup.delim.p5.rdb configure -state disabled

    frame .dup.cont
    button .dup.cont.d  -text [_ "Dismiss"] -command {destroy .dup}
    button .dup.cont.s  -text [_ "Set"] -command {SetDupDelims}
    pack .dup.cont.d .dup.cont.s -side left -expand 1 -fill both -pady 2 -padx 25

    balloonhelpd_for .dup.cont.s [_ "Store the unpaired delimiter or XML tag\nwithout dismissing the dialog. (All other\nsettings are stored as soon as they are made.)"]
    balloonhelpd_for .dup.cont.d [_ "Dismiss the dialog without storing the new\nunpaired delimiter or XML tag. (All other\nsettings are stored as soon as they are made.)"]

    frame .dup.sep1  -height 3 -borderwidth 1 -relief ridge
    frame .dup.sep2  -height 3 -borderwidth 1 -relief ridge
    frame .dup.sep3  -height 3 -borderwidth 1 -relief ridge

    pack .dup.top     -side top -expand 1 -fill both  -pady 5 -padx 3
    pack .dup.sep1    -side top -expand 1 -fill x     -pady 3
    pack .dup.delim   -side top -expand 1 -fill both  -pady 5 -padx 3
    pack .dup.sep2    -side top -expand 1 -fill x     -pady 3
    pack .dup.cont    -side top -expand 1 -fill both  -pady 5 -padx 3

    trace add variable ::DelimitDuplicateP   write ToggleDupChooseDelimiterTypeActive
    trace add variable ::DuplicateSelection  write ToggleDupDealWithDelimitersActive
    trace add variable ::DuplicateDelimiterChoice  write ControlDuplicateChoiceActive

    .dup.top.xlw.rdb${::DuplicateSelection} invoke
    .dup.top.rd.rbf.rdb${::DelimitDuplicateP} invoke
    if {$::DuplicateDelimiterChoice > 0} {
	.dup.delim.p${::DuplicateDelimiterChoice}.rdb invoke
    }
    .dup.delim.p2.rb${::DupPairedDelimiter} invoke
    ControlDuplicateChoiceActive x y z
}

proc SetDupDelims {} {
    if {[winfo exists .dup] == 0} return;
    set ::DuplicateXMLTag [.dup.delim.p4.ent get]
    set ::DuplicateUnpairedDelimiter [.dup.delim.p3.ent get]
    set ::DuplicatePairedStringOpen   [.dup.delim.p5.oent get]
    set ::DuplicatePairedStringClose  [.dup.delim.p5.cent get]
}

proc ControlDuplicateChoiceActive {n e o} {
#Disable everything, then enable whatever is appropriate. 
    .dup.delim.p2.rb1 configure -state disabled
    .dup.delim.p2.rb2 configure -state  disabled
    .dup.delim.p2.rb3 configure -state  disabled
    .dup.delim.p2.rb4 configure -state  disabled
    .dup.delim.p2.rb5 configure -state  disabled
    .dup.delim.p2.rb6 configure -state  disabled
    .dup.delim.p2.rb7 configure -state  disabled
    .dup.delim.p2.rb8 configure -state  disabled
    .dup.delim.p3.ent configure -state  disabled
    .dup.delim.p4.ent configure -state disabled
    .dup.delim.p5.oent configure -state   disabled
    .dup.delim.p5.cent configure -state   disabled
    switch $::DuplicateDelimiterChoice {
	2 {
	    .dup.delim.p2.rb1 configure -state normal
	    .dup.delim.p2.rb2 configure -state normal
	    .dup.delim.p2.rb3 configure -state normal
	    .dup.delim.p2.rb4 configure -state normal
	    .dup.delim.p2.rb5 configure -state normal
	    .dup.delim.p2.rb6 configure -state normal
	    .dup.delim.p2.rb7 configure -state normal
	    .dup.delim.p2.rb8 configure -state normal
	}
	3 {
	    .dup.delim.p3.ent configure -state normal
	}
	4 {
	    .dup.delim.p4.ent configure -state normal
	}
	5 {
	    .dup.delim.p5.oent  configure -state normal
	    .dup.delim.p5.cent configure -state normal
	}
    }
}

proc ToggleDupDealWithDelimitersActive {n e o} {
    if {$::DuplicateSelection == 0} {
	.dup.top.rd.rbf.rdb0  configure -state disabled
	.dup.top.rd.rbf.rdb1  configure -state disabled
	.dup.delim.p2.rb1 configure -state disabled
	.dup.delim.p2.rb2 configure -state disabled
	.dup.delim.p2.rb3 configure -state disabled
	.dup.delim.p2.rb4 configure -state disabled
	.dup.delim.p2.rb5 configure -state disabled
	.dup.delim.p2.rb6 configure -state disabled
	.dup.delim.p2.rb7 configure -state disabled
	.dup.delim.p2.rb8 configure -state disabled
	.dup.delim.p3.ent configure -state disabled
	.dup.delim.p4.ent configure -state disabled
    } else {
	.dup.top.rd.rbf.rdb0  configure -state normal
	.dup.top.rd.rbf.rdb1  configure -state normal
    }
}

proc ToggleDupChooseDelimiterTypeActive {n e o} {
    if {$::DelimitDuplicateP == 0} {
	.dup.delim.p1.rdb configure -state disabled
	.dup.delim.p2.rdb configure -state disabled
	.dup.delim.p3.rdb configure -state disabled
	.dup.delim.p4.rdb configure -state disabled
	.dup.delim.p5.rdb configure -state disabled
    } else {
	.dup.delim.p1.rdb configure -state normal
	.dup.delim.p2.rdb configure -state normal
	.dup.delim.p3.rdb configure -state normal
	.dup.delim.p4.rdb configure -state normal
	.dup.delim.p5.rdb configure -state normal
    }
}

proc ToggleDuplicatePairedDelimiterChoiceActive {n e o} {
    if {$::DuplicateDelimiterChoice == 2} {
	.dup.delim.p2.rb1 configure -state normal
	.dup.delim.p2.rb2 configure -state normal
	.dup.delim.p2.rb3 configure -state normal
	.dup.delim.p2.rb4 configure -state normal
	.dup.delim.p2.rb5 configure -state normal
	.dup.delim.p2.rb6 configure -state normal
	.dup.delim.p2.rb7 configure -state normal
	.dup.delim.p2.rb8 configure -state normal
    } else {
	.dup.delim.p2.rb1 configure -state disabled
	.dup.delim.p2.rb2 configure -state  disabled
	.dup.delim.p2.rb3 configure -state  disabled
	.dup.delim.p2.rb4 configure -state  disabled
	.dup.delim.p2.rb5 configure -state  disabled
	.dup.delim.p2.rb6 configure -state  disabled
	.dup.delim.p2.rb7 configure -state  disabled
	.dup.delim.p2.rb8 configure -state  disabled
    }
}

set DuplicationDialogIsDisplayedP 0;
proc ToggleDuplicationDialog {} {
    global DuplicationDialogIsDisplayedP;
    global m;

    if {$DuplicationDialogIsDisplayedP == 0} {
	if {[winfo exists .dup]} {
	    wm deiconify .dup;
	    raise .dup;
	} else { 
	    PopupDuplicationDialog;
	}
	set DuplicationDialogIsDisplayedP 1;
	$m.misc entryconfigure $::ToggleDuplicationDialogIndex \
	    -label [_ "Remove Duplication Dialog"];
    } else {
	if {[winfo exists .dup]} {
	    wm iconify .dup;
	}
	set DuplicationDialogIsDisplayedP 0;
	$m.misc entryconfigure $::ToggleDuplicationDialogIndex \
	    -label [_ "Display Duplication Dialog"];
    }
}

proc ExplainWhatShouldBeTransliterated {} {
    global HPWidth;
    global HPLines;
    global PopupList;
    if {[info exists PopupList(WhatShouldBeTransliterated)]} {
	if {[winfo exists $PopupList(WhatShouldBeTransliterated)]} {
	    destroy $PopupList(WhatShouldBeTransliterated);
	    return;
	}
    }
    set po [CreateTextDisplay [_ "What Should Be Transliterated"] $HPWidth $HPLines]
    BindKeys $po;
    set PopupList(WhatShouldBeTransliterated) $po;
    AppendToTextDisplay $po [_ "In some cases an entire piece of text is to be transliterated, but in other cases only certain portions should be. If the text is divided into matrix and embedded regions by delimiters, it is usually the case that only the embedded text is to be transliterated. However, in some cases it is the matrix text that requires transliteration. Consider, for example, a conversation in a language using a non-roman writing system in the course of which some English is used. In this case, it is most likely the bits in English that will be enclosed within delimiters, but for most purposes it is the non-roman matrix text that will require transliteration."];
}

proc CmdMoveEntry {} {
    set m [.cf.ctr.ebx.ent1 get];
    set n [.cf.ctr.ebx.ent2 get]
    if {$m == $n} {
	ShowMessage [_ "That doesn't change anything."];
	return ;
    }
    GetTransliterationFromWindows; #We do this just to get an up-to-date count
    if {$m > $::XLRulesInWindows} {
	ShowMessage [format [_ "Error: attempt to move entry %d - there are only %d entries"] \
		     $m $::XLRulesInWindows];
	return;
    }
    if {$n > $::XLRulesInWindows} {
	ShowMessage [format [_ "Error: attempt to move to entry %d - there are only %d entries"] \
		 $n $::XLRulesInWindows];
	return;
    }
    MoveEntry $m $n 0;
    ShowEntry $n;
    ShowMessage [format [_ "Entry %d moved to position %d."] $m $n];
}

proc CmdDeleteEntry {} {
    set k [.cf.ctr.ebx.ent1 get];
    GetTransliterationFromWindows; #We do this just to get an up-to-date count
    if {$k > $::XLRulesInWindows} {
	ShowMessage [format [_ "Error: attempt to delete entry %d - there are only %d entries"] \
		     $k $::XLRulesInWindows];
	return ;
    }
    DeleteEntry $k;
    ShowMessage [format [_ "Entry %d deleted."] $k];
}

proc CmdShowEntry {} {
    set k [.cf.ctr.ebx.ent1 get];
    GetTransliterationFromWindows; #We do this just to get an up-to-date count
    if {$k > $::XLRulesInWindows} {
	ShowMessage [format [_ "Error: there is no entry %d - there are only %d entries"] \
			 $k $::XLRulesInWindows];
	return ;
    }
    ShowEntry $k;
}

#Move entry m to the position immediately following n.
proc MoveEntry {m n {UndoP 1}} {
    if {$m > $n} {set k [expr $m + 1]} else {set k $m }
    set eleft  [$::XLINP get $m.0 "$m.end + 1 char"];
    set eright [$::XLOUT get $m.0 "$m.end + 1 char"];
    $::XLINP insert $n.0 $eleft; 
    $::XLOUT insert $n.0 $eright;
    $::XLINP delete $k.0 "$k.end + 1 char"
    $::XLOUT delete $k.0 "$k.end + 1 char"
    #Push information on stack for undo,
    #but not if we're undoing a previous move. 
    if {$UndoP == 0} {
	if {$m > $n } {
	    PushES [list MoveEntry $n [expr $m + 1]];
	} else {
	    PushES [list MoveEntry [expr $n -1] $m];
	}
    } else {
	ShowMessage [_ "Entry restored to previous position."]
    }
}

proc DeleteEntry {k} {
    set lhs [$::XLINP get $k.0 $k.end];
    $::XLINP delete $k.0 "$k.end + 1 char"
    set rhs [$::XLOUT get $k.0 $k.end];
    $::XLOUT delete $k.0 "$k.end + 1 char"
    PushES [list InsertEntry $k $lhs $rhs]; #Push information on stack for undo
}

# Insert an entry at the specified position.
# This is needed as the inverse of DeleteEntry for undos.
proc InsertEntry {k lhs rhs} {
    $::XLINP insert $k.0 [format "%s\n" $lhs];
    $::XLOUT insert $k.0 [format "%s\n" $rhs];
    ShowMessage [format [_ "Entry %d restored."] $k];
}

set EditStack [list];
proc PushES {es} {
    lappend ::EditStack $es;
}

proc DumpEditStack {} {
    set Entries [llength $::EditStack];
    if {$Entries == 0} {
	ShowMessage [_ "The edit stack is empty."];
	return ;
    }
    set j 0;
    for {set i [expr $Entries -1]} {$i >= 0} {incr i -1} {
	incr j;
	puts [format "%2d   %s" $j   [lindex $::EditStack $i]]
    }
}

proc Undo {} {
    set inst [PopES];
    if {$inst == ""} {
	ShowMessage [_ "The edit stack is empty. Nothing more can be undone."];
	return ;
    } else {
	eval $inst;
	dmsg $inst;
    }
}

#Pop the stack, returning the value popped.
proc PopES {} {
    set Last [expr [llength $::EditStack] -1];
    if {$Last < 0} {return ""}
    set val [lindex $::EditStack $Last];
    set ::EditStack [lreplace $::EditStack $Last $Last];
    return $val;
}

proc ShowEntry {k} {
    $::XLINP tag delete line
    $::XLOUT tag delete line
    $::XLINP tag configure "line" -background $::ColorSpecs(Entry,Background);
    $::XLOUT tag configure "line" -background $::ColorSpecs(Entry,Background);
    $::XLINP tag add line $k.0 $k.end
    $::XLOUT tag add line $k.0 $k.end
    set TopLine [expr $k - $::HalfHeight -1]
    $::XLINP yview $TopLine;
    $::XLOUT yview $TopLine;
    UpdateEntryLineNumbers;
}




#Initialization file stuff


set CustomCharacterChartFileNameList "";
proc StoreCustomCharacterChartFileName {fn} {
    lappend ::CustomCharacterChartFileNameList $fn;
}

set CustomCharacterChartDataList "";
proc StoreCustomCharacterChartInPlace {dl} {
    lappend ::CustomCharacterChartDataList $dl;
}

#ParseChoice controls how if at all the input text
#is parsed into domains
#  0  do not parse - implies XLWhat = 1
#  1  use HZ escapes
#  2  use simple paired delimiters (e.g. <>)
#  3  use simple unpaired delimiters  (e.g. ||)
#  4  XML
#  5  use paired strings
proc SetParseChoice {s} {
    set ::ParseChoice [ParseChoiceStringToNumber $s]
}

proc SetOutputDelimiterChoice {s} {
    set ::OutputDelimiterChoice [ParseChoiceStringToNumber $s]
}

proc ParseChoiceStringToNumber {s} {
    switch -exact -- $s {
	None {
	    return 0
	}
	HZ {
	    return 1
	}
	Paired         -
	PairedCharacter    -
	PairedCharacters    -
	"Paired Character" -
	"Paired Characters" {
	    return 2
	}
	Separator          -
	SeparatorCharacter     -
	SeparatorCharacters    -
	"Separator Character"  -
	"Separator Characters" {
	    return 3
	}
	XML {
	    return 4
	}
	"PairedString"   -
	"Paired String"  -
	"PairedStrings"  -
	"Paired Strings" {
	    return 5
	}
	default {
	    return 0;
	    ShowMessage [_ "Unknown parse choice"]
	}
    }
}


proc ParseChoiceNumberToString {n} {
    switch -exact -- $n {
	0 {return "None"}
	1 {return HZ}
	2 {return Paired}
	3 {return Separator}
	4 {return XML}
	5 {return "PairedStrings"}
    }
}

proc SetDuplicateDelimiterChoice {s} {
    set ::DuplicateDelimiterChoice [ParseChoiceStringToNumber $s]
}

#The user can specify either or both members of the pair. 
proc SetPairedDelimiter {s} {
    set ::PairedDelimiter [PairedDelimiterStringToNumber [string index $s 0]];
}

proc SetOutputPairDelimiters {s} {
    set ::OutputPairDelimiters [PairedDelimiterStringToNumber [string index $s 0]];
}

#The user can specify either or both members of the pair. 
proc SetDuplicatePairedDelimiter {s} {
    set ::DupPairedDelimiter [PairedDelimiterStringToNumber [string index $s 0]];
}

proc PairedDelimiterStringToNumber {s} {
    switch -exact -- $s {
	\u2329 -
	\u232A {set ::PairedDelimiter 1}
	\u0028 -
       	\u0029 {set ::PairedDelimiter 2}
	\u005B -
	\u005D {set ::PairedDelimiter 3}
	\u007B -
	\u007D {set ::PairedDelimiter 4}
	\u00AB -
	\u00BB {set ::PairedDelimiter 5}
	\u003C - 
	\u003E {set ::PairedDelimiter 6}
	\u2018 -
	\u2019 {set ::PairedDelimiter 7}
	\u201C -
	\u201D {set ::PairedDelimiter 8}
    }
}

proc PairedDelimiterNumberToString {n} {
    switch -exact -- $n {
	1 {return \u2329\u232A}
	2 {return \u0028\u0029}
	3 {return \u005B\u005D}
	4 {return \u007B\u007D}
	5 {return \u00AB\u00BB}
	6 {return \u003C\u003E}
	7 {return \u2018\u2019}
	8 {return \u201C\u201D}
    }
}

proc SetPairedStringDelimiters {o c} {
    set ::PairedStringOpen  [IfNoneNull $o];
    set ::PairedStringClose [IfNoneNull $c];
}

proc SetOutputPairedStringDelimiters {o c} {
    set ::OutputPairedStringOpen [ IfNoneNull $o];
    set ::OutputPairedStringClose [IfNoneNull $c];
}

proc SetDuplicatePairedStringDelimiters {o c} {
    set ::DuplicatePairedStringOpen  [IfNoneNull $o];
    set ::DuplicatePairedStringClose [IfNoneNull $c];
}

proc SetXMLTag {s} {
    set ::XMLTag [IfNoneNull $s];
}

proc SetOutputXMLTag {s} {
    set ::OutputXMLTag [IfNoneNull $s];
}

proc SetDuplicateXMLTag {s} {
    set ::DuplicateXMLTag [IfNoneNull $s];
}

proc SetSeparatorCharacter {s} {
    set ::UnpairedDelimiter [IfNoneNull $s];
}

proc SetOutputUnpairedDelimiter {s} {
    set ::OutputUnpairedDelimiter [IfNoneNull $s];
}

proc SetDuplicateUnpairedDelimiter {s} {
    set ::DuplicateUnpairedDelimiter [IfNoneNull $s];
}

#This variable controls what gets transliterated.
#  0  transliterate the entire input	everything
#  1  transliterate embedded domains only	embedded
# -1  transliterate matrix domains only		matrix
proc SetTransliterateWhat {s} {
    set ::XLWhat [XLWhatStringToNumber $s];
}

proc XLWhatNumberToString {n} {
    switch -exact -- $n {
	1 {return "Everything"}
	2 {return Embedded}
	3 {return Matrix}
    }
}

proc XLWhatStringToNumber {s} {
    switch  -regexp -- $s {
	[Ee]verything {return 1}
	[Ee]mbedded {return 2}
	[Mm]atrix {return 3}
    }
}

proc SetDelimitDuplicateP {b} {
    set ::DelimitDuplicateP [Boolean $b];
}

proc SetDuplicateSelection {s} {
    set ::DuplicateSelection [DuplicateSelectionStringToNumber $s];
}

proc DuplicateSelectionStringToNumber {s} {
    switch -regexp -- $s {
	"[Nn]o(ne)?" {return 0}
	[Bb]efore	-
	"[Bb]efore Transliteration" {return 1}
	[Aa]fter		-
	"[Aa]fter Transliteration" {return 2}
	default {
	    ShowMessage [_ "Unknown duplicate selection"]
	    return 0;
	}
    }
}


proc DuplicateSelectionNumberToString {n} {
    switch -exact -- $n {
	0 {return "No"}
	1 {return "Before"}
	2 {return "After"}
    }
}


proc GetExternalCommandFromEntry {} {
    set ::ExternalCommandList [split [string trim [.eca.ent get]]]
    set cmd [lindex $::ExternalCommandList 0]
    if {[string length [auto_execok $cmd]] == 0} {
	ShowMessage [format "Command %s not found." $cmd]
	set ::ExternalCommandList [list]
	.eca.ent delete 0 end
    }
}

proc GetPluginFromEntry {} {
    set path [.plg.path.ent get]
    set name [.plg.pnam.ent get]
    if { [catch {interp invokehidden safeinterp source $path}] != 0} {
	ShowMessage [_ "An error occurred while reading $name from $path."]
	set ::WhichProcedure 0
	.plg.path.ent delete 0 end
    } elseif {[llength [interp eval safeinterp info commands $name]] == 0} {
	ShowMessage [_ "The file $path does not contain a procedure named $name."]
	set ::WhichProcedure 0
	.plg.pnam.ent delete 0 end
    } else {
	set ::WhichProcedure 2
	set ::PluginPath $path
	set ::PluginName $name
    }
}

proc SetStrictParseP {b} {
    set ::StrictParseP [Boolean $b];
}

proc SetExternalCommandList {s} {
    set ::ExternalCommandList [split $s]
    set ::WhichProcedure 1
    if {$::TkPresentP} {
	.eca.ent delete 0 end
	.eca.ent insert 0 $s
    }
}

#Takes two arguments. The first is the path to the file containing the
#plugin code. The second is the name of the plugin procedure, which
#must take a string as input and return a string.
proc LoadPlugin {path name} {
    if {$::TkPresentP} {
	.plg.path.ent delete 0 end
	.plg.pnam.ent delete 0 end
    }
    if { [catch {interp invokehidden safeinterp source $path}] != 0} {
	ShowMessage [_ "An error occurred while reading $name from $path."]
	set ::WhichProcedure 0;
    } elseif {[info commands $name] == 0} {
	ShowMessage [_ "The file $path does not contain a procedure named $name."]
	set ::WhichProcedure 0;
    } else {
	set ::PluginPath $path;
	set ::PluginName $name;
	set ::WhichProcedure 2;
	if {$::TkPresentP} {
	    .plg.path.ent insert 0 $path
	    .plg.pnam.ent insert 0 $name
	}
    }
}

proc TellPlugin {} {
    if {$::WhichProcedure == 2} {
	return "$::PluginPath $::PluginName"
    } else {
	return ""
    }
}

proc TellExternalCommand {} {
    if {$::WhichProcedure == 1} {
	return [join $::ExternalCommandList]
    } else {
	return ""
    }
}

proc SetBalloonHelpShowP {b} {
    set v [Boolean $b]
    set ::BalloonHelpP $v;
    balloonhelp_control $v;
}

proc SetFoldCaseP {b} {
    set ::FoldCaseP [Boolean $b];
}

proc SetMapLongestFirstP {b} {
    set ::MapLongestFirstP [Boolean $b];
}

proc SetSideBySideLayoutP {b} {
    set ::SideBySideLayoutP [Boolean $b];
}

proc SetDisplayEditingControlsP {b} {
    set ::DisplayEditingControlsP [Boolean $b];
}

proc SelectBrowser {s} {
    set ::DefaultBrowser $s;
}

proc SetTransliterationDelimiter {s} {
    set ::TransliterationDelimiter $s;
}



#This creates a procedure for setting the color of each combination of
#object and feature. The procedure validates the color specification, so
#invalid color specifications are trapped in the slave interpreter.
proc DefineColorSettingProcs {} {
    global ColorProcList;
    foreach cs [array names ::ColorSpecs] {
	set csl [split $cs ","]
	set obj   [lindex $csl 0]
	set which [lindex $csl 1]
	set procname [format "Set%s%sColor" $obj $which]
	lappend ColorProcList $procname
	eval [list "proc" $procname \
		   \
		  "\{cs\}" \
		  "if \{!\[::validcolor::IsColorSpecQ \$cs\]\} \{
		  puts \[format \[_ \"\\\"%s\\\" is not a valid color specification.\"\] \$cs\]
		  return;
		 \}
		  set ::ColorSpecs($obj,$which) \$cs"]
    }
}

proc DefineNoopColorSettingProcs {} {
    global ColorProcList;
    foreach cs [array names ::ColorSpecs] {
	set csl [split $cs ","]
	set obj   [lindex $csl 0]
	set which [lindex $csl 1]
	set procname [format "Set%s%sColor" $obj $which]
	lappend ColorProcList $procname
	eval [list "proc" $procname \
		  \
		  "\{cs\}" \
		  "return;
	      \}"]
    }
}

#Returns a list of init file commands representing the current color settings.
proc SaveColorSettings {} {
    set cl [list]
    foreach cs [array names ::ColorSpecs] {
	set csl [split $cs ","]
	set obj   [lindex $csl 0]
	set which [lindex $csl 1]
	set ps $::ColorSpecs($cs)
	if {[string index $ps 0] == "\#"} {
	    set ps \\$::ColorSpecs($cs)
	} else {
	    set ps [list $ps];
	}
	lappend cl [format "%s%sColor %s" $obj $which $ps]
    }
    return [lsort $cl];
}

#Check whether the argument represents a Boolean value
#and interpret it if it is.
proc Boolean {s} {
     switch -regexp $s {
	 1	{return 1}
	 T.*	{return 1}
	 t.*	{return 1}
	 Y.*	{return 1}
	 y.*	{return 1}
	 ok	{return 1}
	 on	{return 1}
	 0 	{return 0}
	 F.*	{return 0}
	 f.*	{return 0}
	 N.*	{return 0}
	 n.*	{return 0}
	 off	{return 0}
	 default {error}
     }
}

proc GetInitCommands {} {
    # We want a fresh interpreter so that we don't
    # list any commands defined in a user's init file.
    if { [interp exists init]} {
	interp delete init
    }
    InitFileSetup;
    set n 0
    #We rely on the fact that all builtin commands have lower-case
    #case names to separate the commands we provide from the builtins.
    return [interp eval init {lsort [info commands \[A-Z\]*]}]
}

proc GetInitTclCommands {} {
    # We want a fresh interpreter so that we don't
    # list any commands defined in a user's init file.
    if { [interp exists init]} {
	interp delete init
    }
    interp create -safe -- init
    set n 0
    #We rely on the fact that all builtin commands have lower-case
    #case names to separate the commands we provide from the builtins.
    return [interp eval init {lsort [info commands \[a-z\]*]}]
}

proc SaveInitCommands {} {
    set SaveFile [tk_getSaveFile -initialfile [_ "InitFileCommands"]];
    if {[string equal $SaveFile ""]} {
	return ;
    } else {
        if { [catch {open $SaveFile "w+"} fh] != 0} {
	    ShowMessage [format \
		     [_ "Unable to open file %s in which to save initialization file commands."] \
		     $SaveFile];
	    return ;
        }
    }
    set n 0;
    foreach l [GetInitCommands] {
	incr n;
	puts $fh [format "\[%03d\] %s" $n $l]
    }
    close $fh;
}

set PopupList(InitCommandList) "";
proc PopupInitCommandList {} {
    if {[PopupDown InitCommandList] ==1} {return}
    set po  [CreateTextDisplay [_ "Initialization File Commands"] 65 10];
    set ::PopupList(InitCommandList) $po;
    BindKeys $po;
    set n 0;
    foreach l [GetInitCommands] {
	incr n;
	AppendToTextDisplay $po [format "\[%03d\] %s\n" $n $l];
    }
}

proc SaveInitTclCommands {} {
    set SaveFile [tk_getSaveFile -initialfile [_ "InitFileTclCommands"]];
    if {[string equal $SaveFile ""]} {
	return ;
    } else {
        if { [catch {open $SaveFile "w+"} fh] != 0} {
	    ShowMessage [format \
		     [_ "Unable to open file %s in which to save Tcl commands."] \
		     $SaveFile];
	    return ;
        }
    }
    set n 0;
    foreach l [GetInitTclCommands] {
	incr n;
	puts $fh [format "\[%03d\] %s" $n $l]
    }
    close $fh;
}

proc PopupInitTclCommandList {} {
    if {[PopupDown InitTclCommandList] ==1} {return}
    set po  [CreateTextDisplay [_ "Initialization File Tcl Commands"] 65 10];
    set ::PopupList(InitTclCommandList) $po;
    BindKeys $po;
    set n 0;
    foreach l [GetInitTclCommands] {
	incr n;
	AppendToTextDisplay $po [format "\[%03d\] %s\n" $n $l];
    }
}

proc IfNullNone {s} {
    if {[string length $s] == 0} {
	return "None"
    } else {
	return $s;
    }
}

proc IfNoneNull {s} {
    if {[string match -nocase "none" $s]} {
	return ""
    } else {
	return $s;
    }
}

proc SaveMiscellaneousSettings {} {
    set MiscellaneousInitCommandList \
	[list \
	     [SlaveName  SetDuplicateDelimiterChoice] \
	     	[ParseChoiceNumberToString $::DuplicateDelimiterChoice]\
	     [SlaveName  SetDuplicatePairedDelimiter] \
	     	[PairedDelimiterNumberToString $::DupPairedDelimiter]\
	     [SlaveName  SetDuplicateUnpairedDelimiter] [IfNullNone $::DuplicateUnpairedDelimiter]\
	     [SlaveName  SetDuplicateXMLTag] [IfNullNone $::DuplicateXMLTag]\
	     [SlaveName  SetParseChoice] [ParseChoiceNumberToString $::ParseChoice]\
	     [SlaveName  SetOutputDelimiterChoice] [ParseChoiceNumberToString $::OutputDelimiterChoice]\
	     [SlaveName  SetOutputPairDelimiters] [PairedDelimiterNumberToString $::OutputPairDelimiters]\
	     [SlaveName  SetOutputUnpairedDelimiter] [IfNullNone $::OutputUnpairedDelimiter]\
	     [SlaveName  SetOutputXMLTag] [IfNullNone $::OutputXMLTag]\
	     [SlaveName  SetPairedDelimiter] [PairedDelimiterNumberToString $::PairedDelimiter]\
	     [SlaveName  SetXMLTag] [IfNullNone $::XMLTag]\
	     [SlaveName  SetSeparatorCharacter] [IfNullNone $::UnpairedDelimiter]\
	     [SlaveName  SetTransliterateWhat] [XLWhatNumberToString $::XLWhat]\
	     [SlaveName  SetStrictParseP] $::StrictParseP\
	     [SlaveName SelectBrowser] $::DefaultBrowser\
	     [SlaveName SetBalloonHelpShowP]			$::BalloonHelpP \
	     [SlaveName SetDelimitDuplicateP]			$::DelimitDuplicateP\
	     [SlaveName SetFoldCaseP]				$::FoldCaseP\
	     [SlaveName SetDisplayEditingControlsP]		$::DisplayEditingControlsP\
	     [SlaveName SetDisplayConsonantChartColumnLabelsP]	$::DisplayConsonantChartColumnLabelsP \
	     [SlaveName SetDisplayConsonantChartRowLabelsP]	$::DisplayConsonantChartRowLabelsP \
	     [SlaveName SetDisplayVowelChartColumnLabelsP]	$::DisplayVowelChartColumnLabelsP \
	     [SlaveName SetDisplayVowelChartRowLabelsP]		$::DisplayVowelChartRowLabelsP \
	     [SlaveName SetDuplicateSelection]			[DuplicateSelectionNumberToString $::DuplicateSelection]\
	     [SlaveName SetMapLongestFirstP]			$::MapLongestFirstP\
	     [SlaveName SetSideBySideLayoutP]			$::SideBySideLayoutP\
	     [SlaveName SetTransliterationDelimiter]		[UnicodeStringToHex $::TransliterationDelimiter]]
    set PluginString [TellPlugin]
    if {[string length $PluginString] > 0} {
	lappend MiscellaneousInitCommandList [SlaveName LoadPlugin]
	lappend MiscellaneousInitCommandList $PluginString
    }

    set ECS [TellExternalCommand]
    if {[string length $ECS] > 0} {
	lappend MiscellaneousInitCommandList [SlaveName  SetExternalCommandList] 
	lappend MiscellaneousInitCommandList $ECS
    }

    if {[info exists ::AccentedLettersPerRow]} {
	lappend MiscellaneousInitCommandList [SlaveName SetAccentedLetterColumns]
	lappend MiscellaneousInitCommandList $::AccentedLettersPerRow
    }

    if {[llength $MiscellaneousInitCommandList] %2 == 1} {
	set msg [_ "The miscellaneous list is ill-formed."]
	append msg [_ "Please file a bug report. Here is the list:"]
	append msg "\n"
	append msg $MiscellaneousInitCommandList
	error $msg
    }
    set cl [list]
    foreach {command value} $MiscellaneousInitCommandList {
	lappend cl [format "%s %s" $command $value]
    }
    foreach title [array names ::SpecialCharacterPalette] {
	if {[string match "*WidgetName" $title]} {continue}
	set info $::SpecialCharacterPalette($title);
	if {[llength $info] > 4} {
	    lappend cl [list ReadCharacterChart [lindex $info 4]];
	} else {
	    lappend cl [ReconstructCharacterChartDefinition $title];
	}
    }
    return $cl;
}

proc ReconstructCharacterChartDefinition {Title} {
    set info $::SpecialCharacterPalette($Title);
    set txt DefineCharacterChart;
    append txt " \{\n"
    append txt [format "%s" [Quote $Title]]
    set flds [llength $info];
    if {$flds > 1} {
	append txt [format "|%d" [lindex $info 1]];# Number of columns
    }
    if {$flds > 2} {
	append txt [format "|%s" [lindex $info 2]];# Font family
    }
    if {$flds > 3} {
	append txt [format "|%s" [lindex $info 3]];# Font size
    }
    append txt "\n"
    #Each entry consists of a unicode string and a gloss
    foreach x [lindex $info 0] {
	set string [lindex $x 0]
	set gloss  [lindex $x 1]
	append txt [format "%s|%s\n" [UnicodeStringToHex $string] [Quote $gloss]];
    }
    append txt "\}"
    return  $txt;
}

#This is a kludge but I encountered quoting hell trying to use (\\s) -> \\ \\1
#e.g.   regsub -all "\\s" $s {\ } r

#Escape spaces and tabs
proc Quote {s} {
    regsub -all "\u0020" $s {\ } r
    regsub -all "\t" $r {\	} ret
    return $ret
}

proc UnicodeStringToHex {s} {
    set rv "";
    foreach c [split $s ""] {
	scan $c "%c" t
	append rv [format "\\u%04X" $t];
    }
    return $rv;
 }

#This package defines a routine IsColorSpecQ that validates X11 color specifications.
#It returns 1 if its argument is either a recognized X11 color name or
#a well formed hexadecimal RGB specification. If so desired, the two routines on
#which it is based, IsColorNameQ and IsColorNumericQ can also be exported.
namespace eval validcolor {
    namespace export IsColorSpecQ;
    namespace export GetColorNames;
    variable CN;

    set CN([list ghost white]) 1
    set CN([list white smoke]) 1
    set CN([list floral white]) 1
    set CN([list antique white]) 1
    set CN([list old lace]) 1
    set CN([list papaya whip]) 1
    set CN([list blanched almond]) 1
    set CN([list peach puff]) 1
    set CN([list navajo white]) 1
    set CN([list lemon chiffon]) 1
    set CN([list mint cream]) 1
    set CN([list alice blue]) 1
    set CN([list lavender blush]) 1
    set CN([list misty rose]) 1
    set CN([list dark slate gray]) 1
    set CN([list dark slate grey]) 1
    set CN([list dim gray]) 1
    set CN([list dim grey]) 1
    set CN([list slate gray]) 1
    set CN([list slate grey]) 1
    set CN([list light slate gray]) 1
    set CN([list light slate grey]) 1
    set CN([list light grey]) 1
    set CN([list light gray]) 1
    set CN([list midnight blue]) 1
    set CN([list navy blue]) 1
    set CN([list cornflower blue]) 1
    set CN([list dark slate blue]) 1
    set CN([list slate blue]) 1
    set CN([list medium slate blue]) 1
    set CN([list light slate blue]) 1
    set CN([list medium blue]) 1
    set CN([list royal blue]) 1
    set CN([list dodger blue]) 1
    set CN([list deep sky blue]) 1
    set CN([list sky blue]) 1
    set CN([list light sky blue]) 1
    set CN([list steel blue]) 1
    set CN([list light steel blue]) 1
    set CN([list light blue]) 1
    set CN([list powder blue]) 1
    set CN([list pale turquoise]) 1
    set CN([list dark turquoise]) 1
    set CN([list medium turquoise]) 1
    set CN([list light cyan]) 1
    set CN([list cadet blue]) 1
    set CN([list medium aquamarine]) 1
    set CN([list dark green]) 1
    set CN([list dark olive green]) 1
    set CN([list dark sea green]) 1
    set CN([list sea green]) 1
    set CN([list medium sea green]) 1
    set CN([list light sea green]) 1
    set CN([list pale green]) 1
    set CN([list spring green]) 1
    set CN([list lawn green]) 1
    set CN([list medium spring green]) 1
    set CN([list green yellow]) 1
    set CN([list lime green]) 1
    set CN([list yellow green]) 1
    set CN([list forest green]) 1
    set CN([list olive drab]) 1
    set CN([list dark khaki]) 1
    set CN([list pale goldenrod]) 1
    set CN([list light goldenrod yellow]) 1
    set CN([list light yellow]) 1
    set CN([list light goldenrod]) 1
    set CN([list dark goldenrod]) 1
    set CN([list rosy brown]) 1
    set CN([list indian red]) 1
    set CN([list saddle brown]) 1
    set CN([list sandy brown]) 1
    set CN([list dark salmon]) 1
    set CN([list light salmon]) 1
    set CN([list dark orange]) 1
    set CN([list light coral]) 1
    set CN([list orange red]) 1
    set CN([list hot pink]) 1
    set CN([list deep pink]) 1
    set CN([list light pink]) 1
    set CN([list pale violet red]) 1
    set CN([list medium violet red]) 1
    set CN([list violet red]) 1
    set CN([list medium orchid]) 1
    set CN([list dark orchid]) 1
    set CN([list dark violet]) 1
    set CN([list blue violet]) 1
    set CN([list medium purple]) 1
    set CN(snow) 1
    set CN(GhostWhite) 1
    set CN(WhiteSmoke) 1
    set CN(gainsboro) 1
    set CN(FloralWhite) 1
    set CN(OldLace) 1
    set CN(linen) 1
    set CN(AntiqueWhite) 1
    set CN(PapayaWhip) 1
    set CN(BlanchedAlmond) 1
    set CN(bisque) 1
    set CN(PeachPuff) 1
    set CN(NavajoWhite) 1
    set CN(moccasin) 1
    set CN(cornsilk) 1
    set CN(ivory) 1
    set CN(LemonChiffon) 1
    set CN(seashell) 1
    set CN(honeydew) 1
    set CN(MintCream) 1
    set CN(azure) 1
    set CN(AliceBlue) 1
    set CN(lavender) 1
    set CN(LavenderBlush) 1
    set CN(MistyRose) 1
    set CN(white) 1
    set CN(black) 1
    set CN(DarkSlateGray) 1
    set CN(DarkSlateGrey) 1
    set CN(DimGray) 1
    set CN(DimGrey) 1
    set CN(SlateGray) 1
    set CN(SlateGrey) 1
    set CN(LightSlateGray) 1
    set CN(LightSlateGrey) 1
    set CN(gray) 1
    set CN(grey) 1
    set CN(LightGrey) 1
    set CN(LightGray) 1
    set CN(MidnightBlue) 1
    set CN(navy) 1
    set CN(NavyBlue) 1
    set CN(CornflowerBlue) 1
    set CN(DarkSlateBlue) 1
    set CN(SlateBlue) 1
    set CN(MediumSlateBlue) 1
    set CN(LightSlateBlue) 1
    set CN(MediumBlue) 1
    set CN(RoyalBlue) 1
    set CN(blue) 1
    set CN(DodgerBlue) 1
    set CN(DeepSkyBlue) 1
    set CN(SkyBlue) 1
    set CN(LightSkyBlue) 1
    set CN(SteelBlue) 1
    set CN(LightSteelBlue) 1
    set CN(LightBlue) 1
    set CN(PowderBlue) 1
    set CN(PaleTurquoise) 1
    set CN(DarkTurquoise) 1
    set CN(MediumTurquoise) 1
    set CN(turquoise) 1
    set CN(cyan) 1
    set CN(LightCyan) 1
    set CN(CadetBlue) 1
    set CN(MediumAquamarine) 1
    set CN(aquamarine) 1
    set CN(DarkGreen) 1
    set CN(DarkOliveGreen) 1
    set CN(DarkSeaGreen) 1
    set CN(SeaGreen) 1
    set CN(MediumSeaGreen) 1
    set CN(LightSeaGreen) 1
    set CN(PaleGreen) 1
    set CN(SpringGreen) 1
    set CN(LawnGreen) 1
    set CN(green) 1
    set CN(chartreuse) 1
    set CN(MediumSpringGreen) 1
    set CN(GreenYellow) 1
    set CN(LimeGreen) 1
    set CN(YellowGreen) 1
    set CN(ForestGreen) 1
    set CN(OliveDrab) 1
    set CN(DarkKhaki) 1
    set CN(khaki) 1
    set CN(PaleGoldenrod) 1
    set CN(LightGoldenrodYellow) 1
    set CN(LightYellow) 1
    set CN(yellow) 1
    set CN(gold) 1
    set CN(LightGoldenrod) 1
    set CN(goldenrod) 1
    set CN(DarkGoldenrod) 1
    set CN(RosyBrown) 1
    set CN(IndianRed) 1
    set CN(SaddleBrown) 1
    set CN(sienna) 1
    set CN(peru) 1
    set CN(burlywood) 1
    set CN(beige) 1
    set CN(wheat) 1
    set CN(SandyBrown) 1
    set CN(tan) 1
    set CN(chocolate) 1
    set CN(firebrick) 1
    set CN(brown) 1
    set CN(DarkSalmon) 1
    set CN(salmon) 1
    set CN(LightSalmon) 1
    set CN(orange) 1
    set CN(DarkOrange) 1
    set CN(coral) 1
    set CN(LightCoral) 1
    set CN(tomato) 1
    set CN(OrangeRed) 1
    set CN(red) 1
    set CN(HotPink) 1
    set CN(DeepPink) 1
    set CN(pink) 1
    set CN(LightPink) 1
    set CN(PaleVioletRed) 1
    set CN(maroon) 1
    set CN(MediumVioletRed) 1
    set CN(VioletRed) 1
    set CN(magenta) 1
    set CN(violet) 1
    set CN(plum) 1
    set CN(orchid) 1
    set CN(MediumOrchid) 1
    set CN(DarkOrchid) 1
    set CN(DarkViolet) 1
    set CN(BlueViolet) 1
    set CN(purple) 1
    set CN(MediumPurple) 1
    set CN(thistle) 1
    set CN(snow1) 1
    set CN(snow2) 1
    set CN(snow3) 1
    set CN(snow4) 1
    set CN(seashell1) 1
    set CN(seashell2) 1
    set CN(seashell3) 1
    set CN(seashell4) 1
    set CN(AntiqueWhite1) 1
    set CN(AntiqueWhite2) 1
    set CN(AntiqueWhite3) 1
    set CN(AntiqueWhite4) 1
    set CN(bisque1) 1
    set CN(bisque2) 1
    set CN(bisque3) 1
    set CN(bisque4) 1
    set CN(PeachPuff1) 1
    set CN(PeachPuff2) 1
    set CN(PeachPuff3) 1
    set CN(PeachPuff4) 1
    set CN(NavajoWhite1) 1
    set CN(NavajoWhite2) 1
    set CN(NavajoWhite3) 1
    set CN(NavajoWhite4) 1
    set CN(LemonChiffon1) 1
    set CN(LemonChiffon2) 1
    set CN(LemonChiffon3) 1
    set CN(LemonChiffon4) 1
    set CN(cornsilk1) 1
    set CN(cornsilk2) 1
    set CN(cornsilk3) 1
    set CN(cornsilk4) 1
    set CN(ivory1) 1
    set CN(ivory2) 1
    set CN(ivory3) 1
    set CN(ivory4) 1
    set CN(honeydew1) 1
    set CN(honeydew2) 1
    set CN(honeydew3) 1
    set CN(honeydew4) 1
    set CN(LavenderBlush1) 1
    set CN(LavenderBlush2) 1
    set CN(LavenderBlush3) 1
    set CN(LavenderBlush4) 1
    set CN(MistyRose1) 1
    set CN(MistyRose2) 1
    set CN(MistyRose3) 1
    set CN(MistyRose4) 1
    set CN(azure1) 1
    set CN(azure2) 1
    set CN(azure3) 1
    set CN(azure4) 1
    set CN(SlateBlue1) 1
    set CN(SlateBlue2) 1
    set CN(SlateBlue3) 1
    set CN(SlateBlue4) 1
    set CN(RoyalBlue1) 1
    set CN(RoyalBlue2) 1
    set CN(RoyalBlue3) 1
    set CN(RoyalBlue4) 1
    set CN(blue1) 1
    set CN(blue2) 1
    set CN(blue3) 1
    set CN(blue4) 1
    set CN(DodgerBlue1) 1
    set CN(DodgerBlue2) 1
    set CN(DodgerBlue3) 1
    set CN(DodgerBlue4) 1
    set CN(SteelBlue1) 1
    set CN(SteelBlue2) 1
    set CN(SteelBlue3) 1
    set CN(SteelBlue4) 1
    set CN(DeepSkyBlue1) 1
    set CN(DeepSkyBlue2) 1
    set CN(DeepSkyBlue3) 1
    set CN(DeepSkyBlue4) 1
    set CN(SkyBlue1) 1
    set CN(SkyBlue2) 1
    set CN(SkyBlue3) 1
    set CN(SkyBlue4) 1
    set CN(LightSkyBlue1) 1
    set CN(LightSkyBlue2) 1
    set CN(LightSkyBlue3) 1
    set CN(LightSkyBlue4) 1
    set CN(SlateGray1) 1
    set CN(SlateGray2) 1
    set CN(SlateGray3) 1
    set CN(SlateGray4) 1
    set CN(LightSteelBlue1) 1
    set CN(LightSteelBlue2) 1
    set CN(LightSteelBlue3) 1
    set CN(LightSteelBlue4) 1
    set CN(LightBlue1) 1
    set CN(LightBlue2) 1
    set CN(LightBlue3) 1
    set CN(LightBlue4) 1
    set CN(LightCyan1) 1
    set CN(LightCyan2) 1
    set CN(LightCyan3) 1
    set CN(LightCyan4) 1
    set CN(PaleTurquoise1) 1
    set CN(PaleTurquoise2) 1
    set CN(PaleTurquoise3) 1
    set CN(PaleTurquoise4) 1
    set CN(CadetBlue1) 1
    set CN(CadetBlue2) 1
    set CN(CadetBlue3) 1
    set CN(CadetBlue4) 1
    set CN(turquoise1) 1
    set CN(turquoise2) 1
    set CN(turquoise3) 1
    set CN(turquoise4) 1
    set CN(cyan1) 1
    set CN(cyan2) 1
    set CN(cyan3) 1
    set CN(cyan4) 1
    set CN(DarkSlateGray1) 1
    set CN(DarkSlateGray2) 1
    set CN(DarkSlateGray3) 1
    set CN(DarkSlateGray4) 1
    set CN(aquamarine1) 1
    set CN(aquamarine2) 1
    set CN(aquamarine3) 1
    set CN(aquamarine4) 1
    set CN(DarkSeaGreen1) 1
    set CN(DarkSeaGreen2) 1
    set CN(DarkSeaGreen3) 1
    set CN(DarkSeaGreen4) 1
    set CN(SeaGreen1) 1
    set CN(SeaGreen2) 1
    set CN(SeaGreen3) 1
    set CN(SeaGreen4) 1
    set CN(PaleGreen1) 1
    set CN(PaleGreen2) 1
    set CN(PaleGreen3) 1
    set CN(PaleGreen4) 1
    set CN(SpringGreen1) 1
    set CN(SpringGreen2) 1
    set CN(SpringGreen3) 1
    set CN(SpringGreen4) 1
    set CN(green1) 1
    set CN(green2) 1
    set CN(green3) 1
    set CN(green4) 1
    set CN(chartreuse1) 1
    set CN(chartreuse2) 1
    set CN(chartreuse3) 1
    set CN(chartreuse4) 1
    set CN(OliveDrab1) 1
    set CN(OliveDrab2) 1
    set CN(OliveDrab3) 1
    set CN(OliveDrab4) 1
    set CN(DarkOliveGreen1) 1
    set CN(DarkOliveGreen2) 1
    set CN(DarkOliveGreen3) 1
    set CN(DarkOliveGreen4) 1
    set CN(khaki1) 1
    set CN(khaki2) 1
    set CN(khaki3) 1
    set CN(khaki4) 1
    set CN(LightGoldenrod1) 1
    set CN(LightGoldenrod2) 1
    set CN(LightGoldenrod3) 1
    set CN(LightGoldenrod4) 1
    set CN(LightYellow1) 1
    set CN(LightYellow2) 1
    set CN(LightYellow3) 1
    set CN(LightYellow4) 1
    set CN(yellow1) 1
    set CN(yellow2) 1
    set CN(yellow3) 1
    set CN(yellow4) 1
    set CN(gold1) 1
    set CN(gold2) 1
    set CN(gold3) 1
    set CN(gold4) 1
    set CN(goldenrod1) 1
    set CN(goldenrod2) 1
    set CN(goldenrod3) 1
    set CN(goldenrod4) 1
    set CN(DarkGoldenrod1) 1
    set CN(DarkGoldenrod2) 1
    set CN(DarkGoldenrod3) 1
    set CN(DarkGoldenrod4) 1
    set CN(RosyBrown1) 1
    set CN(RosyBrown2) 1
    set CN(RosyBrown3) 1
    set CN(RosyBrown4) 1
    set CN(IndianRed1) 1
    set CN(IndianRed2) 1
    set CN(IndianRed3) 1
    set CN(IndianRed4) 1
    set CN(sienna1) 1
    set CN(sienna2) 1
    set CN(sienna3) 1
    set CN(sienna4) 1
    set CN(burlywood1) 1
    set CN(burlywood2) 1
    set CN(burlywood3) 1
    set CN(burlywood4) 1
    set CN(wheat1) 1
    set CN(wheat2) 1
    set CN(wheat3) 1
    set CN(wheat4) 1
    set CN(tan1) 1
    set CN(tan2) 1
    set CN(tan3) 1
    set CN(tan4) 1
    set CN(chocolate1) 1
    set CN(chocolate2) 1
    set CN(chocolate3) 1
    set CN(chocolate4) 1
    set CN(firebrick1) 1
    set CN(firebrick2) 1
    set CN(firebrick3) 1
    set CN(firebrick4) 1
    set CN(brown1) 1
    set CN(brown2) 1
    set CN(brown3) 1
    set CN(brown4) 1
    set CN(salmon1) 1
    set CN(salmon2) 1
    set CN(salmon3) 1
    set CN(salmon4) 1
    set CN(LightSalmon1) 1
    set CN(LightSalmon2) 1
    set CN(LightSalmon3) 1
    set CN(LightSalmon4) 1
    set CN(orange1) 1
    set CN(orange2) 1
    set CN(orange3) 1
    set CN(orange4) 1
    set CN(DarkOrange1) 1
    set CN(DarkOrange2) 1
    set CN(DarkOrange3) 1
    set CN(DarkOrange4) 1
    set CN(coral1) 1
    set CN(coral2) 1
    set CN(coral3) 1
    set CN(coral4) 1
    set CN(tomato1) 1
    set CN(tomato2) 1
    set CN(tomato3) 1
    set CN(tomato4) 1
    set CN(OrangeRed1) 1
    set CN(OrangeRed2) 1
    set CN(OrangeRed3) 1
    set CN(OrangeRed4) 1
    set CN(red1) 1
    set CN(red2) 1
    set CN(red3) 1
    set CN(red4) 1
    set CN(DeepPink1) 1
    set CN(DeepPink2) 1
    set CN(DeepPink3) 1
    set CN(DeepPink4) 1
    set CN(HotPink1) 1
    set CN(HotPink2) 1
    set CN(HotPink3) 1
    set CN(HotPink4) 1
    set CN(pink1) 1
    set CN(pink2) 1
    set CN(pink3) 1
    set CN(pink4) 1
    set CN(LightPink1) 1
    set CN(LightPink2) 1
    set CN(LightPink3) 1
    set CN(LightPink4) 1
    set CN(PaleVioletRed1) 1
    set CN(PaleVioletRed2) 1
    set CN(PaleVioletRed3) 1
    set CN(PaleVioletRed4) 1
    set CN(maroon1) 1
    set CN(maroon2) 1
    set CN(maroon3) 1
    set CN(maroon4) 1
    set CN(VioletRed1) 1
    set CN(VioletRed2) 1
    set CN(VioletRed3) 1
    set CN(VioletRed4) 1
    set CN(magenta1) 1
    set CN(magenta2) 1
    set CN(magenta3) 1
    set CN(magenta4) 1
    set CN(orchid1) 1
    set CN(orchid2) 1
    set CN(orchid3) 1
    set CN(orchid4) 1
    set CN(plum1) 1
    set CN(plum2) 1
    set CN(plum3) 1
    set CN(plum4) 1
    set CN(MediumOrchid1) 1
    set CN(MediumOrchid2) 1
    set CN(MediumOrchid3) 1
    set CN(MediumOrchid4) 1
    set CN(DarkOrchid1) 1
    set CN(DarkOrchid2) 1
    set CN(DarkOrchid3) 1
    set CN(DarkOrchid4) 1
    set CN(purple1) 1
    set CN(purple2) 1
    set CN(purple3) 1
    set CN(purple4) 1
    set CN(MediumPurple1) 1
    set CN(MediumPurple2) 1
    set CN(MediumPurple3) 1
    set CN(MediumPurple4) 1
    set CN(thistle1) 1
    set CN(thistle2) 1
    set CN(thistle3) 1
    set CN(thistle4) 1
    set CN(gray0) 1
    set CN(grey0) 1
    set CN(gray1) 1
    set CN(grey1) 1
    set CN(gray2) 1
    set CN(grey2) 1
    set CN(gray3) 1
    set CN(grey3) 1
    set CN(gray4) 1
    set CN(grey4) 1
    set CN(gray5) 1
    set CN(grey5) 1
    set CN(gray6) 1
    set CN(grey6) 1
    set CN(gray7) 1
    set CN(grey7) 1
    set CN(gray8) 1
    set CN(grey8) 1
    set CN(gray9) 1
    set CN(grey9) 1
    set CN(gray10) 1
    set CN(grey10) 1
    set CN(gray11) 1
    set CN(grey11) 1
    set CN(gray12) 1
    set CN(grey12) 1
    set CN(gray13) 1
    set CN(grey13) 1
    set CN(gray14) 1
    set CN(grey14) 1
    set CN(gray15) 1
    set CN(grey15) 1
    set CN(gray16) 1
    set CN(grey16) 1
    set CN(gray17) 1
    set CN(grey17) 1
    set CN(gray18) 1
    set CN(grey18) 1
    set CN(gray19) 1
    set CN(grey19) 1
    set CN(gray20) 1
    set CN(grey20) 1
    set CN(gray21) 1
    set CN(grey21) 1
    set CN(gray22) 1
    set CN(grey22) 1
    set CN(gray23) 1
    set CN(grey23) 1
    set CN(gray24) 1
    set CN(grey24) 1
    set CN(gray25) 1
    set CN(grey25) 1
    set CN(gray26) 1
    set CN(grey26) 1
    set CN(gray27) 1
    set CN(grey27) 1
    set CN(gray28) 1
    set CN(grey28) 1
    set CN(gray29) 1
    set CN(grey29) 1
    set CN(gray30) 1
    set CN(grey30) 1
    set CN(gray31) 1
    set CN(grey31) 1
    set CN(gray32) 1
    set CN(grey32) 1
    set CN(gray33) 1
    set CN(grey33) 1
    set CN(gray34) 1
    set CN(grey34) 1
    set CN(gray35) 1
    set CN(grey35) 1
    set CN(gray36) 1
    set CN(grey36) 1
    set CN(gray37) 1
    set CN(grey37) 1
    set CN(gray38) 1
    set CN(grey38) 1
    set CN(gray39) 1
    set CN(grey39) 1
    set CN(gray40) 1
    set CN(grey40) 1
    set CN(gray41) 1
    set CN(grey41) 1
    set CN(gray42) 1
    set CN(grey42) 1
    set CN(gray43) 1
    set CN(grey43) 1
    set CN(gray44) 1
    set CN(grey44) 1
    set CN(gray45) 1
    set CN(grey45) 1
    set CN(gray46) 1
    set CN(grey46) 1
    set CN(gray47) 1
    set CN(grey47) 1
    set CN(gray48) 1
    set CN(grey48) 1
    set CN(gray49) 1
    set CN(grey49) 1
    set CN(gray50) 1
    set CN(grey50) 1
    set CN(gray51) 1
    set CN(grey51) 1
    set CN(gray52) 1
    set CN(grey52) 1
    set CN(gray53) 1
    set CN(grey53) 1
    set CN(gray54) 1
    set CN(grey54) 1
    set CN(gray55) 1
    set CN(grey55) 1
    set CN(gray56) 1
    set CN(grey56) 1
    set CN(gray57) 1
    set CN(grey57) 1
    set CN(gray58) 1
    set CN(grey58) 1
    set CN(gray59) 1
    set CN(grey59) 1
    set CN(gray60) 1
    set CN(grey60) 1
    set CN(gray61) 1
    set CN(grey61) 1
    set CN(gray62) 1
    set CN(grey62) 1
    set CN(gray63) 1
    set CN(grey63) 1
    set CN(gray64) 1
    set CN(grey64) 1
    set CN(gray65) 1
    set CN(grey65) 1
    set CN(gray66) 1
    set CN(grey66) 1
    set CN(gray67) 1
    set CN(grey67) 1
    set CN(gray68) 1
    set CN(grey68) 1
    set CN(gray69) 1
    set CN(grey69) 1
    set CN(gray70) 1
    set CN(grey70) 1
    set CN(gray71) 1
    set CN(grey71) 1
    set CN(gray72) 1
    set CN(grey72) 1
    set CN(gray73) 1
    set CN(grey73) 1
    set CN(gray74) 1
    set CN(grey74) 1
    set CN(gray75) 1
    set CN(grey75) 1
    set CN(gray76) 1
    set CN(grey76) 1
    set CN(gray77) 1
    set CN(grey77) 1
    set CN(gray78) 1
    set CN(grey78) 1
    set CN(gray79) 1
    set CN(grey79) 1
    set CN(gray80) 1
    set CN(grey80) 1
    set CN(gray81) 1
    set CN(grey81) 1
    set CN(gray82) 1
    set CN(grey82) 1
    set CN(gray83) 1
    set CN(grey83) 1
    set CN(gray84) 1
    set CN(grey84) 1
    set CN(gray85) 1
    set CN(grey85) 1
    set CN(gray86) 1
    set CN(grey86) 1
    set CN(gray87) 1
    set CN(grey87) 1
    set CN(gray88) 1
    set CN(grey88) 1
    set CN(gray89) 1
    set CN(grey89) 1
    set CN(gray90) 1
    set CN(grey90) 1
    set CN(gray91) 1
    set CN(grey91) 1
    set CN(gray92) 1
    set CN(grey92) 1
    set CN(gray93) 1
    set CN(grey93) 1
    set CN(gray94) 1
    set CN(grey94) 1
    set CN(gray95) 1
    set CN(grey95) 1
    set CN(gray96) 1
    set CN(grey96) 1
    set CN(gray97) 1
    set CN(grey97) 1
    set CN(gray98) 1
    set CN(grey98) 1
    set CN(gray99) 1
    set CN(grey99) 1
    set CN(gray100) 1
    set CN(grey100) 1

    proc GetColorNames {} {
	variable CN
	return [lsort -dictionary [array names CN]]
    }

    proc IsColorNameQ {s} {
	variable CN
	if {[info exists CN($s)]} {
	    return 1;
	} else {
	    return 0;
	}
    }

    proc IsColorNumericQ {s} {
	return [regexp \
		    {^\#([[:xdigit:]]{6}|[[:xdigit:]]{3}|[[:xdigit:]]{9}|[[:xdigit:]]{12})$} $s]
    }
		
    proc IsColorSpecQ {s} {
	if {[IsColorNameQ $s]} {
	    return 1;
	} elseif {[IsColorNumericQ $s]} {
	    return 1;
	} else {
	    return 0;
	}
    }
}
#End of namespace validcolor

proc PopupColorNameList {} {
    if {[PopupDown ColorNameList] ==1} {return}
    set po  [CreateTextDisplay [_ "Color Names"] 65 10];
    set ::PopupList(ColorNameList) $po;
    BindKeys $po;
    set n 0;
    foreach l [::validcolor::GetColorNames] {
	incr n;
	AppendToTextDisplay $po [format "\[%03d\] %s\n" $n $l];
    }
}

#Ths procedure takes as its sole argument the name
#of an existing procedure in the same interpreter
#and creates a new master level procedure that
#wraps the other procedure with a test for the
#correct argument count.

proc WrapMaster {pn} {
    regsub "^::fontsel::" $pn "" Stem
    set WrappedName Wrapped${Stem};
    set sp [format "proc $WrappedName \{args\} \{\n"];
    set ArgsNeeded [llength [info args $pn]];
    append sp [format "\tif \{\[llength \$args\] != %d\} \{\n" $ArgsNeeded]
    if {$ArgsNeeded == 1} {
	set emsg [format [_ "The command %s in the file %%s expects one argument"] \
	[SlaveName $WrappedName]]
	set cmd [format "\t\tputs \"$emsg\"\n" \$::InitFile]
	append sp $cmd
    } else {
	set emsg [format [_ "The command %s in the file %%s expects %d arguments\n"] \
	      [SlaveName $WrappedName] $ArgsNeeded]
	set cmd [format "\t\tputs \"$emsg\"\n" \$::InitFile]; 
	append sp $cmd;
    }
    append sp "\t\treturn;\n\t\}\n\t"
    append sp  $pn
    for {set i 0} {$i < $ArgsNeeded} {incr i} {
	append sp " \[lindex \$args $i\]"
    }
    append sp "\n\}"
    eval $sp;
    return $WrappedName;
}

proc ReadInitFile {name} {
    if { [catch {interp invoke init source $name} msg] } {
	puts [format [_  "Error on reading init file %s: %s"] $name $msg]
    } else {
	if {$::VerboseP} {
	    ShowMessage [format [_ "Executed init file %s"] $name]
	}
    }
    interp delete init
}

proc SlavePutsStderr {msg} {
    puts stderr $msg;
}

proc SlaveMsgcat {} {
    package require msgcat
    ::msgcat::mcload [file join [file dirname [info script]] msgs];
}

#This is the list of the master interpreter commands
#other than those generated from data structures (currently
#color- and font-setting) that we wish to expose for use in
#the init file slave interpreter.

set MiscellaneousSlaveCommandList {
    DefineShortcut
    LoadPlugin
    SelectBrowser
    SetBalloonHelpShowP
    SetDelimitDuplicateP
    SetDisplayEditingControlsP
    SetDisplayConsonantChartColumnLabelsP
    SetDisplayConsonantChartRowLabelsP
    SetDisplayVowelChartColumnLabelsP
    SetDisplayVowelChartRowLabelsP
    SetAccentedLetterColumns
    SetDuplicateDelimiterChoice
    SetDuplicatePairedDelimiter
    SetDuplicatePairedStringDelimiters
    SetDuplicateSelection
    SetDuplicateUnpairedDelimiter
    SetDuplicateXMLTag
    SetExternalCommandList
    SetFoldCaseP
    SetMapLongestFirstP
    SetOutputDelimiterChoice
    SetOutputPairDelimiters
    SetOutputPairedStringDelimiters
    SetOutputUnpairedDelimiter
    SetOutputXMLTag
    SetPairedDelimiter
    SetPairedStringDelimiters
    SetParseChoice
    SetSideBySideLayoutP
    SetStrictParseP
    SetTransliterateWhat
    SetTransliterationDelimiter
    SetSeparatorCharacter
    SetXMLTag
    SlavePutsStderr
    StoreCustomCharacterChartFileName
    StoreCustomCharacterChartInPlace
}

#This procedure creates the master procedures that will be exposed
#in the slave interpreter. It takes care of creating the base procedures
#where, as in the case of color and font setting, these are created
#programmatically, then wraps these as well as the manually
#written procedures that are to be exposed.
proc CreateInitProcs {} {
    set ::FontProcList "";
    if {$::BatchModeP} {
	fontsel::DefineNoopFontSettingProcs
	DefineNoopColorSettingProcs;
    } else { 
	fontsel::DefineFontSettingProcs
	DefineColorSettingProcs;
    }
    #This is the complete list of commands that we want to expose 
    #in the slave init file interpreter. These are the raw, unwrapped, master
    #level commands.
    set ::MasterCommandsToBeExposedInInitList [concat \
	   [lsort $::ColorProcList] \
	   [lsort $::FontProcList]  \
	   [lsort $::MiscellaneousSlaveCommandList]];
    #Now we wrap them all so as to trap valence errors
    set ::WrappedMasterCommandList [list]
    foreach c $::MasterCommandsToBeExposedInInitList {
	lappend ::WrappedMasterCommandList [WrapMaster $c]
    }
}

#This procedure takes the list of wrapped master
#procedures and exposes them in a specified, already
#created, slave interpreter.
proc ExposeInitProcs {Interpreter} {
    foreach c $::WrappedMasterCommandList {
	set sl [SlaveName $c];
	$Interpreter alias $sl $c;
    }
}

set MasterToSlaveName(SetDisplayConsonantChartColumnLabelsP) ShowConsonantChartColumnLabelsP
set MasterToSlaveName(SetDisplayConsonantChartRowLabelsP) ShowConsonantChartRowLabelsP
set MasterToSlaveName(SetDisplayVowelChartColumnLabelsP) ShowVowelChartColumnLabelsP
set MasterToSlaveName(SetDisplayVowelChartRowLabelsP) ShowVowelChartRowLabelsP
set MasterToSlaveName(SelectBrowser) Browser
set MasterToSlaveName(SlavePutsStderr) PutError
set MasterToSlaveName(StoreCustomCharacterChartInPlace) DefineCharacterChart 
set MasterToSlaveName(StoreCustomCharacterChartFileName) ReadCharacterChart 

#Given the name of a wrapped command in the master interpreter,
#returns the name that the command is to have when
#exposed in the slave interpreter. In most cases the alias
#is computed from the master name, but in some cases
#the relationship is irregular. 
proc SlaveName {command} {
    regsub "^Wrapped" $command "" Unwrapped
    if {[info exists ::MasterToSlaveName($Unwrapped)]} {
	set EnglishSlaveName $::MasterToSlaveName($Unwrapped);
    } else {
	regsub "^Set" $Unwrapped "" EnglishSlaveName
    }
    return [_ $EnglishSlaveName]
}

proc InitFileSetup {} {
    interp create -safe -- init
    #We use slaveMsgcat rather than SlaveMsgcat so that the name does not
    #appear in the list that the user sees.
    init alias slaveMsgcat SlaveMsgcat
    init eval slaveMsgcat
    init alias mc ::msgcat::mc
    ExposeInitProcs init
    #This takes care of errors in the name of commands while still in the
    #slave interpreter, without aborting, and using the user-name in the error-message.
    init eval {
	proc unknown {args} {
	    set cmd [lindex $args 0]
	    Puts [format  [mc "Unknown command %s in script file %s."]  $cmd [info script]];
#	    Puts [format  [mc "Unknown command %1\$s in script file %2\$s."]  $cmd [info script]];
	    return ;
	}
    }
}

proc SlaveBoolean {s} {
    switch -regexp $s {
	1	{return 1}
	T.*	{return 1}
	t.*	{return 1}
	Y.*	{return 1}
	y.*	{return 1}
	ok	{return 1}
	on	{return 1}
	0 	{return 0}
	F.*	{return 0}
	f.*	{return 0}
	N.*	{return 0}
	n.*	{return 0}
	off	{return 0}
	default {
	    puts [format [msgcat::mc "%s is not a Boolean argument."] $s]
	    return -1; 
	}
    }
}

#Write out a list of init file commands that will reconstruct
#the current configuration 
proc SaveState {{SaveFile Ask}} {
    if {[string equal $SaveFile Ask]} { 
	set SaveFile [tk_getSaveFile -initialfile [_ "SavedState"]];
	if {$SaveFile == ""} {
	    ShowMessage [_ "File selection cancelled."]
	    return ;
	}
    }
    if {[catch {open $SaveFile "w"} fh ] != 0} {
	ShowMessage [format [_ "Unable to open file %s."] $SaveFile];
	return ;
    }
    set cl [concat \
	[list [format "\#%s" [clock format [clock seconds]]]] \
	[list [format "\#%s" [_ Fonts]]] \
	[fontsel::SaveFontSettings] \
	[list [format "\#%s" [_ Colors]]] \
	[SaveColorSettings] \
	[list [format "\#%s" [_ Miscellaneous]]] \
	[SaveMiscellaneousSettings]];
    foreach x $cl {
	puts $fh $x;
    }
    #Parameters with more than one value require special handling
    puts $fh [format "%s %s %s" \
	  [SlaveName  SetPairedStringDelimiters] \
	  [IfNullNone $::PairedStringOpen]\
	  [IfNullNone $::PairedStringClose]]
    puts $fh [format "%s %s %s" \
	  [SlaveName  SetDuplicatePairedStringDelimiters] \
	  [IfNullNone $::DuplicatePairedStringOpen] \
	  [IfNullNone $::DuplicatePairedStringClose]]
    puts $fh [format "%s %s %s" \
	  [SlaveName SetOutputPairedStringDelimiters] \
	  [IfNullNone $::OutputPairedStringOpen] \
	  [IfNullNone $::OutputPairedStringClose]]
    close $fh;
    ShowMessage [format [_ "The configuration has been saved in file %s."] \
		     [MinimizeFileName $SaveFile]]
}


proc RepackEditingControls {args} {
    UnpackXL;
    if {$::WhichProcedure == 0} {
	PackXL;
    }
}

proc RepackScrollbars {args} {
    RepackEditingControls;
    LayoutTextWindows;
}

proc PackMain {} {
    pack forget .cf
    pack forget .txt
    pack forget .top
    pack forget .eca
    pack forget .plg
    if {$::WhichProcedure == 1} {
	pack .eca -in .top -side top -expand 0 -fill x -padx 2 -pady 2
    } elseif {$::WhichProcedure == 2} {
	pack .plg -in .top -side top -expand 0 -fill x -padx 2 -pady 2
    } else {
	pack .cf  -in .top -side top -expand 1 -fill both -padx 2 -pady 3
    }
    pack .txt -in .top -side top -expand 1 -fill both -padx 2 -pady 3
    pack .top -expand 1 -fill both -padx 3 -pady 0
}

proc LayoutTextWindows {args} {
    pack forget $::TXTINP
    pack forget $::TXTOUT
    pack forget $::TXTISB
    pack forget $::TXTOSB
    pack forget .txt.vwi
    pack forget .txt.vwo
    if {$::SideBySideLayoutP} {
	$::TXTINP  configure -width $::HalfWidth
	$::TXTOUT  configure -width $::HalfWidth
	if {$::UseScrollbarsP} {
	    pack $::TXTISB -side left  -expand 0 -fill y
	    pack $::TXTOSB -side right -expand 0 -fill y
	}
	pack $::TXTINP -side left -expand 1 -fill both  -padx 3
	pack $::TXTOUT -side right -expand 1 -fill both -padx 3
    } else {
	if {$::UseScrollbarsP} {
	    pack $::TXTISB -side right  -expand 0 -fill y -in .txt.vwi
	}
	pack $::TXTINP -side left -expand 1 -fill both -in .txt.vwi
	if {$::UseScrollbarsP} {
	    pack $::TXTOSB -side right -expand 0 -fill y -in .txt.vwo
	}
	pack $::TXTOUT -side left -expand 1 -fill both -in .txt.vwo
	pack .txt.vwi  -side top    -fill both -expand 1 -pady 3
	pack .txt.vwo  -side bottom -fill both -expand 1 -pady 3
    }
    pack .txt -side bottom -expand 1 -fill both
}

proc Usage {} {
    puts {Usage: xlit [options]};
    puts [_ "    -b                       batch mode - do not put up graphical interface"];
    puts [_ "    -d <DuplicateSelection>  set DuplicateSelection"];
    puts [_ "    -f <file>                read configuration from <file>"];
    puts [_ "    -h                       help"];
    puts [_ "    -i                       do not read initialization file"];
    puts [_ "    -l <file>                read left hand side of transliteration definition from <file>"];
    puts [_ "    -p <ParseChoice>         set TransliterateWhat"];
    puts [_ "    -q                       quiet mode - don't chat so much"];
    puts [_ "    -t <file>                read text to transliterate from <file>"];
    puts [_ "    -x <file>                read transliteration definition from <file>"];
    puts [_ "    -v                       version"];
}


proc PrintCopyright {fh} {
    puts $fh "Copyright (C) 2005-2008 William J. Poser.";
    puts $fh [_ "This program is free software; you can redistribute
it and/or modify it under the terms of version 3 of the
GNU General Public License as published by the Free
Software Foundation."];
}

proc PrintVersion {fh} {
    puts $fh "xlit $::Version";
}

proc PathToFileURL {p} {
    return [format "file://%s" $p];
}

#If the filename passed as argument is a pathname
#leading to a file in the current working directory,
#return just the basename+extension. Otherwise
#return the argument.
proc MinimizeFileName {s} {
    set cwd [pwd];
    set sdir [file dirname $s]
    if {[string equal $cwd $sdir]} {
	return [file tail $s]
    } else {
	return $s;
    }
}

proc ControlBalloonHelp {} {
    if {$::BalloonHelpToggle} {
	balloonhelp_control 1
    } else {
	balloonhelp_control 0
    }
}

proc PopupDown {n} {
    global PopupList;
    if {[info exists PopupList($n)]} {
	if {[winfo exists $PopupList($n)]} {
	    destroy $PopupList($n);
	    return 1;
	}
    }
    return 0;
}

proc argshift {k} {
    set ::argv [lrange $::argv $k end];
}

proc DefineShortcut {command keyspec} {
    #Validate command
    if {![info exists ::CommandGlosses($command)]} {
	ShowMessage [format [_ "%s is not a command for which a shortcut can be created"] $command]
	return ;
    }
    #Validate keyspec
    if {![regexp -- {(<Control-[[:alpha:]]>|<KeyPress-F[1-9]>|<KeyPress-F1[0-2]>|<Control-KeyPress-F[1-9]>|<Control-KeyPress-F1[0-2]>)+} $keyspec]} {
	ShowMessage [format [_ "%s is not a valid key specification"] $keyspec]
	return ;
    }
    #Set
    set ::KeyBindings($keyspec) $command;
}

proc SetCommandGlosses {} {
    global CommandGlosses;

    set Pairs [list\
		   LoadTransliterationFile "[_ "Load Transliteration Definition"]"\
		   SaveTransliterationNativeFormat  "[_ "Save Transliteration Definition (Native Format)"]"\
		   SaveOutput  "[_ "Save Transliterated Text"]"\
		   ShutDown  "[_ "Quit"]"\
		   Transliterate  "[_ "Transliterate"]"\
		   Abort  "[_ "Abort"]"\
		   ClearInputText  "[_ "Clear Input Text"]"\
		   ClearTransliteratedText  "[_ "Clear Output Text"]"\
		   ClearTransliteration  "[_ "Clear Transliteration Definition"]"\
		   fontsel::CreateFontControlPanel  "[_ "Select Font"]"\
		   search::SearchResults  "[_ "Search Results"]"\
		   search::SearchInput "[_ "Search Input"]"\
		   search::SearchXLLHS  "[_ "Search Transliteration LHS"]"\
		   search::SearchXLRHS  "[_ "Search Transliteration RHS"]"\
		   SwapXLInOut  "[_ "Swap Transliteration Input and Output"]"\
		   ToggleBalloonHelp  "[_ "Toggle Balloon Help"]"\
	   ToggleCharEntryByCode  "[_ "Enter characters by their hexadecimal Unicode codes"]"\
	   ToggleBalloonHelp "[_ "Toggle Balloon Help"]"\
	   ToggleDescribeKeyBindings "[_ "Toggle Key Binding Help"]"\
	   ToggleParseDialog "[_ "Toggle the dialog for making parser settings ."]"\
	   ToggleDuplicationDialog "[_ "Toggle the dialog for making duplication settings."]"\
	   ToggleIPAA "[_ "Toggle the accented letter entry widget."]"\
	   ToggleIPAC "[_ "Toggle the IPA consonant entry widget."]"\
	   ToggleIPAV "[_ "Toggle the IPA vowel entry widget."]"\
           ToggleIPAD "[_ "Toggle the IPA diacritic entry widget."]"];
    array set CommandGlosses $Pairs;
}

proc BindKeys {w {bl KeyBindings}} {
    upvar \#0 $bl kb;
    foreach KeySeq [array names kb] {
	bind $w $KeySeq $kb($KeySeq);
    }
}

set PopupList(DescribeKeyBindings) "";
proc DescribeKeyBindings {{bl KeyBindings}} {
    global CommandGlosses;
    upvar \#0 $bl kb;
    if {[PopupDown DescribeKeyBindings] ==1} {return}
    set po [CreateTextDisplay [_ "Key Bindings"] 88 9]
    set ::PopupList(DescribeKeyBindings) $po;
    BindKeys $po
    #Invert array and get maximum length of gloss while we are at it.
    set MaxGlossLength 0;
    foreach KeySeq [array names kb] {
	set Gloss $CommandGlosses($kb($KeySeq));
	set GlossLength [string length $Gloss];
	if {$MaxGlossLength < $GlossLength} {set MaxGlossLength $GlossLength}
	set GlossToKeySeq($Gloss) $KeySeq;
    }
    #Now generate entries in alphabetical order of command glosses
    foreach Gloss [lsort [array names GlossToKeySeq]] {
	AppendToTextDisplay $po \
	    [format "%-*s\t%s\n" $MaxGlossLength $Gloss $GlossToKeySeq($Gloss)];
    }
}

proc ToggleDescribeKeyBindings {} {
    set po $::PopupList(DescribeKeyBindings);
    if {[winfo exists $po] == 0} {
	DescribeKeyBindings;
    } else {
	destroy $po;
    }
}

proc PopupShortcutCommands {} {
    global PopupList;
    if {[PopupDown PopupShortcutCommands] ==1} {return}
    set CommandNames [array names ::CommandGlosses]
    set maxglosslen 0;
    set maxtotallen 0;
    foreach c $CommandNames {
	set GlossLength [string length $::CommandGlosses($c)]
	set CommandLength [string length $c]
	set TotalLength [expr $CommandLength + $GlossLength]
	if {$GlossLength > $maxglosslen} {
	    set maxglosslen $GlossLength;
	}
	if {$TotalLength > $maxtotallen} {
	    set maxtotallen $TotalLength;
	}
    }
    set ht [llength $CommandNames]
    if {$ht > 15} {set ht 15}
    set po [CreateTextDisplay [_ "Commands Available for Shortcuts"] [expr $maxtotallen + 8] $ht]
    BindKeys $po;
    set PopupList(PopupShortcutCommands) $po;
    foreach n $CommandNames {
	lappend GlossList [list $::CommandGlosses($n) $n]
    }
    foreach pair [lsort -dictionary -index 0 $GlossList] {
	set gloss [lindex $pair 0]
	set command [lindex $pair 1]
	AppendToTextDisplay $po [format "%-*s  %s\n" $maxglosslen $gloss $command];
    }
}

array set KeyBindings {
"<Control-a>"             ToggleIPAA
"<Control-b>" 		  ToggleBalloonHelp
"<Control-D>"             ToggleDuplicationDialog
"<Control-d>"             ToggleIPAD
"<Control-i>"             search::SearchInput
"<Control-j>"             ToggleIPAC
"<Control-k>"             ToggleDescribeKeyBindings
"<Control-p>"             ToggleParseDialog
"<Control-r>"             search::SearchResults
"<Control-u>"             ToggleCharEntryByCode
"<Control-w>"             ToggleIPAV
};

set HomeDir [GetHomeDir];
set ScriptPath [info script];
if {$DebugP} {
    set ManualDirectory [file join [file dir $ScriptPath] Manual];
} else {
    set ManualDirectory [file join $NonBinPath Manual];
}
set ManualPath [file join $ManualDirectory Manual.html];
set ManualURL [PathToFileURL $ManualPath];


#Open the journal file.
set JournalFullPath [file join $HomeDir $JournalFile];
if {[catch {open $JournalFullPath "a"} JournalHandle ] != 0} {
    puts [format [_ "Unable to open journal file %s."] $JournalFullPath];
    exit 2;
}
WriteJournal [format [_ "Session started: %s in %s"] [clock format [clock seconds]] [pwd]];

#Can't use -c at present as can't figure out quoting to preserve additional args
#Get command line arguments
while {[string match -* [lindex $argv 0]]} {
    switch -glob -- [lindex $argv 0] {
        -p* {
	    set CommandLineTransliterateWhat [lindex $argv 1];
	    argshift 2;
	}
	-d* {
	    set CommandLineDuplicateSelection [lindex $argv 1];
	    argshift 2;
	}
	-b* {
	    set BatchModeP 1;
	    argshift 1;
	}
	-f* {
	    set InitFile [lindex $argv 1];
	    argshift 2;
	} 
	-l* {
	    set CommandLineLHSTransliterationDefinition [lindex $argv 1];
	    argshift 2;
	} 
	-t* {
	    set CommandLineTextFile [lindex $argv 1];
	    argshift 2;
	} 
	-h* {
	    Usage;
	    exit 0;
	}
	-i* {
	    set ReadInitFileP 0;
	    argshift 1;
	}
	-q* {
	    set ::VerboseP 0;
	    argshift 1;
	}
	-x* {
	    set CommandLineTransliterationDefinition [lindex $argv 1];
	    argshift 2;
	}
	-v* {
	    PrintVersion stdout
	    exit 0;
	}
	default {
	    puts [format [_ "Command line option %s not recognized."] [lindex $argv 0]];
	    PrintVersion stdout
	    Usage;
	    exit 0;
	}
    }
};					# End of while

SetCommandGlosses;
if {$::VerboseP} {
    PrintVersion stderr
    PrintCopyright stderr
}
if {!$::BatchModeP} {
    package require Tk
    package require Iwidgets
    set ::TkPresentP 1;
    DetermineGraphicsSystem
    set ::TkEightFivePlusOnX11P [CheckTkEightFivePlusOnX11]
    if {$::TkEightFivePlusOnX11P} {
	set FontInfo(MainFont,size) 10
	set FontInfo(LHSFont,size) 10
	set FontInfo(RHSFont,size) 10
	set FontInfo(MenuFont,size) 10
	set FontInfo(CharacterEntryLabelFont,size) 11
	set FontInfo(CharacterEntryFont,size) 12
	set FontInfo(DiacriticFont,size) 11
	set FontInfo(AccentedLetterFont,size) 12
	set FontInfo(BalloonHelpFont,size) 11
    }
    SetupEvents $System;
    SetCharacterEntryOptions
    SetupImages;
    SetProgressbarOptions
    fontsel::SetFontInfoDefaults;
    fontsel::CreateFonts;
    wm withdraw [fontsel::CreateFontControlPanel]
    fontsel::SetFontSelectionDefaults
    fontsel::RecordDefaults
}


if {[string equal $CommandLineTransliterateWhat ""] == 0} {
    set ::XLWhat [XLWhatStringToNumber $CommandLineTransliterateWhat]
}

if {[string equal $CommandLineDuplicateSelection ""] == 0} {
    set ::DuplicateSelection [DuplicateSelectionStringToNumber $CommandLineDuplicateSelection]
}

set HalfWidth  [expr int($MainWidth/2.0)]
set HalfHeight [expr int($MainHeight/2.0)]

if {!$::BatchModeP} {
    option add *selectFont*Font FontControlPanelFont 100
    option add *Label.Font MainFont 100
    #option add *.Menu.font MenuFont 100
    option add *Menu.Font MenuFont 100
    option add *Radiobutton.Font MainFont 100
    option add *Scale.Font MainFont 100

    option add *Background $ColorSpecs(Default,Background) 60;
    option add *Entry.Background  $::ColorSpecs(Entry,Background) 100; 
    option add *DisabledBackground "\#949CD2" 60;
    option add *Button.activeBackground "\#FF0044" 60;
    option add *Radiobutton.activeBackground $::ColorSpecs(RadioButton,ActiveBackground) 100;
    option add *Radiobutton.selectColor  $::ColorSpecs(RadioButton,SelectedBackground) 100;

    option add *Menu.Background $ColorSpecs(Menu,Background) 100
    option add *Menu.Foreground $ColorSpecs(Menu,Foreground) 100
    option add *Menu*selectColor $ColorSpecs(Menu,Select) 100
    option add *Menu.activeBackground $ColorSpecs(Menu,ActiveBackground) 100
    option add *Menu.activeForeground  $ColorSpecs(Menu,ActiveForeground) 100

    SetupBalloonHelp

    #Create menubar
    frame .top -background $ColorSpecs(MainFrame,Background)
    set m [menu .menubar -tearoff 0 -font MenuFont\
	       -bg $ColorSpecs(Menubar,Background)\
	       -fg $ColorSpecs(Menubar,Foreground)\
	       -activebackground $ColorSpecs(Menubar,ActiveBackground)\
	       -activeforeground $ColorSpecs(Menubar,ActiveForeground)
	  ];

    set TopIndex 0;
    if {$AquaP} {incr TopIndex -1}
    incr TopIndex;
    $m add cascade -label [_ "File"]   -menu [menu $m.file];

    set FileIndex 0
    if {$AquaP} {incr FileIndex -1}
    incr FileIndex;
    $m.file add command -label [_ "Load Text To Be Transliterated"] -command LoadInputText

    incr FileIndex;
    $m.file add command -label [_ "Save Text To Be Transliterated"] -command SaveInputText

    incr FileIndex;
    $m.file add command -label [_ "Lock Text To be Transliterated"] -command LockInputText;

    incr FileIndex;
    $m.file add command -label [_ "Unlock Text To be Transliterated"] -command UnlockInputText;

    incr FileIndex;
    $m.file add separator

    incr FileIndex;
    $m.file add command -label [_ "Load Transliteration Definition"] \
	-command LoadTransliterationFile;

    incr FileIndex;
    $m.file add command -label [_ "Load Yudit Kmap File"] \
	-command {LoadTransliterationFile ParseYuditKmapEntry}

    incr FileIndex;
    $m.file add command -label [_ "Load Transliteration Lefthand Side"] -command LoadTransliterationLHS;
    incr FileIndex;
    $m.file add command -label [_ "Load Transliteration Righthand Side"] -command LoadTransliterationRHS;
    incr FileIndex;
    $m.file add command -label [_ "Save Transliteration Definition (Native Format)"] -command {SaveTransliteration 1}

    incr FileIndex;
    $m.file add command -label [_ "Save Transliteration Definition (Yudit Kmap Format)"] -command {SaveTransliteration 0}

    incr FileIndex;
    $m.file add separator

    incr FileIndex;
    $m.file add command -label [_ "Save Transliterated Text"] -command SaveOutput

    incr FileIndex;
    $m.file add separator

    incr FileIndex;
    $m.file add command -label [_ "Quit"]  -command ShutDown

    if {!$AquaP} {
	incr TopIndex;
	$m add command -label [_ "Transliterate"]  -command Transliterate;

	incr TopIndex;
	$m add command -label [_ "Abort"]  -command Abort;
    }

    incr TopIndex;
    $m add cascade -label [_ "Clear"] -menu [menu $m.clear]

    set ClearIndex 0;
    if {$AquaP} {incr ClearIndex -1}
    incr ClearIndex;
    $m.clear add command -label [_ "Clear Input Text"]  -command ClearInputText;

    incr ClearIndex;
    $m.clear add command -label [_ "Clear Output Text"]  -command ClearTransliteratedText;

    incr ClearIndex;
    $m.clear add command -label [_ "Clear Transliteration Definition"]  -command ClearTransliteration;

    incr TopIndex;
    if {$AquaP} {incr CharacterEntryIndex -1}
    set CharacterEntryIndex 0;
    $m add cascade -label [_ "Character Entry"] -menu [menu $m.charentry];

    incr CharacterEntryIndex;
    set CustomCharacterChartIndex $CharacterEntryIndex;
    $m.charentry add command -label [_ "Load Custom Character Chart Definition"] \
	-command ReadCustomCharacterChartPopup;

    incr CharacterEntryIndex;
    set ToggleIPAAIndex $CharacterEntryIndex;
    $m.charentry add command -label [_ "Display Accented Letter Chart"] -command ToggleIPAA;

    incr CharacterEntryIndex;
    set ToggleIPACIndex $CharacterEntryIndex;
    $m.charentry add command -label [_ "Display IPA Consonant Chart"] -command ToggleIPAC;

    incr CharacterEntryIndex;
    set ToggleIPADIndex $CharacterEntryIndex;
    $m.charentry add command -label [_ "Display IPA Diacritic Chart"] -command ToggleIPAD;

    incr CharacterEntryIndex;
    set ToggleIPAVIndex $CharacterEntryIndex;
    $m.charentry add command -label [_ "Display IPA Vowel Chart"] -command ToggleIPAV;

    incr CharacterEntryIndex;
    set ToggleCharEntryByCodeIndex $CharacterEntryIndex;
    $m.charentry add command -label [_ "Display Widget for Entering Characters by Unicode Code"] -command ToggleCharEntryByCode;

    incr TopIndex;
    $m add cascade -label [_ "Miscellaneous"]   -menu [menu $m.misc];
    set MiscIndex 0;
    if {$AquaP} {incr MiscIndex -1}
    incr MiscIndex;
    $m.misc add cascade -label [_ "Configuration"] -menu [menu $m.misc.configure];

    set ParameterIndex 0;
    if {$AquaP} {incr ParameterIndex -1}

    incr ParameterIndex;
    $m.misc.configure add radiobutton -label [_ "Use Map"] \
	-value 0 -variable WhichProcedure

    incr ParameterIndex;
    $m.misc.configure add radiobutton -label [_ "Use External Command"] \
	-value 1 -variable WhichProcedure

    incr ParameterIndex;
    $m.misc.configure add radiobutton -label [_ "Use Plugin"] \
	-value 2 -variable WhichProcedure

    incr ParameterIndex;
    $m.misc.configure add separator

    $m.misc.configure add cascade  -label [_ "Character Entry"] -menu [menu $m.misc.configure.charentry];
    $m.misc.configure.charentry add checkbutton -label [_ "Display Consonant Chart Column Labels"] \
	-variable DisplayConsonantChartColumnLabelsP -onvalue 1 -offvalue 0 \
	-command ControlDisplayConsonantChartColumnLabels
    $m.misc.configure.charentry add checkbutton -label [_ "Display Consonant Chart Row Labels"] \
	-variable DisplayConsonantChartRowLabelsP -onvalue 1 -offvalue 0 \
	-command ControlDisplayConsonantChartRowLabels
    $m.misc.configure.charentry add checkbutton -label [_ "Display Vowel Chart Column Labels"] \
	-variable DisplayVowelChartColumnLabelsP -onvalue 1 -offvalue 0 \
	-command ControlDisplayVowelChartColumnLabels
    $m.misc.configure.charentry add checkbutton -label [_ "Display Vowel Chart Row Labels"] \
	-variable DisplayVowelChartRowLabelsP -onvalue 1 -offvalue 0 \
	-command ControlDisplayVowelChartRowLabels

    incr ParameterIndex;
    $m.misc.configure add cascade  -label [_ "Layout"] -menu [menu $m.misc.configure.layout];
    $m.misc.configure.layout add checkbutton -label [_ "Use Side-By-Side Layout"] \
	-onvalue 1 -offvalue 0 -variable SideBySideLayoutP;
    trace add variable SideBySideLayoutP write LayoutTextWindows

    $m.misc.configure.layout add checkbutton -label [_ "Display Editing Controls"] \
	-onvalue 1 -offvalue 0 -variable DisplayEditingControlsP;
    trace add variable DisplayEditingControlsP write RepackEditingControls

    $m.misc.configure.layout add checkbutton -label [_ "Display Scrollbars"] \
	-onvalue 1 -offvalue 0 -variable UseScrollbarsP;
    trace add variable UseScrollbarsP write RepackScrollbars;

    incr ParameterIndex;
    $m.misc.configure add command -label [_ "Save Configuration"] -command {SaveState};

    incr ParameterIndex;
    $m.misc.configure add command -label [_ "Select Font"] \
	-command fontsel::CreateFontControlPanel;

    incr ParameterIndex;
    $m.misc.configure add checkbutton -label [_ "Balloon Help"]  -command ControlBalloonHelp \
	-onvalue 1 -offvalue 0 -variable BalloonHelpToggle
    set BalloonHelpToggle $BalloonHelpP
    ControlBalloonHelp;

    incr ParameterIndex;
    $m.misc.configure add checkbutton -label [_ "Fold Case"] \
	-onvalue 1 -offvalue 0 -variable FoldCaseP;

    incr ParameterIndex;
    $m.misc.configure add checkbutton -label [_ "Map Longest First"] \
	-onvalue 1 -offvalue 0 -variable MapLongestFirstP;

    incr ParameterIndex;
    $m.misc.configure add checkbutton -label [_ "Use Delimiter 0xFDDF in Transliteration Files"] \
	-onvalue \uFDDF -offvalue \u0009 -variable TransliterationDelimiter;

    #incr ParameterIndex;
    #set FocusFollowsMouseIndex $ParameterIndex;
    #$m.misc.configure add command -label [_ "Focus Follows Mouse"] -command MakeFocusFollowMouse
    #Note. We delete this menu item once it is used since it is irreversible.
    #If any following item on this submenu needs to set an index, this will disrupt the
    #index values, so be careful when changing things.

    incr MiscIndex;
    set SearchIndex 0;
    if {$AquaP} {incr SearchIndex -1}
    $m.misc add cascade -label [_ "Search"] -menu [menu $m.misc.search];

    incr SearchIndex;
    $m.misc.search add command -label [_ "Search Input"]  -command search::SearchInput;

    incr SearchIndex;
    $m.misc.search add command -label [_ "Search Results"]  -command search::SearchResults;

    incr SearchIndex;
    $m.misc.search add command -label [_ "Search Transliteration LHS"]  -command search::SearchXLLHS;

    incr SearchIndex;
    $m.misc.search add command -label [_ "Search Transliteration RHS"]  -command search::SearchXLRHS;

    incr SearchIndex;
    $m.misc.search add command -label "Go to Character in Input" -command search::GotoCharInput

    incr SearchIndex;
    $m.misc.search add command -label "Go to Character in Results" -command search::GotoCharResults

    incr SearchIndex;
    $m.misc.search add command -label "Go to Character in Transliteration LHS" -command search::GotoCharXLLHS

    incr SearchIndex;
    $m.misc.search add command -label "Go to Character in Transliteration RHS" -command search::GotoCharXLRHS

    incr MiscIndex;
    $m.misc add command -label [_ "Restore Saved Transliteration"] -command RestoreSavedTransliteration

    incr MiscIndex;
    $m.misc add command -label [_ "Swap Transliteration Input and Output"] -command SwapXLInOut;

    incr MiscIndex;
    set ToggleDuplicationDialogIndex $MiscIndex;
    $m.misc add command -label [_ "Display Duplication Dialog"] -command ToggleDuplicationDialog

    incr MiscIndex;
    set ToggleParseDialogIndex $MiscIndex;
    $m.misc add command -label [_ "Display Parse Dialog"]  -command ToggleParseDialog

    incr TopIndex;
    set HelpIndex 0;
    if {$AquaP} {incr HelpIndex -1}
    $m add cascade -label [_ "Help"]   -menu [menu $m.help];

    incr HelpIndex;
    $m.help add command -label [_ "About"] -command About;
    incr HelpIndex;
    $m.help add command -label [_ "Bug Reports"] -command BugReports;
    incr HelpIndex;
    $m.help add command -label [_ "Command Line Options"] -command CommandLineOptions;
    incr HelpIndex;
    $m.help add command -label [_ "Home Page"] -command {ShowWebPage http://www.billposer.org/Software/xlit.html}
    incr HelpIndex;
    $m.help add command -label [_ "Illustrated Web Manual"] -command WebManual;
    incr HelpIndex;
    $m.help add cascade -label [_ "Initialization File"]  -menu [menu $m.help.ifc];
    $m.help.ifc add command -label [_ "Color Names"] -command PopupColorNameList;
    $m.help.ifc add command -label [_ "Commands Available for Shortcuts"] \
	-command PopupShortcutCommands
    $m.help.ifc add command -label [_ "Tcl Commands - Popup"] \
	-command PopupInitTclCommandList;
    $m.help.ifc add command -label [_ "Tcl Commands - Save to File"] \
	-command SaveInitTclCommands;
    $m.help.ifc add command -label [_ "Xlit Commands - Popup"] \
	-command PopupInitCommandList;
    $m.help.ifc add command -label [_ "Xlit Commands - Save to File"] \
	-command SaveInitCommands;

    incr HelpIndex;
    $m.help add command -label [_ "Key Bindings"] -command DescribeKeyBindings;
    incr HelpIndex;
    $m.help add command -label [_ "License"] -command ShowGPL;
    . configure -menu .menubar

    if {$AquaP} {
	frame .com;
	button .com.trans -text [_ "Transliterate"] -command Transliterate \
	    -bg $::ColorSpecs(Menubar,Background) -fg $::ColorSpecs(Menubar,Foreground)
	button .com.abort -text [_ "Abort"] -command Abort \
	    -bg $::ColorSpecs(Menubar,Background) -fg $::ColorSpecs(Menubar,Foreground)
	pack .com.trans .com.abort .com.parse -side left -expand 1 -fill both
	pack .com -side top -expand 1 -fill both
    }

    #Create message region

    frame .mf 
    text $::MSG -bg $ColorSpecs(Message,Background) -fg $ColorSpecs(Message,Foreground) -height 1\
	-relief flat  -font MainFont
    Progress $::PBAR 150

    PackMessageRegion;
    pack .mf -side top -expand 0 -fill x -padx 3 -pady {3 0}

    set XLHeight [expr 2 * $MainHeight];

    proc UpdateEntryLineNumbers {} {
	.cf.num configure -state normal;
	.cf.num delete 1.0 end
	set width [string length $:::XLRulesInWindows]
	if {$width < 2} {set width 2}
	.cf.num configure -width $width
	set TopEntry [lindex [split [$::XLOUT index @1,1] "."] 0];
	for {set i 0} {$i < $::XLHeight} {incr i} {
	    .cf.num insert insert \
		[format "%0*d" $width [expr $i + $TopEntry]]
	}
	.cf.num configure -state disabled
    }

    frame .cf  -background $ColorSpecs(MainFrame,Background)
    scrollbar .cf.sbarleft  -command {YViewBoth}\
	-troughcolor $ColorSpecs(TransliterationIn,Background) \
	-bg $ColorSpecs(TransliterationIn,Background) \
	-activebackground $ColorSpecs(TransliterationIn,ActiveBackground);
    scrollbar .cf.sbarright -command {YViewBoth}\
	-troughcolor $ColorSpecs(TransliterationOut,Background) \
	-bg $ColorSpecs(TransliterationOut,Background) \
	-activebackground $ColorSpecs(TransliterationOut,ActiveBackground);
    text $::XLINP  -yscrollcommand {SetBoth} -width $MainWidth -height $MainHeight\
	-bg $ColorSpecs(TransliterationIn,Background) \
	-fg $ColorSpecs(TransliterationIn,Foreground) \
	-exportselection 1 -font LHSFont
    text $::XLOUT  -yscrollcommand {SetBoth} -width $MainWidth -height $MainHeight\
	-bg $ColorSpecs(TransliterationOut,Background) \
	-fg $ColorSpecs(TransliterationOut,Foreground) \
	-exportselection 1 -font RHSFont
    text .cf.num -width 3 -height 12 \
	-bg $ColorSpecs(TransliterationLineNumbers,Background) \
	-fg $ColorSpecs(TransliterationLineNumbers,Foreground) \
	-exportselection 0 -font MainFont
    UpdateEntryLineNumbers;

    frame .cf.ctr;
    proc AssembleControls {} {
	pack .cf.ctr.ebx.ent1 -side left -expand 1 -fill both -padx 1 -anchor w
	pack .cf.ctr.ebx.ent2 -side right -expand 1 -fill both -padx 1 -anchor e
	pack .cf.ctr.ebx -side top  -expand 1 -fill both -padx 1
	pack .cf.ctr.shw -side left -expand 1 -fill both -padx 1
	pack .cf.ctr.del -side top -expand 1 -fill both -padx 3 -pady 2
	pack .cf.ctr.mov -side top -expand 1 -fill both -padx 3 -pady 2
	pack .cf.ctr.shw -side top -expand 1 -fill both -padx 3 -pady 2
	if {$::DebugP} {
	    pack .cf.ctr.dmp -side top -expand 1 -fill both -padx 3 -pady 2;
	}
	pack .cf.ctr.und -side top -expand 1 -fill both -padx 3 -pady 2;
    }

    option add *cf*Button.background \#B6AFAF 80

    button .cf.ctr.del -text [_ "Delete Entry"] -command CmdDeleteEntry;
    button .cf.ctr.mov -text [_ "Move Entry"] -command CmdMoveEntry;
    button .cf.ctr.shw -text [_ "Show Entry"] -command CmdShowEntry;
    if {$DebugP} {
	button .cf.ctr.dmp -text [_ "Show Stack"] -command DumpEditStack;
    }
    button .cf.ctr.und -text [_ "Undo"] -command Undo;
    frame .cf.ctr.ebx
    entry .cf.ctr.ebx.ent1 -width 5
    entry .cf.ctr.ebx.ent2 -width 5
    .cf.ctr.ebx.ent1 insert end "1"
    .cf.ctr.ebx.ent2 insert end "1"

    set msg [_ "Scroll so that the entry at the line number\nin the left-hand box is visible and highlight it."]
    balloonhelp_for .cf.ctr.shw $msg
    set msg [_ "Delete the entry at the line number\nin the left-hand box."]
    balloonhelp_for .cf.ctr.del $msg
    set msg [_ "Move the entry at the line number in the left-hand\nbox so that it immediately precedes the entry at the\nline number in the right-hand box."]
    balloonhelp_for .cf.ctr.mov $msg
    set msg [_ "Undo the previous deletion or move. Pressing this button\nagain will undo the action before that."]
    balloonhelp_for .cf.ctr.und $msg

    proc PackXL {} {
	if {$::UseScrollbarsP} {
	    pack .cf.sbarleft  -side left    -expand 0 -fill y
	}
	if {$::DisplayEditingControlsP} {
	    pack .cf.ctr      -side right    -expand 0 -fill both
	}
	if {$::UseScrollbarsP} {
	    pack .cf.sbarright -side right   -expand 0 -fill y
	}
	pack $::XLINP     -side left    -expand 1 -fill both
	pack .cf.num      -side left    -expand 0 -fill none
	pack $::XLOUT    -side right   -expand 1 -fill both
	pack .cf -expand 1 -fill both
    }

    proc UnpackXL {} {
	pack forget .cf.ctr;
	pack forget .cf.sbarleft
	pack forget .cf.sbarright
	pack forget $::XLINP;
	pack forget .cf.num
	pack forget $::XLOUT;
    }

    AssembleControls;
    PackXL;

#This is the place for entering external commands.
    frame .eca
    label .eca.lab -text "External command"
    entry .eca.ent -relief flat
    .eca.ent insert end [join $ExternalCommandList " "]
    pack .eca.lab -side left -expand 0 -fill none -anchor w
    pack .eca.ent -side left -expand 1 -fill x -padx {5 0}
    bind .eca.ent <Return> GetExternalCommandFromEntry

#This is the place for entering plugins
    frame .plg
    frame .plg.path
    frame .plg.pnam
    label .plg.path.lab -text "File containing plugin"
    label .plg.pnam.lab -text "Name of plugin"
    entry .plg.path.ent -relief flat
    entry .plg.pnam.ent -relief flat
    set btxt [_ "Browse"]
    set blen [string length $btxt];
    button .plg.path.brb  -text $btxt -width $blen -anchor w -command SelectPluginFile
    .plg.path.ent insert end $::PluginPath
    .plg.pnam.ent insert end  $::PluginName
    pack .plg.path.lab -side left -expand 0 -fill none -anchor w
    pack .plg.path.brb  -expand 0 -anchor w -side left  -padx {8 3}
    pack .plg.path.ent -side left -expand 1 -fill x -padx {5 0}
    pack .plg.pnam.lab -side left -expand 0 -fill none -anchor w
    pack .plg.pnam.ent -side left -expand 1 -fill x -padx {5 0}
    pack .plg.pnam -side top -expand 1 -fill x
    pack .plg.path -side top -expand 1 -fill x
    bind .plg.path.ent <Return> GetPluginFromEntry
    bind .plg.pnam.ent <Return> GetPluginFromEntry

    proc SelectPluginFile {} {
	set File [tk_getOpenFile];
	if {$File == ""} {
	    ShowMessage [_ "File selection aborted."]
	    .plg.path.ent delete 0 end
	    return
	} else { 
	    .plg.path.ent delete 0 end
	    .plg.path.ent insert 0 [MinimizeFileName $File]
	}
    }

#These are the windows containing text to be transliterated and the result
    frame .txt  -background $ColorSpecs(MainFrame,Background)
    #These are used only in vertically stacked mode.
    frame .txt.vwi
    frame .txt.vwo

    text $::TXTINP -yscrollcommand {$::TXTISB set}  -height 8\
	-bg $ColorSpecs(InputText,Background) -exportselection 1 -font LHSFont -wrap word
    text $::TXTOUT -yscrollcommand {$::TXTOSB set}  -height 8\
	-bg $ColorSpecs(OutputText,Background) -exportselection 1 -font RHSFont -wrap word
    scrollbar $::TXTISB -command {$::TXTINP yview} -troughcolor $ColorSpecs(InputText,Background)\
	-bg $ColorSpecs(InputText,Background)  -activebackground $ColorSpecs(InputText,Background);
    scrollbar $::TXTOSB -command {$::TXTOUT yview}  -troughcolor $ColorSpecs(OutputText,Background)\
	-bg $ColorSpecs(OutputText,Background)  -activebackground $ColorSpecs(OutputText,Background);

    PackMain;
    LayoutTextWindows;

    set InsertionTarget $::XLINP;
    BindKeys $::XLINP;
    BindKeys $::XLOUT;
    BindKeys $::TXTINP;
    BindKeys $::TXTOUT;
    BindKeys $::MSG;

    bind $::XLINP <KeyPress> RecordXLGenTime;
    bind $::XLOUT <KeyPress> RecordXLGenTime;
    bind $::XLINP  <Tab>	{focus -force $::XLOUT;break}
    bind $::XLOUT  <Tab>	{focus -force $::XLINP;break}
    bind $::XLINP <FocusIn> {SetInsertionTargets $::XLINP}
    bind $::XLOUT <FocusIn> {SetInsertionTargets $::XLOUT}
    bind $::TXTINP <FocusIn> {SetInsertionTargets $::TXTINP}
    bind $::TXTISB <<B3>> "ScrollbarMoveBigIncrement $::TXTISB 0.20 %x %y"
    bind $::TXTOSB <<B3>> "ScrollbarMoveBigIncrement $::TXTOSB 0.20 %x %y"
    bind .cf.sbarright <<B3>> "ScrollbarMoveBigIncrement .cf.sbarright 0.20 %x %y"
    bind .cf.sbarleft <<B3>> "ScrollbarMoveBigIncrement .cf.sbarleft 0.20 %x %y"
    bind $::XLINP <Control-v> InsertSelection;
    bind $::XLOUT <Control-v> InsertSelection;
    bind $::TXTINP <Control-v> InsertSelection;
    bind $::XLINP <<B3>> InsertSelection;
    bind $::XLOUT <<B3>> InsertSelection;
    bind $::TXTINP <<B3>> InsertSelection;
    bind $::TXTINP <Button-1> {+focus %W}

    bind $::XLINP <ButtonPress-1> {+GetEntryNumber %W %x %y}
    bind $::XLOUT <ButtonPress-1> {+GetEntryNumber %W %x %y}

    bind $::CTRLENT1 <FocusIn>  [list set ::CurrentControlEntry $::CTRLENT1]
    bind $::CTRLENT2 <FocusIn>  [list set ::CurrentControlEntry $::CTRLENT2]

    proc GetEntryNumber {w x y} {
	set ::CurrentEntryNumber [lindex [split [$w index @$x,$y] "."] 0];
    }

    trace add variable CurrentEntryNumber write {SetEntryNumberInControls}
    set CurrentControlEntry $CTRLENT1;

    proc SetEntryNumberInControls {n e o} {
	set cw $::CurrentControlEntry;
	$cw delete 0 end
	$cw insert 0 $::CurrentEntryNumber
    } 

    balloonhelp_for $::XLINP   [_ "Enter strings to be transliterated"]
    balloonhelp_for $::XLOUT   [_ "Enter transliterations."]
    balloonhelp_for $::TXTINP  [_ "Text to be transliterated"]
    balloonhelp_for $::TXTOUT  [_ "Result of transliteration"]
    balloonhelp_for $::MSG     [_ "Messages from the program appear here."]
    balloonhelp_for .eca.ent   [_ "Enter external command to be executed here."]
    balloonhelp_for .plg.path.ent   [_ "Enter the name of the file containing the plugin here."]
    balloonhelp_for .plg.pnam.ent   [_ "Enter the name of the plugin procedure here."]

    $::XLINP configure -state normal
    $::XLOUT configure -state normal
    $::TXTINP configure -state disabled
    $::TXTOUT configure -state disabled
    search::SearchInitialize $::XLINP;
    search::SearchInitialize $::XLOUT;
    search::SearchInitialize $::TXTINP;
    search::SearchInitialize $::TXTOUT;

    tk_focusFollowsMouse;
    wm title . [format "xlit %s \[%s\]" $Version [ProgramTimeDateStamp]]
    . configure -background  $ColorSpecs(MainFrame,Background)
    after idle {
	update idletasks
	set w [winfo reqwidth  .]
	set h [winfo reqheight .]
	wm minsize  . $w $h
    }

    foreach fn $CustomCharacterChartFileNameList {
	wm withdraw [ReadCustomCharacterChartPopup $fn]
    }
    foreach dl $CustomCharacterChartDataList {
	wm withdraw [DefineCustomCharacterChartPopup $dl]
    }

    #If the default browser is on the list, remove it.
    set di [lsearch -exact $BrowserList $DefaultBrowser]
    if {$di >= 0} {
	set BrowserList [lreplace $BrowserList $di $di]
    }
    #Add the default browser to the beginning of the list.
    set BrowserList [linsert $BrowserList 0 $DefaultBrowser];

    proc ControlXLDisplay  {e o n} {
	PackMain
	if {$::WhichProcedure > 0} {
	    UnpackXL
	} else {
	    PackXL
	}
    }

    trace add variable WhichProcedure write ControlXLDisplay
}

if {$TkPresentP} {
    ControlXLDisplay x y z
}

CreateInitProcs;
interp create -safe -- safeinterp
interp share {} stderr safeinterp;# Allow plugins to write on stderr
if {$ReadInitFileP} {
    InitFileSetup;
    #Look first in current directory, then in home directory.
    if { [file exists $InitFile] } {
	ReadInitFile $InitFile;
    } else {
	set cwd [pwd];
	cd;
	if {[file exists $InitFile] } {
	    ReadInitFile $InitFile;
	}
	cd $cwd;
    }
}

if {[string equal $CommandLineTextFile ""] == 0} {
    LoadInputText $CommandLineTextFile;
}

if {[string equal $CommandLineTransliterationDefinition ""] == 0} {
    LoadTransliterationFile ParseNativeTransliterationEntry $CommandLineTransliterationDefinition;
}

if {[string equal $CommandLineLHSTransliterationDefinition ""] == 0} {
    if {$BatchModeP} {
	ShowMessage [_ "The option of loading just the LHS is not available in batch mode."]
    } else { 
	LoadTransliterationLHS $CommandLineLHSTransliterationDefinition;
    }
}

if {$BatchModeP} {
    Transliterate
    if {$::ParseErrorCnt} {
	puts stderr [format "Encountered %d parse errors.\n" $::ParseErrorCnt]
    }
    exit 0;
}
